SCNet, or SCN, is a new artificial neural networks architecture that applies Euclidean distance and consensus similarity for vector matching plus deep learning without backpropagation using division calculations in textual pattern recognition, enabling the development of optimized and efficient large language models (LLMs).

# SCNet

The SCNet is a neural network architecture created by Ben-Hur Varriano focused on the development of language models. It is an extension of the "[Semantic Comparison Network](https://github.com/sapiens-technology/semantic_comparison_network)" module from Sapiens Technology®️ and has additional features for greater generalization and extended capacity in training and inference, enabling the processing of large volumes of data by "streaming" without memory overflow on limited "hardware". The code uses proximity calculations and a [HurNet](https://github.com/sapiens-technology/HurNet) neural network in pattern recognition. Its approximation calculation does not use "backpropagation" and both training and fine-tuning have a discrepant speedup when compared to conventional architectures such as "RNNs" and "Transformers". Its context window is infinite, with no limitations on the length of "embeddings" in the training and tuning phase, in addition to compressing the input vectors when necessary. With its advanced mathematics and sophisticated logic, its concept is innovative and revolutionary, with the potential to become a new paradigm in language models.


## Installation

Use the package manager [pip](https://pip.pypa.io/en/stable/) to install the Semantic Comparison Network.

```bash
pip install scnetwork
```

## Usage
Basic usage example:
```python
from scnetwork import SCNet # imports the main class "SCNet" from the module "scnetwork"
scnet = SCNet() # assigns the "scnet" variable the instantiation of the "SCNet" class, constructing the object with "()"
# now just use the "scnet" object to access the class's resources
# use the "addFit" method to add input and output pairs to the training dataset
# assign a sample input in the first argument and a desired output in the second argument, calling the same method as many times as necessary
scnet.addFit('Hello! Who are you?', 'Hello! I am Sapiens Chat, an AI model created by Sapiens Technology.')
scnet.addFit('Who discovered Brazil?', 'Brazil was discovered by Portuguese navigators led by Pedro Alvares Cabral in 1500.')
scnet.addFit('What is the main language spoken in Spain?', 'The main language spoken in Spain is Spanish.')
scnet.train() # use the "train" method to train the model with the assigned dataset
# with the "print_predict" method you can display the output for a given input
scnet.print_predict('What is the main language spoken in Spain?')

```
Prediction response:
```bash
Reading data: 100%|██████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 1/1 [00:00<00:00, 8848.74it/s]
Embedding data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████████████████| 286/286 [00:00<00:00, 4904.74it/s]
Tokenizing data: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 99.51it/s]
Training model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 115.57it/s]
The main language spoken in Spain is Spanish.
```
```python
from scnetwork import SCNet
scnet = SCNet()

scnet.addFit('Hello! Who are you?', 'Hello! I am Sapiens Chat, an AI model created by Sapiens Technology.')
scnet.addFit('Who discovered Brazil?', 'Brazil was discovered by Portuguese navigators led by Pedro Alvares Cabral in 1500.')
scnet.addFit('What is the main language spoken in Spain?', 'The main language spoken in Spain is Spanish.')
scnet.train(progress=False) # set "False" to the "progress" parameter if you want to hide the training progress
# assign in the first argument the input for which you want to obtain an output
# the training input output that is most semantically similar to the prediction input, will be returned
scnet.print_predict('What is the main language spoken in Spain?') # if the answer is wrong, add more data to your training

```
Prediction response:
```bash
The main language spoken in Spain is Spanish.
```

You can also use the "string" training parameter to add input and output pairs to your dataset.
```python
from scnetwork import SCNet
scnet = SCNet()
# variable with the training data set (can receive strings of any size)
# place each output below each input and separate the pairs with a blank line
# to increase the quality of the answers, preferably use two or more different examples for the same output
dataset = '''
What does the acronym GPT mean in language models?
GPT in language models stands for Generative Pre-trained Transformer.

Tell me what is the meaning of GPT in language models.
GPT in language models stands for Generative Pre-trained Transformer.

What is the capital of Spain?
The capital of Spain is Madrid.

Tell me the name of the capital of Spain.
The capital of Spain is Madrid.
'''
scnet.train(string=dataset) # uses the value of the "dataset" variable as the training database
# call the prediction method as many times as you want
scnet.print_predict('What does the acronym GPT mean in language models?')
scnet.print_predict('What is the capital of Spain?')
scnet.print_predict('Tell me the name of the capital of Spain.')

```
```bash
Reading data: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 1/1 [00:00<00:00, 11096.04it/s]
Embedding data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████████████████| 384/384 [00:00<00:00, 6653.39it/s]
Tokenizing data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 101.83it/s]
Training model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 118.23it/s]
GPT in language models stands for Generative Pre-trained Transformer.
The capital of Spain is Madrid.
The capital of Spain is Madrid.
```
You can combine multiple data sources to build your model by inputting data using the "addFit" and "train" methods.
```python
from scnetwork import SCNet
scnet = SCNet()
# it is also possible to join the data from the "string" parameter with the pairs from the "addFit" method in a single training
# when possible, always use multiple input examples for the same output
dataset = '''
What does the acronym GPT mean in language models?
GPT in language models stands for Generative Pre-trained Transformer.

Tell me what is the meaning of GPT in language models.
GPT in language models stands for Generative Pre-trained Transformer.
'''
# in the "addFit" method also give preference to add multiple inputs to the same output
scnet.addFit('What is the capital of Spain?', 'The capital of Spain is Madrid.')
scnet.addFit('Tell me the name of the capital of Spain.', 'The capital of Spain is Madrid.')
scnet.train(string=dataset)

scnet.print_predict('What does the acronym GPT mean in language models?')
scnet.print_predict('What is the capital of Spain?')
scnet.print_predict('Tell me the name of the capital of Spain.')

```
```bash
Reading data: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 1/1 [00:00<00:00, 12826.62it/s]
Embedding data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████████████████| 384/384 [00:00<00:00, 6523.95it/s]
Tokenizing data: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 99.92it/s]
Training model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 116.02it/s]
GPT in language models stands for Generative Pre-trained Transformer.
The capital of Spain is Madrid.
The capital of Spain is Madrid.
```
The best way to train a language model is through structured data in JSON format. The file should have a key named "data" with a list of dictionaries containing input and output strings with the keys named "input" and "output". You can work with files of any size, the limitation will be determined by the capacity of your hardware.

We will use as an example a file named "dataset.json" with the content below:
```json
{
	"data": [
				{
					"input": "Hello! Who are you?",
					"output": "Hello! I am Sapiens Chat, an AI model created by Sapiens Technology."
				},
				{
					"input": "Who discovered Brazil?",
					"output": "Brazil was discovered by Portuguese navigators led by Pedro Alvares Cabral in 1500."
				},
				{
					"input": "What is the main language spoken in Spain?",
					"output": "The main language spoken in Spain is Spanish."
				},
				{
					"input": "What is the capital of Portugal?",
					"output": "The capital of Portugal is Lisbon."
				},
				{
					"input": "How much is 2 + 2?",
					"output": "The sum of 2 + 2 in mathematics is 4."
				},
				{
					"input": "What is five minus three?",
					"output": "The result of five minus three is two."
				},
				{
					"input": "What is your name?",
					"output": "My name is Sapiens Chat."
				}
	]
}
```
```python
from scnetwork import SCNet
scnet = SCNet()
# use the "dataset_path" parameter to assign the path of a "json", "csv" or "txt" file as the training database
scnet.train(dataset_path='dataset.json')
scnet.print_predict('What is the main language spoken in Spain?')

```
```bash
Reading data: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 1/1 [00:00<00:00, 10754.63it/s]
Embedding data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████| 1000000/1000000 [00:00<00:00, 17477067.70it/s]
Tokenizing data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 102.52it/s]
Training model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 119.07it/s]
The main language spoken in Spain is Spanish.
```
You can also supplement the training data with the "addFit" method.
```python
from scnetwork import SCNet
scnet = SCNet()
# call the "addFit" method before the "train" method to add more data to your dataset
scnet.addFit('What is the capital of Spain?', 'The capital of Spain is Madrid.')
scnet.addFit('Tell me the name of the capital of Spain.', 'The capital of Spain is Madrid.')
scnet.train(dataset_path='dataset.json')
scnet.print_predict('What is the capital of Spain?')
scnet.print_predict('What is the main language spoken in Spain?')

```
```bash
Reading data: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 1/1 [00:00<00:00, 12520.31it/s]
Embedding data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████████████████| 136/136 [00:00<00:00, 2288.03it/s]
Tokenizing data: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 98.72it/s]
Training model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 114.65it/s]
The capital of Spain is Madrid.
The main language spoken in Spain is Spanish.
```
In addition to the "addFit" method, the "string" parameter can also be used as a complement of data.
```python
from scnetwork import SCNet
scnet = SCNet()
# you can also supplement the file data with the data from the "string" parameter
string = '''
What does the acronym GPT mean in language models?
GPT in language models stands for Generative Pre-trained Transformer.

Tell me what is the meaning of GPT in language models.
GPT in language models stands for Generative Pre-trained Transformer.
'''
scnet.train(dataset_path='dataset.json', string=string)
scnet.print_predict('Tell me what is the meaning of GPT in language models.')
scnet.print_predict('What is the main language spoken in Spain?')

```
```bash
Reading data: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 1/1 [00:00<00:00, 13025.79it/s]
Embedding data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████████████████| 246/246 [00:00<00:00, 4189.03it/s]
Tokenizing data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 100.07it/s]
Training model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 116.23it/s]
GPT in language models stands for Generative Pre-trained Transformer.
The main language spoken in Spain is Spanish.
```
You can combine different ways of inserting data when assembling your dataset. Below we are combining data coming from a file, with data from variables.
```python
from scnetwork import SCNet
scnet = SCNet()
# both the "string" parameter and the "addFit" method can be used together or separately when inserting additional data together with the file
string = '''
What does the acronym GPT mean in language models?
GPT in language models stands for Generative Pre-trained Transformer.

Tell me what is the meaning of GPT in language models.
GPT in language models stands for Generative Pre-trained Transformer.

What is the closest star?
The closest star is the Sun.
'''
scnet.addFit('What is the capital of Spain?', 'The capital of Spain is Madrid.')
scnet.addFit('Tell me the name of the capital of Spain.', 'The capital of Spain is Madrid.')
scnet.train(dataset_path='dataset.json', string=string)
scnet.print_predict('What is the main language spoken in Spain?')
scnet.print_predict('Tell me the name of the capital of Spain.')
scnet.print_predict('What is the closest star?')

```
```bash
Reading data: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 1/1 [00:00<00:00, 12052.60it/s]
Embedding data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████████████████| 440/440 [00:00<00:00, 7569.89it/s]
Tokenizing data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 101.02it/s]
Training model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 117.28it/s]
The main language spoken in Spain is Spanish.
The capital of Spain is Madrid.
The closest star is the Sun.
```
The "dataset_path" parameter in the training method can also receive unstructured files in TXT format. But still the preferable and most reliable format remains the JSON format.

We will use as an example a file named "dataset.txt" with the content below:
```txt
What is Artificial Intelligence (AI)?
Artificial Intelligence is a field of computer science that develops systems capable of performing tasks that would normally require human intelligence, such as learning, reasoning, perception, and decision-making.

What is the difference between weak AI and strong AI?
Weak AI (or narrow AI) is designed for specific tasks, such as virtual assistants and recommendation systems. Strong AI, on the other hand, would have self-awareness and the ability to understand and reason about any subject like a human being.

What is Machine Learning and how is it related to AI?
Machine Learning is a subfield of AI that teaches machines to learn patterns from data without explicit programming. It allows systems to improve their performance over time with experience.

What are the main types of machine learning?
The three main types are:
- **Supervised Learning**: The model learns from labeled data.
- **Unsupervised Learning**: The model identifies patterns without labels.
- **Reinforcement Learning**: The model learns through trial and error, receiving rewards or penalties.

What are artificial neural networks?
They are models inspired by the functioning of the human brain, composed of layers of artificial neurons that process information and adjust their weights to recognize patterns and make decisions.

What are the risks of Artificial Intelligence?
Some risks include algorithmic bias, job loss due to automation, misuse in surveillance or warfare, and the possibility of superintelligence beyond human control.

How is AI used in everyday life?
AI is present in virtual assistants (Siri, Alexa), recommendation systems (Netflix, Spotify), facial recognition, autonomous cars, medical diagnostics, chatbots, and much more.

What is natural language processing (NLP)?
It is a field of AI that enables machines to understand, interpret, and generate human language, used in machine translation, chatbots, and voice assistants.

Can AI replace humans in the workforce?
AI can automate repetitive and analytical tasks, but it is unlikely to fully replace humans in creative, emotional, and critical thinking jobs.  

What is a generative AI model?
It is a type of AI that can create new content, such as images, text, and music, based on patterns learned from large amounts of data. Examples include ChatGPT and DALL·E.

```
```python
from scnetwork import SCNet
scnet = SCNet()
scnet.train(dataset_path='dataset.txt')
scnet.print_predict('What is Artificial Intelligence (AI)?')

```
```bash
Reading data: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 1/1 [00:00<00:00, 12018.06it/s]
Embedding data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████| 1000000/1000000 [00:00<00:00, 16033578.62it/s]
Tokenizing data: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 93.81it/s]
Training model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 109.02it/s]
Artificial Intelligence is a field of computer science that develops systems capable of performing tasks that would normally require human intelligence, such as learning, reasoning, perception, and decision-making.
```
Remember that it is possible to complement the file data with the "string" parameter and/or the "addFit" method.
```python
from scnetwork import SCNet
scnet = SCNet()

string = '''
What does the acronym GPT mean in language models?
GPT in language models stands for Generative Pre-trained Transformer.

Tell me what is the meaning of GPT in language models.
GPT in language models stands for Generative Pre-trained Transformer.

What is the closest star?
The closest star is the Sun.
'''
scnet.addFit('What is the capital of Spain?', 'The capital of Spain is Madrid.')
scnet.addFit('Tell me the name of the capital of Spain.', 'The capital of Spain is Madrid.')
scnet.train(dataset_path='dataset.txt', string=string)
scnet.print_predict('What is Artificial Intelligence (AI)?')
scnet.print_predict('Tell me the name of the capital of Spain.')
scnet.print_predict('What is the closest star?')

```
```bash
Reading data: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 1/1 [00:00<00:00, 10106.76it/s]
Embedding data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████████████████| 440/440 [00:00<00:00, 7043.95it/s]
Tokenizing data: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 94.26it/s]
Training model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 109.26it/s]
Artificial Intelligence is a field of computer science that develops systems capable of performing tasks that would normally require human intelligence, such as learning, reasoning, perception, and decision-making.
The capital of Spain is Madrid.
The closest star is the Sun.
```
You can set "True" for the "stream" parameter of the prediction method when you want to display the response token by token.
```python
from scnetwork import SCNet
scnet = SCNet()
scnet.train(dataset_path='dataset.txt', progress=False) # progress=False inhibits the display of progress bars
scnet.print_predict('What is Artificial Intelligence (AI)?', stream=True) # stream=True display one token at a time

```
```bash
Artificial Intelligence is a field of computer science that develops systems capable of performing tasks that would normally require human intelligence, such as learning, reasoning, perception, and decision-making.
```
With the "interval" parameter you can define a minimum interval in seconds between the display of one token and another. This parameter will only work if "stream" is equal to True.
```python
from scnetwork import SCNet
scnet = SCNet()
# "dataset_path" receives the path to a structured file in "json" or "csv" format, or to an unstructured file in "txt" format
# "progress" is set to True to display progress or False to not display progress; the default value is True
scnet.train(dataset_path='dataset.txt', progress=False)
# interval=0.1 causes each generated token to wait 10% of 1 second before displaying the next token
# the default value of the "interval" parameter is 0.0, but you can set any positive floating value
scnet.print_predict('What is Artificial Intelligence (AI)?', stream=True, interval=0.1)

```
```bash
Artificial Intelligence is a field of computer science that develops systems capable of performing tasks that would normally require human intelligence, such as learning, reasoning, perception, and decision-making.
```
With the "precision" parameter you can define a percentage value between 0 and 1 for the default size of the "embeddings", referring to the proportion of the largest "embedding" found in the training data set. This can be especially useful if you need to reduce the processing load on limited hardware, but it will make the prediction responses less precision.
```python
from scnetwork import SCNet
scnet = SCNet()
# precision=0.8 makes all embeddings 80% the length of the largest embedding
# very small precision values may cause incomplete or inadequate predictions
scnet.train(dataset_path='dataset.txt', precision=0.8, progress=False)
scnet.print_predict('What is Artificial Intelligence (AI)?', stream=True, interval=0.1)

```
```bash
Artificial Intelligence is a field of computer science that develops systems capable of performing tasks that would normally require human intelligence, such as learning, reasoning, perception, and decision-making.
```
You can also work with structured files in CSV format with a column named "input", "prompt" or "question" for input and a column named "output", "answer" or "response" for output. The CSV file may contain other columns, but only these two will be read. Each line in the file will represent a pair of input and output. Always give preference to structured files in JSON or CSV format.

We will use as an example a file named "dataset.csv" with the content below:
```csv
"input","output"
"Hello! Who are you?","Hello! I am Sapiens Chat, an AI model created by Sapiens Technology."
"Who discovered Brazil?","Brazil was discovered by Portuguese navigators led by Pedro Alvares Cabral in 1500."
"What is the main language spoken in Spain?","The main language spoken in Spain is Spanish."
"What is the capital of Portugal?","The capital of Portugal is Lisbon."
"How much is 2 + 2?","The sum of 2 + 2 in mathematics is 4."
"What is five minus three?","The result of five minus three is two."
"What is your name?","My name is Sapiens Chat."
```
```python
from scnetwork import SCNet
scnet = SCNet()
# although "dataset_path" can handle "json", "csv" and "txt" files, always give preference to datasets in "json" format when possible
scnet.train(dataset_path='dataset.csv')
# interval=0.01 for 1% intervals of 1 second between one token and another
scnet.print_predict('What is the main language spoken in Spain?', stream=True, interval=0.01)

```
```bash
Reading data: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 1/1 [00:00<00:00, 12483.05it/s]
Embedding data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████| 1000000/1000000 [00:00<00:00, 17119747.92it/s]
Tokenizing data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 100.63it/s]
Training model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 116.87it/s]
The main language spoken in Spain is Spanish.
```
Now see an example complementing the file data with the "string" parameter and with the "addFit" method.
```python
from scnetwork import SCNet
scnet = SCNet()

string = '''
What does the acronym GPT mean in language models?
GPT in language models stands for Generative Pre-trained Transformer.

Tell me what is the meaning of GPT in language models.
GPT in language models stands for Generative Pre-trained Transformer.

What is the closest star?
The closest star is the Sun.
'''
scnet.addFit('What is the capital of Spain?', 'The capital of Spain is Madrid.')
scnet.addFit('Tell me the name of the capital of Spain.', 'The capital of Spain is Madrid.')
scnet.train(dataset_path='dataset.csv', string=string, progress=False)

stream = True
interval = 0.01
scnet.print_predict('What is the main language spoken in Spain?', stream=stream, interval=interval)
scnet.print_predict('Tell me the name of the capital of Spain.', stream=stream, interval=interval)
scnet.print_predict('What is the closest star?', stream=stream, interval=interval)

```
```bash
The main language spoken in Spain is Spanish.
The capital of Spain is Madrid.
The closest star is the Sun.
```
With the "tokenizer" name parameter contained in the training method, you can set the name of a specific type of tokenizer to be used in converting texts into embeddings chains.
```python
from scnetwork import SCNet
scnet = SCNet()
# "tokenizer" takes a string with the name of the encoder to be used for tokenization; the default value is 'gpt-4'
# tokenizers available: 'gpt-2', 'gpt-3', 'gpt-3.5-turbo', 'gpt-4', 'gpt-4-turbo', 'text-davinci-001', 'text-davinci-002' and 'text-davinci-003'
scnet.train(dataset_path='dataset.csv', tokenizer='gpt-2') # tokenizer='gpt-2' uses gpt-2 tokenizer
scnet.print_predict('Hello! Who are you?', stream=True, interval=0.01)

```
```bash
Reading data: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 1/1 [00:00<00:00, 13706.88it/s]
Embedding data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████| 1000000/1000000 [00:00<00:00, 11898802.26it/s]
Tokenizing data: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 70.40it/s]
Training model: 100%|██████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 81.91it/s]
Hello! I am Sapiens Chat, an AI model created by Sapiens Technology.
```
The "precision" parameter sets a specific percentage of the total length of the largest tokenized embedding and applies this same length to all embeddings.
```python
from scnetwork import SCNet
scnet = SCNet()
# precision=0.95 so that all embedding vectors have a size of 95% of the size of the largest embedding
# the default value of "precision" is 1.0, i.e. all embedding vectors will have the same length as 100% of the length of the largest vector
# always use float values between 0 and 1 for "precision"
scnet.train(dataset_path='dataset.csv', precision=0.95, tokenizer='gpt-4-turbo')
scnet.print_predict('What is five minus three?', stream=True, interval=0.01)

```
```bash
Reading data: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 1/1 [00:00<00:00, 13357.66it/s]
Embedding data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████| 1000000/1000000 [00:00<00:00, 16882767.06it/s]
Tokenizing data: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 99.28it/s]
Training model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 115.32it/s]
The result of five minus three is two.
```
With the "context_window" parameter of the training method, you can set an integer to limit the size of your model's context window. The size defined in this parameter will be applied to the length of all embeddings. By default, the context window size is infinite, which means that all embeddings will always use their maximum length.
```python
from scnetwork import SCNet
scnet = SCNet()
# in the example below we are applying a limit of 50 tokens to the size of the context window
# in this case, any prediction input with a length other than 50 tokens will have its size resized to 50
# the "context_window" parameter must always receive an integer greater than zero; the default value for "context_window" is float('inf'), that is, infinity
# the smaller your context window, the faster your training and inference will be
# the bigger your context window, the slow your training and inference will be
# larger context windows decrease the chances of incorrect responses, but smaller context windows increase the chances of your model issuing an inaccurate response
scnet.train(dataset_path='dataset.csv', context_window=50)
scnet.print_predict('What is the main language spoken in Spain?', stream=True, interval=0.01)

```
```bash
Reading data: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 1/1 [00:00<00:00, 12157.40it/s]
Embedding data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████| 1000000/1000000 [00:00<00:00, 17437395.81it/s]
Tokenizing data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 102.43it/s]
Training model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 118.93it/s]
The main language spoken in Spain is Spanish.
```
With the "end_tag" parameter of the "train" method, you can define a substring contained in the text of your file or training string to be used as a separator of the input and output pairs. In this case, the algorithm will divide your text into smaller parts, separated by the value of "end_tag" and consider that each of these parts contains a specific output and/or input.

We will use as an example a file named "dataset_with_tag.txt" with the content below:
```txt
What is Artificial Intelligence (AI)?
Artificial Intelligence is a field of computer science that develops systems capable of performing tasks that would normally require human intelligence, such as learning, reasoning, perception, and decision-making.
<|end|>
What is the difference between weak AI and strong AI?
Weak AI (or narrow AI) is designed for specific tasks, such as virtual assistants and recommendation systems. Strong AI, on the other hand, would have self-awareness and the ability to understand and reason about any subject like a human being.
<|end|>
What is Machine Learning and how is it related to AI?
Machine Learning is a subfield of AI that teaches machines to learn patterns from data without explicit programming. It allows systems to improve their performance over time with experience.
<|end|>
What are the main types of machine learning?
The three main types are:
- **Supervised Learning**: The model learns from labeled data.
- **Unsupervised Learning**: The model identifies patterns without labels.
- **Reinforcement Learning**: The model learns through trial and error, receiving rewards or penalties.
<|end|>
What are artificial neural networks?
They are models inspired by the functioning of the human brain, composed of layers of artificial neurons that process information and adjust their weights to recognize patterns and make decisions.
<|end|>
What are the risks of Artificial Intelligence?
Some risks include algorithmic bias, job loss due to automation, misuse in surveillance or warfare, and the possibility of superintelligence beyond human control.
<|end|>
How is AI used in everyday life?
AI is present in virtual assistants (Siri, Alexa), recommendation systems (Netflix, Spotify), facial recognition, autonomous cars, medical diagnostics, chatbots, and much more.
<|end|>
What is natural language processing (NLP)?
It is a field of AI that enables machines to understand, interpret, and generate human language, used in machine translation, chatbots, and voice assistants.
<|end|>
Can AI replace humans in the workforce?
AI can automate repetitive and analytical tasks, but it is unlikely to fully replace humans in creative, emotional, and critical thinking jobs.  
<|end|>
What is a generative AI model?
It is a type of AI that can create new content, such as images, text, and music, based on patterns learned from large amounts of data. Examples include ChatGPT and DALL·E.
<|end|>
```
```python
from scnetwork import SCNet
scnet = SCNet()
# the "end_tag" parameter will only take effect when "dataset_path" receives a "txt" file or when you are using the "string" parameter
# below we are considering the '<|end|>' tag contained in the example file as the value that separates one possible answer from the other, but you can use any substring you prefer
# the default value of the "end_tag" parameter is '\n\n', that is, the text will use two line breaks to separate the parts
scnet.train(dataset_path='dataset_with_tag.txt', end_tag='<|end|>')
scnet.print_predict('Explain artificial neural networks to me.', stream=True, interval=0.01)

```
```bash
Reading data: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 1/1 [00:00<00:00, 12300.01it/s]
Embedding data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████| 1000000/1000000 [00:00<00:00, 15905830.96it/s]
Tokenizing data: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 93.59it/s]
Training model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 108.81it/s]
They are models inspired by the functioning of the human brain, composed of layers of artificial neurons that process information and adjust their weights to me patterns and make decisions.
```
If you are using the "string" parameter together with a text file, both must have the same value of "end_tag" in their content.
```python
from scnetwork import SCNet
scnet = SCNet()
# below we are using the "string" parameter with different inputs for the same outputs of file, this way your model will become more accurate
# note that both the value of the "string" parameter and the content of the file in "dataset_path" have '<|end|>' as a separator
string = '''
Tell me the main types of machine learning.
The three main types are:
- **Supervised Learning**: The model learns from labeled data.
- **Unsupervised Learning**: The model identifies patterns without labels.
- **Reinforcement Learning**: The model learns through trial and error, receiving rewards or penalties.
<|end|>
Explain to me what Artificial Intelligence is.
Artificial Intelligence is a field of computer science that develops systems capable of performing tasks that would normally require human intelligence, such as learning, reasoning, perception, and decision-making.
<|end|>
'''
scnet.train(dataset_path='dataset_with_tag.txt', string=string, end_tag='<|end|>')
prompt = 'Explain artificial neural networks to me.'
print('Prompt:', prompt)
print('Answer: ', end='', flush=True)
scnet.print_predict('Explain artificial neural networks to me.', stream=True, interval=0.01)
print()
prompt = 'Tell me the main types of machine learning.'
print('Prompt:', prompt)
print('Answer: ', end='', flush=True)
scnet.print_predict('Tell me the main types of machine learning.', stream=True, interval=0.01)

```
```bash
Reading data: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 1/1 [00:00<00:00, 13189.64it/s]
Embedding data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████████████████| 589/589 [00:00<00:00, 9442.98it/s]
Tokenizing data: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 94.11it/s]
Training model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 109.41it/s]
Prompt: Explain artificial neural networks to me.
Answer: They are models inspired by the functioning of the human brain, composed of layers of artificial neurons that process information and adjust their weights to me patterns and make decisions.

Prompt: Tell me the main types of machine learning.
Answer: The three main types are:
- **Supervised Learning**: The model learns from labeled data.
- **Unsupervised Learning**: The model identifies patterns without labels.
- **Reinforcement Learning**: The model learns through trial and error, receiving rewards or penalties.
```
The "validate" parameter of the training method receives a float for a percentage between 0 and 1 that represents the part of the data set that will be separated for validation and testing of the model's learning at the end of the training process.
```python
from scnetwork import SCNet
scnet = SCNet()
# the "validate" parameter receives a float with the percentage of data that will be used to validate the learning
# the default value of "validate" is 0.0, i.e. there will be no validation upon completion of training
# in the scnet network the validation data will also be part of the training; in this case if you want you can use 1.0 to validate 100% of the data
# the larger the sample set aside for validation and testing, the longer it will take to complete the training method
# you can use the return of the "train" function to get your training metrics
training_metrics = scnet.train(dataset_path='dataset.txt', validate=0.2) # validate=0.2 so that 20% of the data is used for validation and testing
print(training_metrics) # displays training metrics
# the 'val_loss' key contains the percentage (between 0 and 1) of the amount of data separated for testing that failed validation (the lower the better)
# the 'loss' key contains the percentage (between 0 and 1) of information loss, that is, the percentage of data that was not fully learned (the lower the better)
# the 'generalization_rate' key contains the percentage (between 0 and 1) of the model's ability to generalize to inputs other than those used in training (the higher the better)
# the 'precision' key contains the percentage (between 0 and 1) of the model's final precision in relation to your possible ability to get the answers right (the higher the better)

```
```bash
Reading data: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 1/1 [00:00<00:00, 13189.64it/s]
Embedding data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████| 1000000/1000000 [00:00<00:00, 15794364.30it/s]
Tokenizing data: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 92.86it/s]
Validating model: 100%|████████████████████████████████████████████████████████████████████████████████████████████████████████| 426/426 [00:00<00:00, 154618.68it/s]
Training model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 103.33it/s]
{'val_loss': 0.2197099781488428, 'loss': 0.08435451958561202, 'generalization_rate': 0.7802900218511573, 'precision': 0.915645480414388}
```
Below is an example with all the training parameters and their default values.
```python
from scnetwork import SCNet
scnet = SCNet()

scnet.train( # train the model
	dataset_path='dataset.txt', # path to training file ("json", "csv" or "txt")
	string='', # complementary content
	precision=1.0, # percentage for the maximum length of the embeddings relative to the length of the largest embedding
	tokenizer='gpt-4', # tokenization standard
	context_window=float('inf'), # context window size
	end_tag='\n\n', # separator used to divide the training text into smaller parts
	validate=0.0, # percentage of the dataset separated for validation
	progress=True # enables or disables progress bars
)

scnet.print_predict(prompt='What is Artificial Intelligence (AI)?')

```
```bash
Reading data: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 1/1 [00:00<00:00, 13357.66it/s]
Embedding data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████| 1000000/1000000 [00:00<00:00, 16308446.03it/s]
Tokenizing data: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 95.95it/s]
Training model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 111.54it/s]
Artificial Intelligence is a field of computer science that develops systems capable of performing tasks that would normally require human intelligence, such as learning, reasoning, perception, and decision-making.
```
To save a trained model so that you can use it in the future on any machine without having to train it again, simply call the method named "saveModel" right after calling the "train" method. The "saveModel" method will save two files with your model: one file with the network settings and another with the weights.
```python
from scnetwork import SCNet
scnet = SCNet()
# training and saving the language model
scnet.train(dataset_path='dataset.json') # train the model
# the "saveModel" method will save two files with your model data: a ".sccon" file with the settings and another ".scnet" file with the weights
# if no name is specified in the "model_path" parameter of the saving method, the model files will be saved with the default name "model" in the local directory
scnet.saveModel() # save the model ("model.sccon" and "model.scnet" files)
# note in the progress bar that the model will be saved with 91U, that is, 91 unit parameters (U for Units, K for Thousands/Kilobytes, M for Millions, B for Billions and T for Trillions)

```
```bash
Reading data: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 1/1 [00:00<00:00, 12633.45it/s]
Embedding data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████| 1000000/1000000 [00:00<00:00, 17341580.39it/s]
Tokenizing data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 101.81it/s]
Training model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 118.28it/s]
Generating 91U parameters: 100%|████████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 38421.11it/s]
Saving model: 100%|██████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 5/5 [00:00<00:00, 4659.30it/s]
```
Once you save your model, you never have to go through the hassle of retraining it or waiting for the weights adjustment process. Simply use the "loadModel" method to load the pre-trained model files and make your predictions as normal.
```python
from scnetwork import SCNet
scnet = SCNet()
# the "loadModel" method will search for files of some pre-trained model in the local directory if no value is passed for the "model_path" parameter
# if no path is given to the "model_path" parameter, the "loadModel" method will look for model files in the local directory and load the first one it finds
scnet.loadModel() # loads a pre-trained model
scnet.print_predict('What is the main language spoken in Spain?') # performs model inference

```
```bash
Loading model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 4559.73it/s]
The main language spoken in Spain is Spanish.
```
With the "progress" parameter of the loading method, you can enable or disable the display of progress when loading the model.
```python
from scnetwork import SCNet
scnet = SCNet()
# use "progress" equal to False if you want to hide the loading progress; the default value of the "progress" parameter is True
# loading and inference of the pre-trained model
scnet.loadModel(progress=False) # progress=False disables the loading progress bar
scnet.print_predict('What is five minus three?')

```
```bash
The result of five minus three is two.
```
You can also use the "progress" parameter of the model save method to inhibit the display of progress when saving your pre-trained model.
```python
from scnetwork import SCNet
scnet = SCNet()
# use the "progress" parameter with False to disable progress of saving, or with True to enable; the default value of "progress" is True
scnet.train(dataset_path='dataset.json')
scnet.saveModel(progress=False) # progress=False disables the saving progress bars

```
```bash
Reading data: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 1/1 [00:00<00:00, 13148.29it/s]
Embedding data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████| 1000000/1000000 [00:00<00:00, 17095743.90it/s]
Tokenizing data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 100.51it/s]
Training model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 116.79it/s]
```
To also hide the training progress, simply set False to "progress" parameters of training and model saving.
```python
from scnetwork import SCNet
scnet = SCNet()

scnet.train(dataset_path='dataset.json', progress=False)
scnet.saveModel(progress=False)

```
```bash

```
You can also choose a name and/or path to save your pre-trained models files, to do this simply define the local address of your preference in the value of the "model_path" parameter of the saving method.
```python
from scnetwork import SCNet
scnet = SCNet()

scnet.train(dataset_path='dataset.json')
# the "model_path" parameter receives a string with the path and name of the model that will be generated in saving
# you can define any path and name you prefer
scnet.saveModel(model_path='my_model') # in this case the ".sccon" and ".scnet" files will be named "my_model"

```
```bash
Reading data: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 1/1 [00:00<00:00, 13357.66it/s]
Embedding data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████| 1000000/1000000 [00:00<00:00, 17187164.24it/s]
Tokenizing data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 100.99it/s]
Training model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 117.33it/s]
Generating 91U parameters: 100%|████████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 36578.23it/s]
Saving model: 100%|██████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 5/5 [00:00<00:00, 5127.51it/s]
```
When there are files for more than one model, you can specify the path of the model you want to load in the "model_path" parameter of the loading method. Just pass the path to the model directory, and/or the name of the model files.
```python
from scnetwork import SCNet
scnet = SCNet()
# for the value of "model_path" of loading, use the same value used in "model_path" of saving
scnet.loadModel(model_path='my_model') # loads the model which files names is "my_model"
scnet.print_predict('What is the main language spoken in Spain?', stream=False, interval=0.05)

```
```bash
Loading model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 5069.08it/s]
The main language spoken in Spain is Spanish.
```
You can also use the extension of one of the saved files if you want to make the code more readable.
```python
from scnetwork import SCNet
scnet = SCNet()
# if you want you can use the extension ".scnet" or ".sccon" at the end of the path in "model_path"
scnet.loadModel(model_path='my_model.scnet')
scnet.print_predict('What is the main language spoken in Spain?', stream=False, interval=0.05)

```
```bash
Loading model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 5196.48it/s]
The main language spoken in Spain is Spanish.
```
Any of the extensions are valid.
```python
from scnetwork import SCNet
scnet = SCNet()
# if you want you can use the extension ".scnet" or ".sccon" at the end of the path in "model_path"
scnet.loadModel(model_path='my_model.sccon')
scnet.print_predict('What is the main language spoken in Spain?', stream=False, interval=0.05)

```
```bash
Loading model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 5278.70it/s]
The main language spoken in Spain is Spanish.
```
The same goes for the saving method. The extension of one of the files is optional.
```python
from scnetwork import SCNet
scnet = SCNet()

scnet.train(dataset_path='dataset.json')
# you can use the extension of one of the files in "model_path" if you prefer
scnet.saveModel(model_path='my_model.scnet')

```
```bash
Reading data: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 1/1 [00:00<00:00, 11781.75it/s]
Embedding data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████| 1000000/1000000 [00:00<00:00, 17162409.11it/s]
Tokenizing data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 100.82it/s]
Training model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 117.11it/s]
Generating 91U parameters: 100%|████████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 28181.21it/s]
Saving model: 100%|██████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 5/5 [00:00<00:00, 4771.68it/s]
```
Note that either model extension is valid and the result will be the same.
```python
from scnetwork import SCNet
scnet = SCNet()

scnet.train(dataset_path='dataset.json')
# you can use the extension of one of the files in "model_path" if you prefer
scnet.saveModel(model_path='my_model.sccon')

```
```bash
Reading data: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 1/1 [00:00<00:00, 13315.25it/s]
Embedding data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████| 1000000/1000000 [00:00<00:00, 17524825.04it/s]
Tokenizing data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 102.92it/s]
Training model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 119.52it/s]
Generating 91U parameters: 100%|████████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 25944.15it/s]
Saving model: 100%|██████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 5/5 [00:00<00:00, 4542.24it/s]
```
You can also specify a directory for the saved model files. If the specified directory does not exist, it will be created.
```python
from scnetwork import SCNet
scnet = SCNet()

scnet.train(dataset_path='dataset.json')
# saves files named "my_model" inside the directory named "models"
# if the directory already exists, it only generates the files inside it, if it does not exist, it creates the directory before generating the files
scnet.saveModel(model_path='./models/my_model')

```
```bash
Reading data: 100%|██████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 1/1 [00:00<00:00, 9915.61it/s]
Embedding data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████| 1000000/1000000 [00:00<00:00, 17298024.93it/s]
Tokenizing data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 101.56it/s]
Training model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 118.00it/s]
Generating 91U parameters: 100%|████████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 47572.45it/s]
Saving model: 100%|██████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 5/5 [00:00<00:00, 4893.03it/s]
```
It is also possible to specify extensions with the directory.
```python
from scnetwork import SCNet
scnet = SCNet()

scnet.train(dataset_path='dataset.json')
# as the directory already exists, it will only generate the new model files inside it
scnet.saveModel(model_path='./models/model01.scnet')

```
```bash
Reading data: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 1/1 [00:00<00:00, 13706.88it/s]
Embedding data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████| 1000000/1000000 [00:00<00:00, 17085228.50it/s]
Tokenizing data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 100.24it/s]
Training model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 116.43it/s]
Generating 91U parameters: 100%|████████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 40009.26it/s]
Saving model: 100%|██████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 5/5 [00:00<00:00, 4747.91it/s]
```
Note that any of the model extensions are valid.
```python
from scnetwork import SCNet
scnet = SCNet()

scnet.train(dataset_path='dataset.json')
# as the directory already exists, it will only generate the new model files inside it
scnet.saveModel(model_path='./models/model02.sccon')

```
```bash
Reading data: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 1/1 [00:00<00:00, 13148.29it/s]
Embedding data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████| 1000000/1000000 [00:00<00:00, 17357512.35it/s]
Tokenizing data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 101.98it/s]
Training model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 118.48it/s]
Generating 91U parameters: 100%|████████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 34711.48it/s]
Saving model: 100%|██████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 5/5 [00:00<00:00, 4641.77it/s]
```
Now to load simply specify the models directory followed by the name of one of the available models.
```python
from scnetwork import SCNet
scnet = SCNet()

scnet.loadModel(model_path='./models/my_model')
scnet.print_predict('What is the main language spoken in Spain?', stream=False, interval=0.05)

```
```bash
Loading model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 4397.20it/s]
The main language spoken in Spain is Spanish.
```
It is also possible to specify the directory and name of the model with the extension when loading.
```python
from scnetwork import SCNet
scnet = SCNet()

scnet.loadModel(model_path='./models/my_model.scnet')
scnet.print_predict('What is the main language spoken in Spain?', stream=False, interval=0.05)

```
```bash
Loading model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 4590.39it/s]
The main language spoken in Spain is Spanish.
```
Any model extension is valid.
```python
from scnetwork import SCNet
scnet = SCNet()

scnet.loadModel(model_path='./models/my_model.sccon')
scnet.print_predict('What is the main language spoken in Spain?', stream=False, interval=0.05)

```
```bash
Loading model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 5044.70it/s]
The main language spoken in Spain is Spanish.
```
You can also treat the save routine as a function by getting its return state where we will have True if the save is successful or False otherwise.
```python
from scnetwork import SCNet
scnet = SCNet()

scnet.train(dataset_path='dataset.json') # trains the model
saving_state = scnet.saveModel(model_path='./models/model01') # gets the saving state
if saving_state: print('Model saved SUCCESSFULLY!') # displays a success message if training is successful
else: print('ERROR saving model.') # displays a failure message if there some error in training

```
```bash
Reading data: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 1/1 [00:00<00:00, 10305.42it/s]
Embedding data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████| 1000000/1000000 [00:00<00:00, 17135274.70it/s]
Tokenizing data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 100.71it/s]
Training model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 116.97it/s]
Generating 91U parameters: 100%|████████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 23151.63it/s]
Saving model: 100%|██████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 5/5 [00:00<00:00, 4193.47it/s]
Model saved SUCCESSFULLY!
```
Note that the loading routine can also be treated as a function.
```python
from scnetwork import SCNet
scnet = SCNet()

loading_state = scnet.loadModel(model_path='./models/model01')
if loading_state: scnet.print_predict('What is the main language spoken in Spain?', stream=False, interval=0.05) # infers, if the load is successful
else: print('ERROR loading pre-trained model!!') # displays an error message if loading fails

```
```bash
Loading model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 4719.52it/s]
The main language spoken in Spain is Spanish.
```
Below is an example with all the parameters of the save routine and their default values.
```python
from scnetwork import SCNet
scnet = SCNet()

scnet.train(dataset_path='dataset.json')
# the default value of "model_path" is an empty string
# if a model with this name already exists, the previous model will be overwritten
saving_state = scnet.saveModel(model_path='./models/model02', progress=True)
if saving_state: print('Model saved SUCCESSFULLY!')
else: print('ERROR saving model.')

```
```bash
Reading data: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 1/1 [00:00<00:00, 11650.84it/s]
Embedding data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████| 1000000/1000000 [00:00<00:00, 17176887.83it/s]
Tokenizing data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 100.76it/s]
Training model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 117.04it/s]
Generating 91U parameters: 100%|████████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 24769.51it/s]
Saving model: 100%|██████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 5/5 [00:00<00:00, 4348.23it/s]
Model saved SUCCESSFULLY!
```
Below is an example with all the parameters of the load routine and their default values.
```python
from scnetwork import SCNet
scnet = SCNet()
# the default value of "model_path" is an empty string
loading_state = scnet.loadModel(model_path='./models/model01', progress=True)
if loading_state: scnet.print_predict('What is the main language spoken in Spain?', stream=False, interval=0.05)
else: print('ERROR loading pre-trained model!!')

```
```bash
Loading model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 5102.56it/s]
The main language spoken in Spain is Spanish.
```
To perform model inference, you can use two methods. The "predict" method works as a function that returns the result of the inference. The "print_predict" method will display the result of the inference. Both have the same parameters and produce exactly the same result.
```python
from scnetwork import SCNet
scnet = SCNet()
# training phase
scnet.train(dataset_path='dataset.json')
scnet.saveModel(model_path='./models/my_model')

```
```bash
Reading data: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 1/1 [00:00<00:00, 13148.29it/s]
Embedding data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████| 1000000/1000000 [00:00<00:00, 17280065.92it/s]
Tokenizing data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 101.50it/s]
Training model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 117.90it/s]
Generating 91U parameters: 100%|████████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 24892.01it/s]
Saving model: 100%|██████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 5/5 [00:00<00:00, 4528.51it/s]
```
After training your model, simply use the "predict" or "print_predict" method. The "print_predict" method is more practical because it does not require manipulation of the return value.
```python
from scnetwork import SCNet
scnet = SCNet()
# inference phase
scnet.loadModel(model_path='./models/my_model')
scnet.print_predict('Hello! Who are you?')

```
```bash
Loading model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 4450.53it/s]
Hello! I am Sapiens Chat, an AI model created by Sapiens Technology.
```
The inference "prompt" parameter is used to receive the input that will be transformed into an output.
```python
from scnetwork import SCNet
scnet = SCNet()
# inference phase
scnet.loadModel(model_path='./models/my_model')
# the "prompt" parameter receives a string with a question or instruction
# since "prompt" is the first parameter, if you want you can hide it by passing only the string value directly in the first position of the method
scnet.print_predict(prompt='Hello! Who are you?')

```
```bash
Loading model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 4652.21it/s]
Hello! I am Sapiens Chat, an AI model created by Sapiens Technology.
```
With the "max_tokens" parameter you can set an integer for the maximum limit of tokens to be returned in the response.
```python
from scnetwork import SCNet
scnet = SCNet()
# inference phase
scnet.loadModel(model_path='./models/my_model')
# the "max_tokens" parameter receives an integer with the approximate maximum number of tokens in the method's response
# the default value of "max_tokens" is None, with None it is implied that there will be no limit to the number of tokens returned
scnet.print_predict(prompt='Hello! Who are you?', max_tokens=2) # max_tokens=2 will return only the part of the response that contains up to 2 tokens

```
```bash
Loading model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 4726.36it/s]
Hello!
```
You can access the probability of the inference with the level of certainty of the answer in percentage numbers between 0 and 1, where 0 will be a completely uncertain answer and 1 will be an answer with 100% certainty of being the correct answer. To do this, simply access the public variable named "probability" through the class object, right after the inference for which you want to know the level of certainty.
```python
from scnetwork import SCNet
scnet = SCNet()
# inference phase
scnet.loadModel(model_path='./models/my_model')
# the "probability" variable can be accessed to obtain the probability of the answer
# the "probability" variable will only have the correct probability after the inference is completed
scnet.print_predict(prompt='Hello! Who are you?') # executes the inference
print('Probability of answer:', scnet.probability) # displays the probability with the certainty level of the immediately preceding prediction

```
```bash
Loading model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 4820.25it/s]
Hello! I am Sapiens Chat, an AI model created by Sapiens Technology.
Probability of answer: 0.9583294388883696
```
With the "min_probability" parameter of the prediction method, it is possible to define a minimum probability for the response to be returned. If this minimum probability is not reached, the response will be an empty string.
```python
from scnetwork import SCNet
scnet = SCNet()
# inference phase
scnet.loadModel(model_path='./models/my_model')
# "min_probability" receives a floating point number between 0 and 1 with the desired minimum probability; its default value is 0.01 for 1% probability
# since the probability for this model's answer to the question below is less than 0.98 (98%), the return will be an empty string
min_probability = 0.98 # minimum probability for a response to be returned
scnet.print_predict(prompt='Hello! Who are you?', min_probability=min_probability)
print('Minimum desired probability:', min_probability)
print('Probability achieved:', scnet.probability)

```
```bash
Loading model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 4912.18it/s]

Minimum desired probability: 0.98
Probability achieved: 0.9583294388883696
```
With the "generalization" parameter set to True, you enable the model to generalize the response to inputs it was not trained on, adapting the response to new information. If "generalization" is set to False, you disable the model's generalization, causing it to not generalize to inputs it was not trained on and returning only the training response closest to the specified input.
```python
from scnetwork import SCNet
scnet = SCNet()
# inference phase
scnet.loadModel(model_path='./models/my_model')
# for the prompt input below the model was trained only with the response "The main language spoken in Spain is Spanish."
# with the "generalization" parameter equal to False, the response will be the same as the training response
# with the "generalization" parameter equal to True, the response will be a generalization/adaptation of the training response for the current prompt
# the default value of the "generalization" parameter is True
# in this example we have two words that the model does not know: 'principal' and 'Mexico'; both will be inserted into the generalized response based on the question pattern and the training response pattern
prompt = 'What is the principal language spoken in Mexico?'
scnet.print_predict(prompt=prompt, generalization=False) # non-generalized response
scnet.print_predict(prompt=prompt, generalization=True) # generalized/adapted response

```
```bash
Loading model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 4967.87it/s]
The main language spoken in Spain is Spanish.
The principal language spoken in Mexico is Spanish.
```
With the "stream" parameter of the prediction method, it is possible to displays token by token with True, or the complete response at once with False.
```python
from scnetwork import SCNet
scnet = SCNet()
# inference phase
scnet.loadModel(model_path='./models/my_model')
# if the "stream" parameter is equal to True, the inference will display token by token, if it is equal to False, it will display the complete response; the default value is False
scnet.print_predict(prompt='Hello! Who are you?', stream=True)

```
```bash
Loading model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 5405.03it/s]
Hello! I am Sapiens Chat, an AI model created by Sapiens Technology.
```
It is also possible to combine enabling the "stream" parameter when it is equal to True with the "interval" parameter, which will receive a floating number referring to the minimum interval in seconds to be waited between one token and another.
```python
from scnetwork import SCNet
scnet = SCNet()
# inference phase
scnet.loadModel(model_path='./models/my_model')
# "interval" receives a float with the time in seconds to wait for the next token to be generated in the response
# always use a positive floating or integer number in "interval"
scnet.print_predict(prompt='Hello! Who are you?', stream=True, interval=0.05) # interval=0.05 wait 5% of 1 second between one token and another

```
```bash
Loading model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 5028.28it/s]
Hello! I am Sapiens Chat, an AI model created by Sapiens Technology.
```
If you call the "addFit" method after model training or loading, a fine-tuning is applied on top of the current model. This can be especially useful for fine-tuning responses to inputs that are being incorrectly inferred by the base model, or for adding new data to the current model without having to retrain it from scratch.
```python
from scnetwork import SCNet
scnet = SCNet()
# adjustment phase
scnet.loadModel(model_path='./models/my_model') # loads the base model
# applies fine-tuning to the loaded model
scnet.addFit('What does the acronym GPT mean in language models?', 'GPT in language models stands for Generative Pre-trained Transformer.')
scnet.addFit('Tell me what is the meaning of GPT in language models.', 'GPT in language models stands for Generative Pre-trained Transformer.')
scnet.addFit('What is the capital of Spain?', 'The capital of Spain is Madrid.')
scnet.addFit('Tell me the name of the capital of Spain.', 'The capital of Spain is Madrid.')
# after the fine-tuning is complete, a file will be generated with the name of the current model and the extension ".scfit"

```
```bash
Loading model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 4919.59it/s]
```
Now just load the adjusted model with the knowledge of the base model plus the knowledge of the fine-tuning.
```python
from scnetwork import SCNet
scnet = SCNet()
# inference phase
scnet.loadModel(model_path='./models/my_model') # loads the model that received the fine-tuning
scnet.print_predict(prompt='Hello! Who are you?', stream=True, interval=0.05) # base model question
scnet.print_predict(prompt='What does the acronym GPT mean in language models?', stream=True, interval=0.05) # fine-tuning question

```
```bash
Loading model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 4085.74it/s]
Hello! I am Sapiens Chat, an AI model created by Sapiens Technology.
GPT in language models stands for Generative Pre-trained Transformer.
```
When your model is fine-tuned, you can access the probability that the current input belongs to the body of knowledge of the data added to the fine-tuning using the global variable named "fit_probability".
```python
from scnetwork import SCNet
scnet = SCNet()
# inference phase
scnet.loadModel(model_path='./models/my_model') # loads the model that received the fine-tuning
scnet.print_predict(prompt='Hello! Who are you?', stream=True, interval=0.05) # base model question
print('Fine-tuning probability:', scnet.fit_probability) # displays the probability of previous prompt belongs knowledge of fine-tuning
scnet.print_predict(prompt='What does the acronym GPT mean in language models?', stream=True, interval=0.05) # fine-tuning question
print('Fine-tuning probability:', scnet.fit_probability) # displays the probability of previous prompt belongs knowledge of fine-tuning
# the "fit_probability" variable returns a percentage level between 0 and 1 for the chance of current input contains your response in the fine-tuning data, it should be accessed after a prediction

```
```bash
Loading model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 4081.76it/s]
Hello! I am Sapiens Chat, an AI model created by Sapiens Technology.
Fine-tuning probability: 0.0
GPT in language models stands for Generative Pre-trained Transformer.
Fine-tuning probability: 1.0
```
Prediction methods also have a parameter named "min_fit_probability" that will receive a float between 0 and 1 with the minimum expected value for the global variable "fit_probability". If "fit_probability" is less than "min_fit_probability", the model will disregard the fine-tuning data and will look for its answer only in the base model data.
```python
from scnetwork import SCNet
scnet = SCNet()
# inference phase
scnet.loadModel(model_path='./models/my_model') # loads the model that received the fine-tuning
# the "min_fit_probability" parameter receives a floating number between 0 and 1 with the minimum desired probability for the fine-tuning response
# the default value of the "min_fit_probability" parameter is 0.7 for 70%
min_fit_probability = 0.8 # example with a minimum probability of 80% expected for the model to consider using the fine-tuning data
# in the first inference the fine-tuning probability reaches the value defined in "min_fit_probability", therefore the most likely response of the fine-tuning will be returned
# in the second inference the fine-tuning probability does not reach the value defined in "min_fit_probability", therefore the most likely response of the base model will be returned
scnet.print_predict(prompt='What is the capital of Spain?', min_fit_probability=min_fit_probability, stream=True, interval=0.05) # returns fine-tuning response
print(f'Expected probability for returning a response of fine-tuning: {min_fit_probability} - Probability achieved in the fine-tuning: {scnet.fit_probability}.')
print()
# the inference below will return the answer "The capital of Spain is Lisbon." adapted from the answer "The capital of Portugal is Lisbon." from the base model
scnet.print_predict(prompt='I would like to know the name of the capital of Spain.', min_fit_probability=min_fit_probability, stream=True, interval=0.05) # returns base model response
print(f'Expected probability for returning a response of fine-tuning: {min_fit_probability} - Probability achieved in the fine-tuning: {scnet.fit_probability}.')

```
```bash
Loading model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 4302.48it/s]
The capital of Spain is Madrid.
Expected probability for returning a response of fine-tuning: 0.8 - Probability achieved in the fine-tuning: 1.0.

The capital of Spain is Lisbon.
Expected probability for returning a response of fine-tuning: 0.8 - Probability achieved in the fine-tuning: 0.7777777777777778.
```
Check out an example below with all the parameters of the "print_predict" method in their respective orders and with their default values.
```python
from scnetwork import SCNet
scnet = SCNet()

scnet.loadModel(model_path='./models/my_model')

scnet.print_predict( # performs model inference and displays the result
	prompt='Hello! Who are you?', # input string for which a response is desired; default value is an empty string
	max_tokens=None, # integer with the maximum number of tokens desired in the response; None for responses with no token limit
	min_fit_probability=0.7, # minimum probability of the closest fine-tuning output found, so that it is returned as the answer
	min_probability=0.01, # minimum probability for an output from the fine-tuning or base model to be returned, if not reached the return will be an empty string
	generalization=True, # True to make the model a generalist with the ability to adapt the response to the user's input, or False to return identical responses to those from the training
	stream=False, # True to display the result of the inference token by token, or False to display all the generated tokens at once
	interval=0.0 # minimum interval in seconds to generate the next token
)

```
```bash
Loading model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 3798.21it/s]
Hello! I am Sapiens Chat, an AI model created by Sapiens Technology.
```
The "predict" function works exactly the same way and with the same parameters as the "print_predict" function, with the difference that instead of displaying the answer, it will return the answer, giving the developer more flexibility to manipulate the output as they wish.
```python
from scnetwork import SCNet
scnet = SCNet()

scnet.loadModel(model_path='./models/my_model')

inference = scnet.predict( # performs model inference and displays the result
	prompt='Hello! Who are you?', # input string for which a response is desired; default value is an empty string
	max_tokens=None, # integer with the maximum number of tokens desired in the response; None for responses with no token limit
	min_fit_probability=0.7, # minimum probability of the closest fine-tuning output found, so that it is returned as the answer
	min_probability=0.01, # minimum probability for an output from the fine-tuning or base model to be returned, if not reached the return will be an empty string
	generalization=True, # True to make the model a generalist with the ability to adapt the response to the user's input, or False to return identical responses to those from the training
	stream=False, # True to display the result of the inference token by token, or False to display all the generated tokens at once
	interval=0.0 # minimum interval in seconds to generate the next token
)
# the result of the return of the "predict" function will be a dictionary with the keys 'answer' containing the answer, 'fit_probability' containing the probability of using the fine-tuning and 'probability' which contains the final probability
print(inference) # displays the inference result returned by the prediction function

```
```bash
Loading model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 3948.38it/s]
{'answer': 'Hello! I am Sapiens Chat, an AI model created by Sapiens Technology.', 'fit_probability': 0.0, 'probability': 0.9583294388883696}
```
See below how to access each of the return values.
```python
from scnetwork import SCNet
scnet = SCNet()

scnet.loadModel(model_path='./models/my_model')

inference = scnet.predict(prompt='Hello! Who are you?') # returns the inference dictionary

answer = inference['answer'] # response text
fit_probability = inference['fit_probability'] # probability of being a fine-tuning response
probability = inference['probability'] # probability of being the correct answer

print('answer:', answer) # displays the answer
print('fit_probability:', fit_probability) # displays the fit probability
print('probability:', probability) # displays the final probability

```
```bash
Loading model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 3937.26it/s]
answer: Hello! I am Sapiens Chat, an AI model created by Sapiens Technology.
fit_probability: 0.0
probability: 0.9583294388883696
```
Now check out an example returning one token at a time with the "stream" parameter equal to True.
```python
from scnetwork import SCNet
scnet = SCNet()

scnet.loadModel(model_path='./models/my_model')

generator = scnet.predict(prompt='Hello! Who are you?', stream=True, interval=0.05) # with "stream" enabled, the return will be a generator object

for inference in generator: # runs through the generator
	token = inference['answer']
	print(token, end='', flush=True) # displays one token at a time on the same line

	fit_probability = inference['fit_probability']
	probability = inference['probability']
print()

print('fit_probability:', fit_probability)
print('probability:', probability)

```
```bash
Loading model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 4067.06it/s]
Hello! I am Sapiens Chat, an AI model created by Sapiens Technology.
fit_probability: 0.0
probability: 0.9583294388883696
```
Below is an example with an effective test on all input prompts contained in the JSON file that was used in the training.
```python
from scnetwork import SCNet
scnet = SCNet()

scnet.loadModel(model_path='./models/my_model')
# the "loadJSON" function reads a json file and returns it in dictionary format
json_data = scnet.loadJSON('dataset.json')
for data in json_data['data']:
	prompt = data['input']
	generator = scnet.predict(prompt=prompt, min_fit_probability=0.95, stream=True, interval=0.05)

	print('Prompt:', prompt)
	print('Answer: ', end='', flush=True)
	for inference in generator:
		token = inference['answer']
		print(token, end='', flush=True)
	print('\n')

```
```bash
Loading model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 4187.13it/s]
Prompt: Hello! Who are you?
Answer: Hello! I am Sapiens Chat, an AI model created by Sapiens Technology.

Prompt: Who discovered Brazil?
Answer: Brazil was discovered by Portuguese navigators led by Pedro Alvares Cabral in 1500.

Prompt: What is the main language spoken in Spain?
Answer: The main language spoken in Spain is Spanish.

Prompt: What is the capital of Portugal?
Answer: The capital of Portugal is Lisbon.

Prompt: How much is 2 + 2?
Answer: The sum of 2 + 2 in mathematics is 4.

Prompt: What is five minus three?
Answer: The result of five minus three is two.

Prompt: What is your name?
Answer: My name is Sapiens Chat.

```
When you train a model and certain inputs are giving incorrect or inaccurate answers, you should fine-tune it on top of the original training. Fine-tuning can also be especially useful when you need to add new data to an already trained model where retraining would be costly and inefficient. To fine-tune a pre-trained model, simply call the "addFit" method right after training or loading a model to add new fine-tuned data to the existing model.
```python
from scnetwork import SCNet
scnet = SCNet()
# first train the model
scnet.train(dataset_path='dataset.json')

prompt = 'Hello! Who are you?'
print('Prompt:', prompt)
print('Answer: ', end='', flush=True)
scnet.print_predict(prompt=prompt, stream=True, interval=0.05)
print()
# after training the model, any call to the "addFit" method will be considered a fine-tuning
INPUT = 'What is the capital of Brazil?'
OUTPUT = 'The capital of Brazil is Brasilia.'
scnet.addFit(prompt=INPUT, answer=OUTPUT)
# fine-tuning test
# the result will be the response added to the fit, even if that response is not in the original training data
print('Prompt:', INPUT)
print('Answer: ', end='', flush=True)
scnet.print_predict(prompt=INPUT, stream=True, interval=0.05)

```
```bash
Reading data: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 1/1 [00:00<00:00, 13662.23it/s]
Embedding data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████| 1000000/1000000 [00:00<00:00, 17397326.30it/s]
Tokenizing data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 102.21it/s]
Training model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 118.71it/s]
Prompt: Hello! Who are you?
Answer: Hello! I am Sapiens Chat, an AI model created by Sapiens Technology.

Prompt: What is the capital of Brazil?
Answer: The capital of Brazil is Brasilia.
```
To add truly effective fine-tuning that can be loaded without retraining, you should train a base model, then save it, and fine-tune only after you have loaded the pre-trained model.
```python
from scnetwork import SCNet
scnet = SCNet()
# creates a base model to be adjusted
scnet.train(dataset_path='dataset.json')
scnet.saveModel('./models_for_adjustment/model01')

```
```bash
Reading data: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 1/1 [00:00<00:00, 12228.29it/s]
Embedding data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████| 1000000/1000000 [00:00<00:00, 17071876.59it/s]
Tokenizing data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 100.26it/s]
Training model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 116.46it/s]
Generating 91U parameters: 100%|████████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 39756.44it/s]
Saving model: 100%|██████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 5/5 [00:00<00:00, 4500.33it/s]
```
In the following example, we demonstrate a case where the model will need to be adjusted for an incorrect answer.
```python
from scnetwork import SCNet
scnet = SCNet()

scnet.loadModel('./models_for_adjustment/model01')
# below we are waiting for the following response contained in the base model: "Brazil was discovered by Portuguese navigators led by Pedro Alvares Cabral in 1500."
# but the answer obtained will not be the answer we are expecting, this shows that the model needs to be adjusted for the same answer with the prompt below
prompt = 'Please, could you tell me who were the discoverers of Brazil?'
print('Prompt:', prompt)
print('Answer: ', end='', flush=True)
scnet.print_predict(prompt=prompt, stream=True, interval=0.05)

```
```bash
Loading model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 4441.77it/s]
Prompt: Please, could you tell me who were the discoverers of Brazil?
Answer: The discoverers of Brazil is Lisbon.
```
Now that we have identified an incorrect answer, we will load the model that is answering incorrectly and adjust it with the answer we want for the previous test input.
```python
from scnetwork import SCNet
scnet = SCNet()

scnet.loadModel('./models_for_adjustment/model01')
# input and output pair definition for fine-tuning
INPUT = 'Please, could you tell me who were the discoverers of Brazil?'
OUTPUT = 'Brazil was discovered by Portuguese navigators led by Pedro Alvares Cabral in 1500.'
scnet.addFit(INPUT, OUTPUT) # the first parameter receives the input and the second the desired output
# you can make as many adjustments as you want
# an additional file named "model01.scfit" will be generated in the current model directory
# you will not need to save the model after adjustment

```
```bash
Loading model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 4710.43it/s]
```
Now after fine-tuning we will test the base model that previously emitted the incorrect response.
```python
from scnetwork import SCNet
scnet = SCNet()

scnet.loadModel('./models_for_adjustment/model01')
# this time, as the model has been adjusted, we will have our desired answer
prompt = 'Please, could you tell me who were the discoverers of Brazil?'
print('Prompt:', prompt)
print('Answer: ', end='', flush=True)
scnet.print_predict(prompt=prompt, stream=True, interval=0.05)

```
```bash
Loading model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 4284.27it/s]
Prompt: Please, could you tell me who were the discoverers of Brazil?
Answer: Brazil was discovered by Portuguese navigators led by Pedro Alvares Cabral in 1500.
```
You can also add new adjustments to an already adjusted model.
```python
from scnetwork import SCNet
scnet = SCNet()

scnet.loadModel('./models_for_adjustment/model01')
# adds new fits without losing previous fits
scnet.addFit('What does the acronym GPT mean in language models?', 'GPT in language models stands for Generative Pre-trained Transformer.')
scnet.addFit('Tell me what is the meaning of GPT in language models.', 'GPT in language models stands for Generative Pre-trained Transformer.')
scnet.addFit('What is the capital of Spain?', 'The capital of Spain is Madrid.')
scnet.addFit('Tell me the name of the capital of Spain.', 'The capital of Spain is Madrid.')

```
```bash
Loading model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 4194.90it/s]
```
Now realize that the base model contains the data from training, first fit, and later fit.
```python
from scnetwork import SCNet
scnet = SCNet()

scnet.loadModel('./models_for_adjustment/model01')

prompt = 'Hello! Who are you?'
print('Prompt:', prompt)
print('Answer: ', end='', flush=True)
scnet.print_predict(prompt=prompt, stream=True, interval=0.05)
print()
prompt = 'What is your name?'
print('Prompt:', prompt)
print('Answer: ', end='', flush=True)
scnet.print_predict(prompt=prompt, stream=True, interval=0.05)
print()
prompt = 'Please, could you tell me who were the discoverers of Brazil?'
print('Prompt:', prompt)
print('Answer: ', end='', flush=True)
scnet.print_predict(prompt=prompt, stream=True, interval=0.05)
print()
prompt = 'What does the acronym GPT mean in language models?'
print('Prompt:', prompt)
print('Answer: ', end='', flush=True)
scnet.print_predict(prompt=prompt, stream=True, interval=0.05)
print()
prompt = 'What is the capital of Spain?'
print('Prompt:', prompt)
print('Answer: ', end='', flush=True)
scnet.print_predict(prompt=prompt, stream=True, interval=0.05)

```
```bash
Loading model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 3914.68it/s]
Prompt: Hello! Who are you?
Answer: Hello! I am Sapiens Chat, an AI model created by Sapiens Technology.

Prompt: What is your name?
Answer: My name is Sapiens Chat.

Prompt: Please, could you tell me who were the discoverers of Brazil?
Answer: Brazil was discovered by Portuguese navigators led by Pedro Alvares Cabral in 1500.

Prompt: What does the acronym GPT mean in language models?
Answer: GPT in language models stands for Generative Pre-trained Transformer.

Prompt: What is the capital of Spain?
Answer: The capital of Spain is Madrid.
```
To make the code easier to read, you can specify the names of the input and output parameters when assigning values.
```python
from scnetwork import SCNet
scnet = SCNet()
# saving a new base model
scnet.train(dataset_path='dataset.json')
scnet.saveModel('./models_for_adjustment/model02')

```
```bash
Reading data: 100%|██████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 1/1 [00:00<00:00, 9709.04it/s]
Embedding data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████| 1000000/1000000 [00:00<00:00, 17336204.58it/s]
Tokenizing data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 101.74it/s]
Training model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 118.15it/s]
Generating 91U parameters: 100%|████████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 42581.77it/s]
Saving model: 100%|██████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 5/5 [00:00<00:00, 4780.38it/s]
```
With the base model saved, we can now apply our fine-tuning.
```python
from scnetwork import SCNet
scnet = SCNet()
# base model loading
scnet.loadModel('./models_for_adjustment/model02')
# application of fine-tuning
# the "prompt" parameter receives an example input and the "answer" parameter receives an answer that should be returned when an input equal or similar to the "prompt" parameter is sent to the inference
scnet.addFit(prompt='What does the acronym GPT mean in language models?', answer='GPT in language models stands for Generative Pre-trained Transformer.')
scnet.addFit(prompt='Tell me what is the meaning of GPT in language models.', answer='GPT in language models stands for Generative Pre-trained Transformer.')
scnet.addFit(prompt='What is the capital of Spain?', answer='The capital of Spain is Madrid.')
scnet.addFit(prompt='Tell me the name of the capital of Spain.', answer='The capital of Spain is Madrid.')

```
```bash
Loading model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 4461.35it/s]
```
Note that after fine-tuning the model will respond to both inputs from the base model and inputs that were added via fine-tuning.
```python
from scnetwork import SCNet
scnet = SCNet()
# loading of the adjusted model
scnet.loadModel('./models_for_adjustment/model01')
min_fit_probability = 0.85 # minimum probability for fine-tuning to be considered
# inference tests with the base model data and with the fine-tuning data
prompt = 'Hello! Who are you?'
print('Prompt:', prompt)
print('Answer: ', end='', flush=True)
scnet.print_predict(prompt=prompt, min_fit_probability=min_fit_probability, stream=True, interval=0.05)
print()
prompt = 'What is your name?'
print('Prompt:', prompt)
print('Answer: ', end='', flush=True)
scnet.print_predict(prompt=prompt, min_fit_probability=min_fit_probability, stream=True, interval=0.05)
print()
prompt = 'Who discovered Brazil?'
print('Prompt:', prompt)
print('Answer: ', end='', flush=True)
scnet.print_predict(prompt=prompt, min_fit_probability=min_fit_probability, stream=True, interval=0.05)
print()
prompt = 'What is the main language spoken in Spain?'
print('Prompt:', prompt)
print('Answer: ', end='', flush=True)
scnet.print_predict(prompt=prompt, min_fit_probability=min_fit_probability, stream=True, interval=0.05)
print()
prompt = 'What is the capital of Portugal?'
print('Prompt:', prompt)
print('Answer: ', end='', flush=True)
scnet.print_predict(prompt=prompt, min_fit_probability=min_fit_probability, stream=True, interval=0.05)
print()
prompt = 'How much is 2 + 2?'
print('Prompt:', prompt)
print('Answer: ', end='', flush=True)
scnet.print_predict(prompt=prompt, min_fit_probability=min_fit_probability, stream=True, interval=0.05)
print()
prompt = 'What is five minus three?'
print('Prompt:', prompt)
print('Answer: ', end='', flush=True)
scnet.print_predict(prompt=prompt, min_fit_probability=min_fit_probability, stream=True, interval=0.05)
print()
prompt = 'What does the acronym GPT mean in language models?'
print('Prompt:', prompt)
print('Answer: ', end='', flush=True)
scnet.print_predict(prompt=prompt, min_fit_probability=min_fit_probability, stream=True, interval=0.05)
print()
prompt = 'Tell me what is the meaning of GPT in language models.'
print('Prompt:', prompt)
print('Answer: ', end='', flush=True)
scnet.print_predict(prompt=prompt, min_fit_probability=min_fit_probability, stream=True, interval=0.05)
print()
prompt = 'What is the capital of Spain?'
print('Prompt:', prompt)
print('Answer: ', end='', flush=True)
scnet.print_predict(prompt=prompt, min_fit_probability=min_fit_probability, stream=True, interval=0.05)
print()
prompt = 'Tell me the name of the capital of Spain.'
print('Prompt:', prompt)
print('Answer: ', end='', flush=True)
scnet.print_predict(prompt=prompt, min_fit_probability=min_fit_probability, stream=True, interval=0.05)
print()

```
```bash
Loading model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 3751.61it/s]
Prompt: Hello! Who are you?
Answer: Hello! I am Sapiens Chat, an AI model created by Sapiens Technology.

Prompt: What is your name?
Answer: My name is Sapiens Chat.

Prompt: Who discovered Brazil?
Answer: Brazil was discovered by Portuguese navigators led by Pedro Alvares Cabral in 1500.

Prompt: What is the main language spoken in Spain?
Answer: The main language spoken in Spain is Spanish.

Prompt: What is the capital of Portugal?
Answer: The capital of Portugal is Lisbon.

Prompt: How much is 2 + 2?
Answer: The sum of 2 + 2 in mathematics is 4.

Prompt: What is five minus three?
Answer: The result of five minus three is two.

Prompt: What does the acronym GPT mean in language models?
Answer: GPT in language models stands for Generative Pre-trained Transformer.

Prompt: Tell me what is the meaning of GPT in language models.
Answer: GPT in language models stands for Generative Pre-trained Transformer.

Prompt: What is the capital of Spain?
Answer: The capital of Spain is Madrid.

Prompt: Tell me the name of the capital of Spain.
Answer: The capital of Spain is Madrid.

```
To fit your model to chat, you must specify a different "id" number for each fit added to the chat. Then simply specify this same "id" to the "relationship_id" parameter of the fit with the answer that should be considered when the previous inference to response with the answer of the fit with the same "id" as the current "relationship_id" and the inputs of the fit and the current inference are the same or similar.
```python
from scnetwork import SCNet
scnet = SCNet()
# saving a new base model
scnet.train(dataset_path='dataset.json')
scnet.saveModel('./models_for_adjustment/chat-model')

```
```bash
Reading data: 100%|██████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 1/1 [00:00<00:00, 2486.25it/s]
Embedding data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████| 1000000/1000000 [00:00<00:00, 14972919.33it/s]
Tokenizing data: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 85.48it/s]
Training model: 100%|██████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 99.08it/s]
Generating 91U parameters: 100%|████████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 38716.65it/s]
Saving model: 100%|██████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 5/5 [00:00<00:00, 4500.33it/s]
```
With the base model trained, we will prepare it so that it can be used as a chat, relating the "relationship_id" parameters with the "id" name parameters in the "addFit" method.
```python
from scnetwork import SCNet
scnet = SCNet()
# base model loading
scnet.loadModel('./models_for_adjustment/chat-model')
# applying fine-tuning for chat mode
# note: we call instruction models the models that were not trained for chat and chat models the models that were trained to maintain a dialogue (which is the case of the current model)
# the "id" parameter receives a unique number for each fit, or 0 for one or more fits that should not be considered as part of a chat
# the "relationship_id" parameter receives the number of any "id" that has already been added previously
# the "relationship_id" parameters with 0 will not take into account the order of the entries and therefore will not be part of the chat mode
# the fits with "relationship_id" different from zero will only be returned as a response when the user's immediately preceding entry has the same "id" as the current "relationship_id"
scnet.addFit(prompt='Hello!', answer='Hello everything is fine?', id=1, relationship_id=0)
scnet.addFit(prompt='Yes.', answer="Very good, I'm glad to know.", id=2, relationship_id=1)
scnet.addFit(prompt='No.', answer="What a shame, I'm so sorry for you.", id=3, relationship_id=1)
scnet.addFit(prompt='Are you a robot?', answer='Yes, I am, and you?', id=4, relationship_id=0)
scnet.addFit(prompt='No.', answer="It's good to know I'm talking to a human.", id=5, relationship_id=4)
scnet.addFit(prompt='Yes.', answer='How interesting, we will be two robots talking.', id=6, relationship_id=4)
scnet.addFit(prompt="Hey how's it going?", answer='Hi, I am fine. Can I help you with anything?', id=7, relationship_id=0)
scnet.addFit(prompt='Yes you can. I would like to know what the capital of Denmark is.', answer='The capital of Denmark is Copenhagen.', id=8, relationship_id=7)
scnet.addFit(prompt="No, you can't.", answer="Okay, if you need anything I'll be here.", id=9, relationship_id=7)

```
```bash
Loading model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 4009.30it/s]
```
Now we will test our chat model.
```python
from scnetwork import SCNet
scnet = SCNet()
# loading chat model
scnet.loadModel('./models_for_adjustment/chat-model')

prompt = 'Hello!'
print('You:', prompt)
print('Bot: ', end='', flush=True)
scnet.print_predict(prompt=prompt, stream=True, interval=0.05)
print()
prompt = 'Yes.'
print('You:', prompt)
print('Bot: ', end='', flush=True)
scnet.print_predict(prompt=prompt, stream=True, interval=0.05)
print()

prompt = 'Hello!'
print('You:', prompt)
print('Bot: ', end='', flush=True)
scnet.print_predict(prompt=prompt, stream=True, interval=0.05)
print()
prompt = 'No.'
print('You:', prompt)
print('Bot: ', end='', flush=True)
scnet.print_predict(prompt=prompt, stream=True, interval=0.05)
print()

prompt = 'Are you a robot?'
print('You:', prompt)
print('Bot: ', end='', flush=True)
scnet.print_predict(prompt=prompt, stream=True, interval=0.05)
print()
prompt = 'Yes.'
print('You:', prompt)
print('Bot: ', end='', flush=True)
scnet.print_predict(prompt=prompt, stream=True, interval=0.05)
print()

prompt = 'Are you a robot?'
print('You:', prompt)
print('Bot: ', end='', flush=True)
scnet.print_predict(prompt=prompt, stream=True, interval=0.05)
print()
prompt = 'No.'
print('You:', prompt)
print('Bot: ', end='', flush=True)
scnet.print_predict(prompt=prompt, stream=True, interval=0.05)
print()

prompt = "Hey how's it going?"
print('You:', prompt)
print('Bot: ', end='', flush=True)
scnet.print_predict(prompt=prompt, stream=True, interval=0.05)
print()
prompt = 'Yes you can. I would like to know what the capital of Denmark is.'
print('You:', prompt)
print('Bot: ', end='', flush=True)
scnet.print_predict(prompt=prompt, stream=True, interval=0.05)
print()

prompt = "Hey how's it going?"
print('You:', prompt)
print('Bot: ', end='', flush=True)
scnet.print_predict(prompt=prompt, stream=True, interval=0.05)
print()
prompt = "No, you can't."
print('You:', prompt)
print('Bot: ', end='', flush=True)
scnet.print_predict(prompt=prompt, stream=True, interval=0.05)
print()

```
```bash
Loading model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 3884.13it/s]
You: Hello!
Bot: Hello everything is fine?

You: Yes.
Bot: Very good, I'm glad to know.

You: Hello!
Bot: Hello everything is fine?

You: No.
Bot: What a shame, I'm so sorry for you.

You: Are you a robot?
Bot: Yes, I am, and you?

You: Yes.
Bot: How interesting, we will be two robots talking.

You: Are you a robot?
Bot: Yes, I am, and you?

You: No.
Bot: It's good to know I'm talking to a human.

You: Hey how's it going?
Bot: Hi, I am fine. Can I help you with anything?

You: Yes you can. I would like to know what the capital of Denmark is.
Bot: The capital of Denmark is Copenhagen.

You: Hey how's it going?
Bot: Hi, I am fine. Can I help you with anything?

You: No, you can't.
Bot: Okay, if you need anything I'll be here.

```
You can also add hidden layers to your network. By adding hidden layers, the fit will be performed with a [HurNet](https://github.com/sapiens-technology/HurNet) type neural network. Hidden layers only apply to fits for instruction models, i.e. where the "id" and "relationship_id" parameters are not defined.
```python
from scnetwork import SCNet
scnet = SCNet()
# saving a new base model
scnet.train(dataset_path='dataset.json')
scnet.saveModel('./models_for_adjustment/hurnet-model')

```
```bash
Reading data: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 1/1 [00:00<00:00, 13148.29it/s]
Embedding data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████| 1000000/1000000 [00:00<00:00, 17084115.04it/s]
Tokenizing data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 100.39it/s]
Training model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 116.59it/s]
Generating 91U parameters: 100%|████████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 41120.63it/s]
Saving model: 100%|██████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 5/5 [00:00<00:00, 4961.32it/s]
```
After training we will apply fine-tuning with hidden layers to improve pattern abstraction.
```python
from scnetwork import SCNet
scnet = SCNet()
# base model loading
scnet.loadModel('./models_for_adjustment/hurnet-model')
# hidden layers architecture of hurnet network
# the name parameter "num_neurons" of the method of layers must receive an integer greater than zero
# the "addHiddenLayer" method can be treated as a function because we can obtain its return which will be True if the layer is added, or False otherwise
# you can add as many hidden layers as you want and with the number of neurons you deem necessary
scnet.addHiddenLayer(num_neurons=2) # adds a hidden layer with 2 neurons
scnet.addHiddenLayer(num_neurons=4) # adds another hidden layer with 4 neurons
scnet.addHiddenLayer(num_neurons=2) # adds another hidden layer with 2 neurons

scnet.addFit(
	prompt='If I have five fingers on one hand and two on the other, do I have an even number or an odd number?',
	answer="You have 7 fingers, so it's an odd number."
)
scnet.addFit(
	prompt='The square root of 25 is what?',
	answer='The square root of 25 is the same as 5 multiplied by 5, so the square root of 25 is 5.'
)

# when adding hidden layers to the network, you will need to invoke the "fit" method, so that the layers are compiled
scnet.fit() # trains the hurnet neural network and saves a file with the ".hutnet" extension in the directory of the loaded model

```
```bash
Loading model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 4850.51it/s]
Fitting model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 4/4 [00:00<00:00, 2073.57it/s]
```
Now we will test the result of the fit by running some inferences.
```python
from scnetwork import SCNet
scnet = SCNet()
# loading hurnet model
scnet.loadModel('./models_for_adjustment/hurnet-model')

prompt = 'If I have five fingers on one hand and two on the other, do I have an even number or an odd number?'
print('Prompt:', prompt)
print('Answer: ', end='', flush=True)
scnet.print_predict(prompt=prompt, stream=True, interval=0.05)
print()
prompt = 'The square root of 25 is what?'
print('Prompt:', prompt)
print('Answer: ', end='', flush=True)
scnet.print_predict(prompt=prompt, stream=True, interval=0.05)
print()

```
```bash
Loading model: 100%|██████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 927.80it/s]
Prompt: If I have five fingers on one hand and two on the other, do I have an even number or an odd number?
Answer: You have 7 fingers, so it's an odd number.

Prompt: The square root of 25 is what?
Answer: The square root of 25 is the same as 5 multiplied by 5, so the square root of 25 is 5.

```
When performing a new adjustment with the [HurNet](https://github.com/sapiens-technology/HurNet) network, any previous adjustment with the [HurNet](https://github.com/sapiens-technology/HurNet) architecture will be overwritten by the current adjustment.
```python
from scnetwork import SCNet
scnet = SCNet()
# base model loading
scnet.loadModel('./models_for_adjustment/hurnet-model')
# if you prefer, you can hide the name of the "num_neurons" parameter, just passing the integer with the number of neurons directly
# in the following example we are creating four hidden layers with two neurons each
scnet.addHiddenLayer(2)
scnet.addHiddenLayer(2)
scnet.addHiddenLayer(2)
scnet.addHiddenLayer(2)
# to verify the replacement of the previous fine-tuning with the current fine-tuning, we will add a greeting with 'Hi!' in each of the responses
scnet.addFit(
	prompt='If I have five fingers on one hand and two on the other, do I have an even number or an odd number?',
	answer="Hi! You have 7 fingers, so it's an odd number."
)
scnet.addFit(
	prompt='The square root of 25 is what?',
	answer='Hi! The square root of 25 is the same as 5 multiplied by 5, so the square root of 25 is 5.'
)
# calling the "fit" method is always mandatory after adding hidden layers to the fine-tuning
scnet.fit()

```
```bash
Loading model: 100%|██████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 956.54it/s]
Fitting model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 4/4 [00:00<00:00, 9874.76it/s]
```
Now we will test the new fine-tunings with the [HurNet](https://github.com/sapiens-technology/HurNet) neural network.
```python
from scnetwork import SCNet
scnet = SCNet()
# loading hurnet model
scnet.loadModel('./models_for_adjustment/hurnet-model')

prompt = 'If I have five fingers on one hand and two on the other, do I have an even number or an odd number?'
print('Prompt:', prompt)
print('Answer: ', end='', flush=True)
scnet.print_predict(prompt=prompt, stream=True, interval=0.05)
print()
prompt = 'The square root of 25 is what?'
print('Prompt:', prompt)
print('Answer: ', end='', flush=True)
scnet.print_predict(prompt=prompt, stream=True, interval=0.05)
print()

```
```bash
Loading model: 100%|██████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 987.03it/s]
Prompt: If I have five fingers on one hand and two on the other, do I have an even number or an odd number?
Answer: Hi! You have 7 fingers, so it's an odd number.

Prompt: The square root of 25 is what?
Answer: Hi! The square root of 25 is the same as 5 multiplied by 5, so the square root of 25 is 5.

```
Bellow, check out the possible parameters for the "fit" method.
```python
from scnetwork import SCNet
scnet = SCNet()
scnet.loadModel('./models_for_adjustment/hurnet-model')

scnet.addHiddenLayer(num_neurons=2)
scnet.addHiddenLayer(num_neurons=4)
scnet.addHiddenLayer(num_neurons=2)

scnet.addFit(
	prompt='If I have five fingers on one hand and two on the other, do I have an even number or an odd number?',
	answer="You have 7 fingers, so it's an odd number."
)
scnet.addFit(
	prompt='The square root of 25 is what?',
	answer='The square root of 25 is the same as 5 multiplied by 5, so the square root of 25 is 5.'
)
# all values below correspond to the default values of the method and belong to the same parameters of the hurnet network training method
scnet.fit( # compiles the hurnet network weights for the current fine-tuning
	interaction=True, # with True enables interactive calculation to abstract more subtle patterns, with False disables interactive calculation
	activation_function='linear', # receives a string with the name of the hurnet network activation function
	bias=0.0, # receives a floating point number with a bias value to force the weights of the hurnet network up or down
	learning_rate=1.0, # receives a floating number between 0 and 1 with the percentage value for the learning rate of the hurnet network
	stochastic_factor=False, # with False disables the addition of stochastic values to the weights of the hurnet network, with True adds stochastic values
	progress=True # with True enables the visualization of the progress of the compilation of the hurnet network weights, with False disables the visualization of the progress
) # consults the hurnet network documentation for more information

```
```bash
Loading model: 100%|██████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 956.26it/s]
Fitting model: 100%|████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 4/4 [00:00<00:00, 10505.46it/s]
```
Below we have a test with the previous fine-tuning.
```python
from scnetwork import SCNet
scnet = SCNet()
scnet.loadModel('./models_for_adjustment/hurnet-model')

prompt = 'If I have five fingers on one hand and two on the other, do I have an even number or an odd number?'
print('Prompt:', prompt)
print('Answer: ', end='', flush=True)
scnet.print_predict(prompt=prompt, stream=True, interval=0.05)
print()
prompt = 'The square root of 25 is what?'
print('Prompt:', prompt)
print('Answer: ', end='', flush=True)
scnet.print_predict(prompt=prompt, stream=True, interval=0.05)
print()

```
```bash
Loading model: 100%|██████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 960.36it/s]
Prompt: If I have five fingers on one hand and two on the other, do I have an even number or an odd number?
Answer: You have 7 fingers, so it's an odd number.

Prompt: The square root of 25 is what?
Answer: The square root of 25 is the same as 5 multiplied by 5, so the square root of 25 is 5.

```
Now look at an example without the fine-tuning progress visualization.
```python
from scnetwork import SCNet
scnet = SCNet()
scnet.loadModel('./models_for_adjustment/hurnet-model', progress=False)

scnet.addHiddenLayer(num_neurons=2)
scnet.addHiddenLayer(num_neurons=4)
scnet.addHiddenLayer(num_neurons=2)

scnet.addFit(
	prompt='If I have five fingers on one hand and two on the other, do I have an even number or an odd number?',
	answer="Hi! You have 7 fingers, so it's an odd number."
)
scnet.addFit(
	prompt='The square root of 25 is what?',
	answer='Hi! The square root of 25 is the same as 5 multiplied by 5, so the square root of 25 is 5.'
)
# with progress=False we disable the visualization of the fine-tuning progress
# the "fit" method can also be treated as a function if you use its return
result = scnet.fit(progress=False) # the return will be a boolean with True if it ran without errors, or False otherwise
if result: print('Fine-tuning completed SUCCESSFULLY!!')
else: print('ERROR completing fine-tuning.')

```
```bash
Fine-tuning completed SUCCESSFULLY!!
```
Now we will test the completed adjustment.
```python
from scnetwork import SCNet
scnet = SCNet()
scnet.loadModel('./models_for_adjustment/hurnet-model', progress=False)

prompt = 'If I have five fingers on one hand and two on the other, do I have an even number or an odd number?'
print('Prompt:', prompt)
print('Answer: ', end='', flush=True)
scnet.print_predict(prompt=prompt, stream=True, interval=0.05)
print()
prompt = 'The square root of 25 is what?'
print('Prompt:', prompt)
print('Answer: ', end='', flush=True)
scnet.print_predict(prompt=prompt, stream=True, interval=0.05)
print()

```
```bash
Prompt: If I have five fingers on one hand and two on the other, do I have an even number or an odd number?
Answer: Hi! You have 7 fingers, so it's an odd number.

Prompt: The square root of 25 is what?
Answer: Hi! The square root of 25 is the same as 5 multiplied by 5, so the square root of 25 is 5.

```
Note how it is possible to control each step of code execution based on the previous execution just by getting the return of each function
```python
from scnetwork import SCNet
scnet = SCNet()
load_model = scnet.loadModel('./models_for_adjustment/hurnet-model', progress=False)

if load_model: add_hidden_layer = scnet.addHiddenLayer(num_neurons=2)
if add_hidden_layer: add_hidden_layer = scnet.addHiddenLayer(num_neurons=4)
if add_hidden_layer: add_hidden_layer = scnet.addHiddenLayer(num_neurons=2)

scnet.addFit(
	prompt='If I have five fingers on one hand and two on the other, do I have an even number or an odd number?',
	answer="Hi! You have 7 fingers, so it's an odd number."
)
scnet.addFit(
	prompt='The square root of 25 is what?',
	answer='Hi! The square root of 25 is the same as 5 multiplied by 5, so the square root of 25 is 5.'
)
'''
as the adjustment inputs are very different from each other, this indicates that the patterns are very discrepant,
so enabling the "interaction" parameter becomes unnecessary, enabling faster training with the hurnet network by setting False for this parameter
'''
fine_tuning = scnet.fit(interaction=False, progress=False)
if fine_tuning: scnet.print_predict(prompt='The square root of 25 is what?', stream=True, interval=0.05)

```
```bash
Hi! The square root of 25 is the same as 5 multiplied by 5, so the square root of 25 is 5.
```
Notice how simple it is to apply an activation function to your training.
```python
from scnetwork import SCNet
scnet = SCNet()
load_model = scnet.loadModel('./models_for_adjustment/hurnet-model', progress=False)

if load_model: add_hidden_layer = scnet.addHiddenLayer(num_neurons=2)
if add_hidden_layer: add_hidden_layer = scnet.addHiddenLayer(num_neurons=4)
if add_hidden_layer: add_hidden_layer = scnet.addHiddenLayer(num_neurons=2)

scnet.addFit(
	prompt='If I have five fingers on one hand and two on the other, do I have an even number or an odd number?',
	answer="Hi! You have 7 fingers, so it's an odd number."
)
scnet.addFit(
	prompt='The square root of 25 is what?',
	answer='Hi! The square root of 25 is the same as 5 multiplied by 5, so the square root of 25 is 5.'
)
# example of applying a "relu" activation function to the hurnet network
fine_tuning = scnet.fit(interaction=False, activation_function='relu', progress=False)
if fine_tuning: scnet.print_predict(prompt='The square root of 25 is what?', stream=True, interval=0.05)

```
```bash
Hi! The square root of 25 is the same as 5 multiplied by 5, so the square root of 25 is 5.
```
Now check out an example below where the inference will not return the value we are expecting.
```python
from scnetwork import SCNet
scnet = SCNet()
load_model = scnet.loadModel('./models_for_adjustment/hurnet-model', progress=False)

if load_model: add_hidden_layer = scnet.addHiddenLayer(num_neurons=2)
if add_hidden_layer: add_hidden_layer = scnet.addHiddenLayer(num_neurons=4)
if add_hidden_layer: add_hidden_layer = scnet.addHiddenLayer(num_neurons=2)

scnet.addFit(prompt='What does the acronym GPT mean in language models?', answer='GPT in language models stands for Generative Pre-trained Transformer.')
scnet.addFit(prompt='Tell me what is the meaning of GPT in language models.', answer='GPT in language models stands for Generative Pre-trained Transformer.')
scnet.addFit(prompt='What is the capital of Spain?', answer='The capital of Spain is Madrid.')
scnet.addFit(prompt='Tell me the name of the capital of Spain.', answer='The capital of Spain is Madrid.')
scnet.addFit(prompt='What is the closest star to our planet?', answer='The closest star to our planet is the Sun.')

fine_tuning = scnet.fit(interaction=False, progress=False)
if fine_tuning: scnet.print_predict(prompt='What is the closest star to our planet?', stream=True, interval=0.05)
# expected: "The closest star to our planet is the Sun."
# returned: "The capital of Portugal is Lisbon."

```
```bash
The capital of Portugal is Lisbon.
```
In case of incorrect outputs, it is worth trying to change the "bias" value until you find a value that adjusts the network weights to the output we are expecting.
```python
from scnetwork import SCNet
scnet = SCNet()
load_model = scnet.loadModel('./models_for_adjustment/hurnet-model', progress=False)

if load_model: add_hidden_layer = scnet.addHiddenLayer(num_neurons=2)
if add_hidden_layer: add_hidden_layer = scnet.addHiddenLayer(num_neurons=4)
if add_hidden_layer: add_hidden_layer = scnet.addHiddenLayer(num_neurons=2)

scnet.addFit(prompt='What does the acronym GPT mean in language models?', answer='GPT in language models stands for Generative Pre-trained Transformer.')
scnet.addFit(prompt='Tell me what is the meaning of GPT in language models.', answer='GPT in language models stands for Generative Pre-trained Transformer.')
scnet.addFit(prompt='What is the capital of Spain?', answer='The capital of Spain is Madrid.')
scnet.addFit(prompt='Tell me the name of the capital of Spain.', answer='The capital of Spain is Madrid.')
scnet.addFit(prompt='What is the closest star to our planet?', answer='The closest star to our planet is the Sun.')
# with a "bias" of 0.5 we will readjust the weights to values closer to what we are expecting
# there may be cases where it is necessary to use a negative "bias"
# experiment with different "bias" values until you find one that suits your needs
fine_tuning = scnet.fit(interaction=False, bias=0.5, progress=False)
if fine_tuning: scnet.print_predict(prompt='What is the closest star to our planet?', stream=True, interval=0.05)

```
```bash
The closest star to our planet is the Sun.
```
Since in the current fine-tuning we have similar inputs, it is worth enabling the "interaction" parameter so that it is possible to abstract these subtle patterns without the need to add a "bias" value.
```python
from scnetwork import SCNet
scnet = SCNet()
load_model = scnet.loadModel('./models_for_adjustment/hurnet-model', progress=False)

if load_model: add_hidden_layer = scnet.addHiddenLayer(num_neurons=2)
if add_hidden_layer: add_hidden_layer = scnet.addHiddenLayer(num_neurons=4)
if add_hidden_layer: add_hidden_layer = scnet.addHiddenLayer(num_neurons=2)

scnet.addFit(prompt='What does the acronym GPT mean in language models?', answer='GPT in language models stands for Generative Pre-trained Transformer.')
scnet.addFit(prompt='Tell me what is the meaning of GPT in language models.', answer='GPT in language models stands for Generative Pre-trained Transformer.')
scnet.addFit(prompt='What is the capital of Spain?', answer='The capital of Spain is Madrid.')
scnet.addFit(prompt='Tell me the name of the capital of Spain.', answer='The capital of Spain is Madrid.')

fine_tuning = scnet.fit(interaction=True, progress=False)
if fine_tuning: scnet.print_predict(prompt='Tell me what is the meaning of GPT in language models.', stream=True, interval=0.05)

```
```bash
GPT in language models stands for Generative Pre-trained Transformer.
```
With the "learning_rate" parameter, you control the network's learning speed. Increasing the learning rate can provide subtle speed gains in fine-tuning training, but very high values ​​can make the network unstable and give nonsensical responses. In some datasets, keeping the learning rate below 1.0 may yield better results if your samples have very complex patterns. Very complex patterns take longer to be learned, and decreasing the learning rate will slow down learning when needed.
```python
from scnetwork import SCNet
scnet = SCNet()
load_model = scnet.loadModel('./models_for_adjustment/hurnet-model', progress=False)

if load_model: add_hidden_layer = scnet.addHiddenLayer(num_neurons=2)
if add_hidden_layer: add_hidden_layer = scnet.addHiddenLayer(num_neurons=4)
if add_hidden_layer: add_hidden_layer = scnet.addHiddenLayer(num_neurons=2)

scnet.addFit(prompt='What does the acronym GPT mean in language models?', answer='GPT in language models stands for Generative Pre-trained Transformer.')
scnet.addFit(prompt='Tell me what is the meaning of GPT in language models.', answer='GPT in language models stands for Generative Pre-trained Transformer.')
scnet.addFit(prompt='What is the capital of Spain?', answer='The capital of Spain is Madrid.')
scnet.addFit(prompt='Tell me the name of the capital of Spain.', answer='The capital of Spain is Madrid.')
# since the patterns to be abstracted are not highly complex, we can increase the learning rate without major negative consequences
# note: changing the learning rate may make fine-tuning training slower or faster depending on the number of fit attempts
fine_tuning = scnet.fit(learning_rate=1.00001, progress=False)
if fine_tuning: scnet.print_predict(prompt='What is the capital of Spain?', stream=True, interval=0.05)

```
```bash
The capital of Spain is Madrid.
```
Note that very subtle changes in the learning rate will not worsen the responses if the values are still very close to 1.0.
```python
from scnetwork import SCNet
scnet = SCNet()
load_model = scnet.loadModel('./models_for_adjustment/hurnet-model', progress=False)

if load_model: add_hidden_layer = scnet.addHiddenLayer(num_neurons=2)
if add_hidden_layer: add_hidden_layer = scnet.addHiddenLayer(num_neurons=4)
if add_hidden_layer: add_hidden_layer = scnet.addHiddenLayer(num_neurons=2)

scnet.addFit(prompt='What does the acronym GPT mean in language models?', answer='GPT in language models stands for Generative Pre-trained Transformer.')
scnet.addFit(prompt='Tell me what is the meaning of GPT in language models.', answer='GPT in language models stands for Generative Pre-trained Transformer.')
scnet.addFit(prompt='What is the capital of Spain?', answer='The capital of Spain is Madrid.')
scnet.addFit(prompt='Tell me the name of the capital of Spain.', answer='The capital of Spain is Madrid.')

fine_tuning = scnet.fit(learning_rate=0.99999, progress=False)
if fine_tuning: scnet.print_predict(prompt='What is the capital of Spain?', stream=True, interval=0.05)

```
```bash
The capital of Spain is Madrid.
```
Enabling the "stochastic_factor" parameter will make your responses more random.
```python
from scnetwork import SCNet
scnet = SCNet()
scnet.loadModel('./models_for_adjustment/hurnet-model', progress=False)

scnet.addHiddenLayer(num_neurons=2)
scnet.addHiddenLayer(num_neurons=4)
scnet.addHiddenLayer(num_neurons=2)

scnet.addFit(prompt='What does the acronym GPT mean in language models?', answer='GPT in language models stands for Generative Pre-trained Transformer.')
scnet.addFit(prompt='Tell me what is the meaning of GPT in language models.', answer='GPT in language models stands for Generative Pre-trained Transformer.')
scnet.addFit(prompt='What is the capital of Spain?', answer='The capital of Spain is Madrid.')
scnet.addFit(prompt='Tell me the name of the capital of Spain.', answer='The capital of Spain is Madrid.')
# the "stochastic_factor" parameter will increase the level of randomness in the results of the hurnet network, this will cause the responses to also become more random
# randomness will be applied between one training and another and not between inferences
# randomness in training weights can be useful where all previous attempts at parameter settings have failed
# by adding randomness to the network weights, you can repeat the training until the weights are given values closer to what you expect
scnet.fit(stochastic_factor=True, progress=False) # stochastic_factor=True will likely issue an incorrect answer due to the added randomness
scnet.print_predict(prompt='What is the capital of Spain?', stream=True, interval=0.05)

```
```bash
The capital of Spain is Lisbon.
```
Note that when you turn off randomness, the correct answer will be issued.
```python
from scnetwork import SCNet
scnet = SCNet()
scnet.loadModel('./models_for_adjustment/hurnet-model', progress=False)

scnet.addHiddenLayer(num_neurons=2)
scnet.addHiddenLayer(num_neurons=4)
scnet.addHiddenLayer(num_neurons=2)

scnet.addFit(prompt='What does the acronym GPT mean in language models?', answer='GPT in language models stands for Generative Pre-trained Transformer.')
scnet.addFit(prompt='Tell me what is the meaning of GPT in language models.', answer='GPT in language models stands for Generative Pre-trained Transformer.')
scnet.addFit(prompt='What is the capital of Spain?', answer='The capital of Spain is Madrid.')
scnet.addFit(prompt='Tell me the name of the capital of Spain.', answer='The capital of Spain is Madrid.')
# the "stochastic_factor" parameter will increase the level of randomness in the results of the hurnet network, this will cause the responses to also become more random
# randomness will be applied between one training and another and not between inferences
# randomness in training weights can be useful where all previous attempts at parameter settings have failed
# by adding randomness to the network weights, you can repeat the training until the weights are given values closer to what you expect
scnet.fit(stochastic_factor=False, progress=False) # stochastic_factor=False will issue an correct answer due to the not added randomness
scnet.print_predict(prompt='What is the capital of Spain?', stream=True, interval=0.05)

```
```bash
The capital of Spain is Madrid.
```
The SCNet module has some useful functions to facilitate the process of developing language model algorithms based on semantic comparison. One of them is the function called "loadJSON" that can load the content of a local or web JSON file, or even a string containing a JSON, always returning a dictionary or a list as a response.
```python
from scnetwork import SCNet
scnet = SCNet()
# the "file_path" parameter receives a string with the local or web path to a json file
json_loaded = scnet.loadJSON(file_path='./dataset.json') # loads the json and returns its content
print(json_loaded) # displays the contents of the converted file in a python variable
print()
# below see an example with a web file
url_path = 'https://raw.githubusercontent.com/sapiens-technology/files/refs/heads/main/dataset_en.json'
json_loaded = scnet.loadJSON(file_path=url_path) # loads the json and returns its content
print(json_loaded) # displays the contents of the converted file in a python variable
print()
# with the name parameter "string_content" you can directly pass the file content
json_content = '''
{
	"data": [
				{
					"input": "Hello! Who are you?",
					"output": "Hello! I am Sapiens Chat, an AI model created by Sapiens Technology."
				},
				{
					"input": "Who discovered Brazil?",
					"output": "Brazil was discovered by Portuguese navigators led by Pedro Alvares Cabral in 1500."
				},
				{
					"input": "What is the main language spoken in Spain?",
					"output": "The main language spoken in Spain is Spanish."
				},
				{
					"input": "What is the capital of Portugal?",
					"output": "The capital of Portugal is Lisbon."
				},
				{
					"input": "How much is 2 + 2?",
					"output": "The sum of 2 + 2 in mathematics is 4."
				},
				{
					"input": "What is five minus three?",
					"output": "The result of five minus three is two."
				},
				{
					"input": "What is your name?",
					"output": "My name is Sapiens Chat."
				}
	]
}
'''
json_loaded = scnet.loadJSON(string_content=json_content) # loads the json and returns its content
print(json_loaded) # displays the contents of the converted file in a python variable

```
```bash
{'data': [{'input': 'Hello! Who are you?', 'output': 'Hello! I am Sapiens Chat, an AI model created by Sapiens Technology.'}, {'input': 'Who discovered Brazil?', 'output': 'Brazil was discovered by Portuguese navigators led by Pedro Alvares Cabral in 1500.'}, {'input': 'What is the main language spoken in Spain?', 'output': 'The main language spoken in Spain is Spanish.'}, {'input': 'What is the capital of Portugal?', 'output': 'The capital of Portugal is Lisbon.'}, {'input': 'How much is 2 + 2?', 'output': 'The sum of 2 + 2 in mathematics is 4.'}, {'input': 'What is five minus three?', 'output': 'The result of five minus three is two.'}, {'input': 'What is your name?', 'output': 'My name is Sapiens Chat.'}]}

{'data': [{'input': 'Hello! Who are you?', 'output': 'Hello! I am Sapiens Chat, an AI model created by Sapiens Technology.'}, {'input': 'Who discovered Brazil?', 'output': 'Brazil was discovered by Portuguese navigators led by Pedro Alvares Cabral in 1500.'}, {'input': 'What is the main language spoken in Spain?', 'output': 'The main language spoken in Spain is Spanish.'}, {'input': 'What is the capital of Portugal?', 'output': 'The capital of Portugal is Lisbon.'}, {'input': 'How much is 2 + 2?', 'output': 'The sum of 2 + 2 in mathematics is 4.'}, {'input': 'What is five minus three?', 'output': 'The result of five minus three is two.'}, {'input': 'What is your name?', 'output': 'My name is Sapiens Chat.'}]}

{'data': [{'input': 'Hello! Who are you?', 'output': 'Hello! I am Sapiens Chat, an AI model created by Sapiens Technology.'}, {'input': 'Who discovered Brazil?', 'output': 'Brazil was discovered by Portuguese navigators led by Pedro Alvares Cabral in 1500.'}, {'input': 'What is the main language spoken in Spain?', 'output': 'The main language spoken in Spain is Spanish.'}, {'input': 'What is the capital of Portugal?', 'output': 'The capital of Portugal is Lisbon.'}, {'input': 'How much is 2 + 2?', 'output': 'The sum of 2 + 2 in mathematics is 4.'}, {'input': 'What is five minus three?', 'output': 'The result of five minus three is two.'}, {'input': 'What is your name?', 'output': 'My name is Sapiens Chat.'}]}
```
You can use the "key_name" parameter of the "loadJSON" function to get the contents of a specific key from your JSON.
```python
from scnetwork import SCNet
scnet = SCNet()
# the "file_path" parameter receives a string with the local or web path to a json file
json_loaded = scnet.loadJSON(file_path='./dataset.json', key_name='data') # loads the contents of the key named "data" contained in the json
print(json_loaded) # displays the contents of the converted file in a python variable
print()
# below see an example with a web file
url_path = 'https://raw.githubusercontent.com/sapiens-technology/files/refs/heads/main/dataset_en.json'
json_loaded = scnet.loadJSON(file_path=url_path, key_name='data') # loads the contents of the key named "data" contained in the json
print(json_loaded) # displays the contents of the converted file in a python variable
print()
# with the name parameter "string_content" you can directly pass the file content
json_content = '''
{
	"data": [
				{
					"input": "Hello! Who are you?",
					"output": "Hello! I am Sapiens Chat, an AI model created by Sapiens Technology."
				},
				{
					"input": "Who discovered Brazil?",
					"output": "Brazil was discovered by Portuguese navigators led by Pedro Alvares Cabral in 1500."
				},
				{
					"input": "What is the main language spoken in Spain?",
					"output": "The main language spoken in Spain is Spanish."
				},
				{
					"input": "What is the capital of Portugal?",
					"output": "The capital of Portugal is Lisbon."
				},
				{
					"input": "How much is 2 + 2?",
					"output": "The sum of 2 + 2 in mathematics is 4."
				},
				{
					"input": "What is five minus three?",
					"output": "The result of five minus three is two."
				},
				{
					"input": "What is your name?",
					"output": "My name is Sapiens Chat."
				}
	]
}
'''
json_loaded = scnet.loadJSON(string_content=json_content, key_name='data') # loads the contents of the key named "data" contained in the json
print(json_loaded) # displays the contents of the converted file in a python variable

```
```bash
[{'input': 'Hello! Who are you?', 'output': 'Hello! I am Sapiens Chat, an AI model created by Sapiens Technology.'}, {'input': 'Who discovered Brazil?', 'output': 'Brazil was discovered by Portuguese navigators led by Pedro Alvares Cabral in 1500.'}, {'input': 'What is the main language spoken in Spain?', 'output': 'The main language spoken in Spain is Spanish.'}, {'input': 'What is the capital of Portugal?', 'output': 'The capital of Portugal is Lisbon.'}, {'input': 'How much is 2 + 2?', 'output': 'The sum of 2 + 2 in mathematics is 4.'}, {'input': 'What is five minus three?', 'output': 'The result of five minus three is two.'}, {'input': 'What is your name?', 'output': 'My name is Sapiens Chat.'}]

[{'input': 'Hello! Who are you?', 'output': 'Hello! I am Sapiens Chat, an AI model created by Sapiens Technology.'}, {'input': 'Who discovered Brazil?', 'output': 'Brazil was discovered by Portuguese navigators led by Pedro Alvares Cabral in 1500.'}, {'input': 'What is the main language spoken in Spain?', 'output': 'The main language spoken in Spain is Spanish.'}, {'input': 'What is the capital of Portugal?', 'output': 'The capital of Portugal is Lisbon.'}, {'input': 'How much is 2 + 2?', 'output': 'The sum of 2 + 2 in mathematics is 4.'}, {'input': 'What is five minus three?', 'output': 'The result of five minus three is two.'}, {'input': 'What is your name?', 'output': 'My name is Sapiens Chat.'}]

[{'input': 'Hello! Who are you?', 'output': 'Hello! I am Sapiens Chat, an AI model created by Sapiens Technology.'}, {'input': 'Who discovered Brazil?', 'output': 'Brazil was discovered by Portuguese navigators led by Pedro Alvares Cabral in 1500.'}, {'input': 'What is the main language spoken in Spain?', 'output': 'The main language spoken in Spain is Spanish.'}, {'input': 'What is the capital of Portugal?', 'output': 'The capital of Portugal is Lisbon.'}, {'input': 'How much is 2 + 2?', 'output': 'The sum of 2 + 2 in mathematics is 4.'}, {'input': 'What is five minus three?', 'output': 'The result of five minus three is two.'}, {'input': 'What is your name?', 'output': 'My name is Sapiens Chat.'}]
```
The "normalization" function will normalize an input text, returning the original text in lowercase letters, without accents and without punctuation. Words with more than 5 characters will have the last 3 characters removed.
```python
from scnetwork import SCNet
scnet = SCNet()
# the parameter named "input_text" will receive the input text
input_text = 'Este é um exemplo de uma frase em Português para um teste de normalização.' # input text
output_text = scnet.normalization(input_text=input_text) # the output will be the normalized input text
print('Original:', input_text) # displays the original input text
print('Normalized:', output_text) # displays the normalized output text

```
```bash
Original: Este é um exemplo de uma frase em Português para um teste de normalização.
Normalized: este e um exem de uma frase em portug para um teste de normaliza
```
With the function named "textForEmbedding" you can convert a text of any size into a chain of embeddings, where each number in the chain represents a token (piece of word). The function's return will always be a numeric vector of the "list" type.
```python
from scnetwork import SCNet
scnet = SCNet()

input_text = 'The result of five minus three is two.' # input text
# the "textForEmbedding" function will receive a text of any size and return a vector of embeddings
# embeddings are numbers that represent a certain text tokens
# a token is a piece of text or a piece of word
embeddings_vector = scnet.textForEmbedding( # returns a vector of embeddings with the vector representation of the input text
	text=input_text, # string with the text that will be converted to a numeric list
	length=None, # desired number of elements in the return list, the smaller the number, the lower the quality of the representation; the default value is None so that the list has the exact amount of elements needed
	encoder='gpt-4' # name of the encoder used in the conversion, each encoder will have a different numeric pattern; it can be 'gpt2'/'gpt-2', 'gpt-3'/'gpt-3.5'/'gpt-3.5-turbo', 'gpt-4', 'gpt-4-turbo', 'text-davinci-001', 'text-davinci-002' or 'text-davinci-003'; default value is 'gpt-4'
) # the return will be a numeric list with the representation of the embeddings vector
print('Input text:', input_text) # displays the original input text
print('Embeddings vector:', embeddings_vector) # displays the vector of embeddings

```
```bash
Input text: The result of five minus three is two.
Embeddings vector: [791, 1121, 315, 4330, 28382, 2380, 374, 1403, 13]
```
You can also assign an integer to the "length" parameter, to set a custom number of elements for your vector. Very large values can make further processing very slow, while very small values increase the chances of incorrect results when converting the vector back to text.
```python
from scnetwork import SCNet
scnet = SCNet()

input_text = 'The result of five minus three is two.' # input text
# as an example we will use the 'gpt-4-turbo' tokenizer in encoding the elements
# length=None to obtain a vector of embeddings with its original size (recommended)
original_length = scnet.textForEmbedding(text=input_text, length=None, encoder='gpt-4-turbo')
# length=5 for an embedding vector with 5 elements (smaller than the original length)
shorter_length = scnet.textForEmbedding(text=input_text, length=5, encoder='gpt-4-turbo')
# length=15 for an embedding vector with 15 elements (larger than the original length)
longer_length = scnet.textForEmbedding(text=input_text, length=15, encoder='gpt-4-turbo')

print('Original length:', original_length) # displays the embeddings vector with its original size
print('Shorter length:', shorter_length) # displays the embeddings vector with its reduced size
print('Longer length:', longer_length) # displays the embeddings vector with its increased size

```
```bash
Original length: [791, 1121, 315, 4330, 28382, 2380, 374, 1403, 13]
Shorter length: [791, 1121, 315, 4330, 6510]
Longer length: [791, 1121, 315, 4330, 28382, 2380, 374, 1403, 13, 220, 220, 220, 220, 220, 220]
```
With the "embeddingForText" function you can convert a vector of embeddings into its original text (when possible).
```python
from scnetwork import SCNet
scnet = SCNet()

input_text = 'The result of five minus three is two.' # input text
# as an example we will use the 'gpt-4-turbo' tokenizer in encoding the elements
tokenizer = 'gpt-4-turbo' # encoding tokenizer name
# length=None to obtain a vector of embeddings with its original size (recommended)
original_length = scnet.textForEmbedding(text=input_text, length=None, encoder=tokenizer)
# length=5 for an embedding vector with 5 elements (smaller than the original length)
shorter_length = scnet.textForEmbedding(text=input_text, length=5, encoder=tokenizer)
# length=15 for an embedding vector with 15 elements (larger than the original length)
longer_length = scnet.textForEmbedding(text=input_text, length=15, encoder=tokenizer)
# attempt to recover the original text
# the greater the distance between the number of elements in the altered vector and the number of elements in the original vector, the more inaccurate the recovery will be
# always use the same value of "encoder" in text recovery that was used in the conversion of the embedding vector
# the "embedding" parameter receives a vector of embeddings to be converted to text (always receives a value of type list)
# the "encoder" parameter must receive the same string that was used in the parameter of the same name in the "textForEmbedding" function
text_of_original_vector = scnet.embeddingForText(embedding=original_length, encoder=tokenizer)
text_of_vector_decreased = scnet.embeddingForText(embedding=shorter_length, encoder=tokenizer)
text_of_augmented_vector = scnet.embeddingForText(embedding=longer_length, encoder=tokenizer)

print(text_of_original_vector) # if there is no decrease in size, the original text will always be recovered
print(text_of_vector_decreased) # if there is a decrease in size, there may be inaccuracies in recovering the original text
print(text_of_augmented_vector) # if the size is increased, the original text can be recovered with trailing white spaces

```
```bash
The result of five minus three is two.
The result of five attr
The result of five minus three is two.
```
The "strip" parameter of the recovery function, if set to True, will eliminate leading and trailing white spaces from the recovered text (if any).
```python
from scnetwork import SCNet
scnet = SCNet()

input_text = 'The result of five minus three is two.' # input text
# length=15 for an embedding vector with 15 elements (larger than the original length)
# when we increase the length of the embeddings vector, elements representing whitespace are added to the end of the vector
longer_length = scnet.textForEmbedding(text=input_text, length=15)
# the "strip" parameter receives a boolean with True to remove whitespace around the text, or False to keep these whitespaces
original_text_with_spaces = scnet.embeddingForText(embedding=longer_length, strip=False)
original_text_without_spaces = scnet.embeddingForText(embedding=longer_length, strip=True)

print(f'Text with {len(original_text_with_spaces)} characters: [{original_text_with_spaces}]')
print(f'Text with {len(original_text_without_spaces)} characters: [{original_text_without_spaces}]')

```
```bash
Text with 44 characters: [The result of five minus three is two.      ]
Text with 38 characters: [The result of five minus three is two.]
```
With the function named "countTokens" you can count the total amount of tokens in a text using a certain counting pattern based on the encoder.
```python
from scnetwork import SCNet
scnet = SCNet()

input_text = 'The result of five minus three is two.'
encoder = 'gpt-4' # tokenizer used to encode the tokens converted into embeddings and the tokens to be counted
# the "text" parameter receives the text to be converted; the "encoder" parameter receives the tokenizer used in the conversion
embeddings = scnet.textForEmbedding(text=input_text, encoder=encoder)
# the "text" parameter receives the text to be counted; the "encoder" parameter receives the tokenizer used in the count
tokens_count = scnet.countTokens(text=input_text, encoder=encoder)
# the total number of tokens counted will always be equal to the number of elements in the embeddings vector
print('Embeddings length:', len(embeddings)) # displays the length of the embeddings vector
print('Tokens count:', tokens_count) # displays the total number of tokens in the input text

```
```bash
Embeddings length: 9
Tokens count: 9
```
With the "truncateTokens" function you can truncate a text by removing a certain amount of tokens from its end.
```python
from scnetwork import SCNet
scnet = SCNet()

original_text = 'The result of five minus three is two.' # text that should be truncated
encoder = 'gpt-4' # encoder of the tokenization used in truncation
# the "max_tokens" parameter receives an integer with the maximum amount of tokens to be obtained from the input text
# if the total amount of tokens in the string of the "text" parameter is less than or equal to that defined in the "max_tokens" parameter, the return value will be the same as the input text
# if the total amount of tokens in the string of the "text" parameter is greater than that defined in the "max_tokens" parameter, the return value will be the input text with only the first n tokens defined in "max_tokens"
# the default value of "max_tokens" is 1000000 and the default value of all parameters with the name "encoder" contained in the module will always be 'gpt-4'
truncated_text = scnet.truncateTokens(text=original_text, max_tokens=5, encoder=encoder)

print('Original text:', original_text) # displays the original text
print('Truncated text:', truncated_text) # displays the truncated text

```
```bash
Original text: The result of five minus three is two.
Truncated text: The result of five minus
```
The "textSummary" function is used to summarize an input text so that it does not exceed a certain number of tokens.
```python
from scnetwork import SCNet
scnet = SCNet()

original_text = '''
What is Artificial Intelligence (AI)?
Artificial Intelligence is a field of computer science that develops systems capable of performing tasks that would normally require human intelligence, such as learning, reasoning, perception, and decision-making.

What is the difference between weak AI and strong AI?
Weak AI (or narrow AI) is designed for specific tasks, such as virtual assistants and recommendation systems. Strong AI, on the other hand, would have self-awareness and the ability to understand and reason about any subject like a human being.

What is Machine Learning and how is it related to AI?
Machine Learning is a subfield of AI that teaches machines to learn patterns from data without explicit programming. It allows systems to improve their performance over time with experience.
'''
summarized_text = scnet.textSummary( # returns the summary of the string contained in the parameter named "text"
	text=original_text, # value of the string to summarize; the default value for all module string parameters is an empty string
	max_tokens=50, # maximum number of tokens desired in the text summary; default value is 1000000
	encoder='gpt-4' # tokenizer used in digest encoding; default value is 'gpt-4'
)

print('Original text:', original_text.strip()) # displays the original text without the summary
print('-'*150) # displays a dashed line to make it easier to understanding between the two visualizations
print('Summarized text:', summarized_text) # displays the summarized text without exceeding the maximum token limit

```
```bash
Original text: What is Artificial Intelligence (AI)?
Artificial Intelligence is a field of computer science that develops systems capable of performing tasks that would normally require human intelligence, such as learning, reasoning, perception, and decision-making.

What is the difference between weak AI and strong AI?
Weak AI (or narrow AI) is designed for specific tasks, such as virtual assistants and recommendation systems. Strong AI, on the other hand, would have self-awareness and the ability to understand and reason about any subject like a human being.

What is Machine Learning and how is it related to AI?
Machine Learning is a subfield of AI that teaches machines to learn patterns from data without explicit programming. It allows systems to improve their performance over time with experience.
------------------------------------------------------------------------------------------------------------------------------------------------------
Summarized text: What is Artificial Intelligence (AI)?
Artificial Intelligence is a field of computer science such as virtual assistants and recommendation systems. Strong AI, on the other hand without explicit programming. It allows systems to improve their performance over time with experience.
```
The "getTrainingText" function retrieves all text from the dataset used in base training and fine-tuning training when possible. Only fine-tuning texts with hidden layers cannot be retrieved.
```python
from scnetwork import SCNet
scnet = SCNet()
# saving a new base model
scnet.train(dataset_path='dataset.json')
scnet.saveModel('./models_for_adjustment/model03')
# base model loading
scnet.loadModel('./models_for_adjustment/model03')
# application of fine-tuning
scnet.addFit(prompt='What does the acronym GPT mean in language models?', answer='GPT in language models stands for Generative Pre-trained Transformer.')
scnet.addFit(prompt='Tell me what is the meaning of GPT in language models.', answer='GPT in language models stands for Generative Pre-trained Transformer.')
scnet.addFit(prompt='What is the capital of Spain?', answer='The capital of Spain is Madrid.')
scnet.addFit(prompt='Tell me the name of the capital of Spain.', answer='The capital of Spain is Madrid.')

```
```bash
Reading data: 100%|██████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 1/1 [00:00<00:00, 9709.04it/s]
Embedding data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████| 1000000/1000000 [00:00<00:00, 18778307.57it/s]
Tokenizing data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 110.06it/s]
Training model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 127.74it/s]
Generating 91U parameters: 100%|████████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 35951.18it/s]
Saving model: 100%|██████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 5/5 [00:00<00:00, 4475.36it/s]
Loading model: 100%|████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 14406.34it/s]
```
Now, to retrieve the training and fine-tuning data, simply call the "getTrainingText" function, which will return all the retrieved text.
```python
from scnetwork import SCNet
scnet = SCNet()
scnet.loadModel('./models_for_adjustment/model03') # load the model you want to retrieve the contents of the dataset
recovered_text = scnet.getTrainingText() # the return will be all the content used in training and fine-tuning (if there is fine-tuning)
print(recovered_text) # displays all content retrieved from the loaded model

```
```bash
Loading model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 3954.76it/s]
Hello! Who are you?
Hello! I am Sapiens Chat, an AI model created by Sapiens Technology.

Who discovered Brazil?
Brazil was discovered by Portuguese navigators led by Pedro Alvares Cabral in 1500.

What is the main language spoken in Spain?
The main language spoken in Spain is Spanish.

What is the capital of Portugal?
The capital of Portugal is Lisbon.

How much is 2 + 2?
The sum of 2 + 2 in mathematics is 4.

What is five minus three?
The result of five minus three is two.

What is your name?
My name is Sapiens Chat.

What does the acronym GPT mean in language models?
GPT in language models stands for Generative Pre-trained Transformer.

Tell me what is the meaning of GPT in language models.
GPT in language models stands for Generative Pre-trained Transformer.

What is the capital of Spain?
The capital of Spain is Madrid.

Tell me the name of the capital of Spain.
The capital of Spain is Madrid.
```
You can also use the "max_tokens" name parameter of the "getTrainingText" function to limit the number of tokens in the retrieved text, so that the returned string will never exceed that amount of tokens.
```python
from scnetwork import SCNet
scnet = SCNet()
scnet.loadModel('./models_for_adjustment/model03') # load the model you want to retrieve the contents of the dataset
# "max_tokens" receives an integer with the maximum amount of tokens that can be returned by the function; the default value is None, so that all the retrieved text is returned without limitations
recovered_text = scnet.getTrainingText(max_tokens=50) # returns 50 tokens from the entire dataset used in training and fine-tuning the loaded model
print(recovered_text) # displays all content retrieved from the loaded model

```
```bash
Loading model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 4091.43it/s]
Hello! Who are you?
Hello! I am Sapiens Chat, an AI model created by Sapiens Technology.

Who discovered Brazil?
Brazil was discovered by Portuguese navigators led by Pedro Alvares Cabral in 1500.
```
The "alternativeVectorComparison" function compares a numeric vector with the numeric vectors of a matrix and returns the index and vector of the matrix that most closely matches the input vector. The comparison is made with differentiation calculations, without considering the order of the elements.
```python
from scnetwork import SCNet
scnet = SCNet()
# example of an input vector that will be compared to each of the vectors in the matrix
input_vector = [20, 50, 10, 40, 30]
# example of a matrix that will have each of its vector elements compared to the input vector
vector_matrix = [
	[2, 1, 5, 4, 3],
	[30, 50, 10, 40, 20],
	[200, 300, 500, 400, 100]
]
# both the elements of the input vector and the elements of each of the matrix vectors must be integers or floating points numbers
# the return will be a tuple with the index of the closest vector in the first position and the closest vector itself in the second position
# both the "input_vector" parameter and the "vector_matrix" parameter must receive values of type list and both have empty lists as default values
index, vector = scnet.alternativeVectorComparison( # returns a tuple with the index and vector closest to the value of "input_vector"
	input_vector=input_vector, # list with the numeric vector that will be compared to the matrix; the list can have as many elements as you want
	vector_matrix=vector_matrix # matrix with the numeric vectors to be compared with the input vector; the matrix can have as many vectors as you want and the vectors can have any number of elements
)

print('Index:', index) # displays the index belonging to the matrix vector that most closely matches the input vector
print('Vector:', vector) # displays the matrix vector that most closely matches the input vector

```
```bash
Index: 1
Vector: [30, 50, 10, 40, 20]
```
Note that the matrix vectors and the input vector may have different numbers of elements.
```python
from scnetwork import SCNet
scnet = SCNet()
# example of an input vector that will be compared to each of the vectors in the matrix
input_vector = [20, 50, 10, 40] # 4 elements
# example of a matrix that will have each of its vector elements compared to the input vector
vector_matrix = [
	[2, 1, 5, 4, 3, 0], # 6 elements
	[30, 50, 10, 40, 20], # 5 elements
	[200, 300, 500] # 3 elements
]

index, vector = scnet.alternativeVectorComparison(input_vector=input_vector, vector_matrix=vector_matrix)

print('Index:', index) # displays the index belonging to the matrix vector that most closely matches the input vector
print('Vector:', vector) # displays the matrix vector that most closely matches the input vector

```
```bash
Index: 1
Vector: [30, 50, 10, 40, 20]
```
The "vectorComparison" function works similarly to the "alternativeVectorComparison" function, but uses a cosine similarity calculation to perform the comparison and in some cases may provide different results than the previous function. You will need to perform experiments to decide which vector comparison function is best for your specific case.
```python
from scnetwork import SCNet
scnet = SCNet()
# example of an input vector that will be compared to each of the vectors in the matrix
input_vector = [20, 50, 10, 40, 30]
# example of a matrix that will have each of its vector elements compared to the input vector
vector_matrix = [
	[2, 1, 5, 4, 3],
	[30, 50, 10, 40, 20],
	[200, 300, 500, 400, 100]
]
# both the elements of the input vector and the elements of each of the matrix vectors must be integers or floating points numbers
# the return will be a tuple with the index of the closest vector in the first position and the closest vector itself in the second position
# both the "input_vector" parameter and the "vector_matrix" parameter must receive values of type list and both have empty lists as default values
index, vector = scnet.vectorComparison( # returns a tuple with the index and vector closest to the value of "input_vector"
	input_vector=input_vector, # list with the numeric vector that will be compared to the matrix; the list can have as many elements as you want
	vector_matrix=vector_matrix, # matrix with the numeric vectors to be compared with the input vector; the matrix can have as many vectors as you want and the vectors can have any number of elements,
	stream=True # if True, will load the matrix by "streaming" without it being completely loaded into memory, if False, it will load the complete matrix into memory; the default value is True to avoid memory overflows with very large matrices
)

print('Index:', index) # displays the index belonging to the matrix vector that most closely matches the input vector
print('Vector:', vector) # displays the matrix vector that most closely matches the input vector

```
```bash
Index: 1
Vector: [30, 50, 10, 40, 20]
```
Note that the matrix vectors and the input vector may have different numbers of elements.
```python
from scnetwork import SCNet
scnet = SCNet()
# example of an input vector that will be compared to each of the vectors in the matrix
input_vector = [20, 50, 10, 40] # 4 elements
# example of a matrix that will have each of its vector elements compared to the input vector
vector_matrix = [
	[2, 1, 5, 4, 3, 0], # 6 elements
	[30, 50, 10, 40, 20], # 5 elements
	[200, 300, 500] # 3 elements
]

index, vector = scnet.vectorComparison(input_vector=input_vector, vector_matrix=vector_matrix, stream=True)

print('Index:', index) # displays the index belonging to the matrix vector that most closely matches the input vector
print('Vector:', vector) # displays the matrix vector that most closely matches the input vector

```
```bash
Index: 1
Vector: [30, 50, 10, 40, 20]
```
The result without using "streaming" will be the same, but memory consumption will be higher.
```python
from scnetwork import SCNet
scnet = SCNet()

input_vector = [20, 50, 10, 40]
vector_matrix = [[2, 1, 5, 4, 3, 0], [30, 50, 10, 40, 20], [200, 300, 500]]
# stream=False to load the complete matrix into available memory
index, vector = scnet.vectorComparison(input_vector=input_vector, vector_matrix=vector_matrix, stream=False)

print('Index:', index) # displays the index belonging to the matrix vector that most closely matches the input vector
print('Vector:', vector) # displays the matrix vector that most closely matches the input vector

```
```bash
Index: 1
Vector: [30, 50, 10, 40, 20]
```
The "similarityBetweenVectors" function returns a percentage level between 0 and 1 of how similar two vectors are.
```python
from scnetwork import SCNet
scnet = SCNet()
# vectors that will be compared
first_vector = [1, 2, 3, 4, 5]
second_vector = [1, 3, 3, 4, 5]
# the "first_vector" parameter receives a vector to be compared to another vector in "second_vector"; the default value for both is an empty list
# this function also calculates the distance between elements
similarity = scnet.similarityBetweenVectors(first_vector=first_vector, second_vector=second_vector) # returns a floating number between 0 and 1 with a similarity percentage
print('Similarity between the vectors:', similarity) # displays the level of similarity between the vectors

# for equal vectors the similarity returned will be 1.0 (100%)
first_vector = [1, 2, 3, 4, 5]
second_vector = [1, 2, 3, 4, 5]
similarity = scnet.similarityBetweenVectors(first_vector=first_vector, second_vector=second_vector)
print('Similarity between the vectors:', similarity)

# the order of the elements is considered in the comparison
first_vector = [1, 2, 3, 4, 5]
second_vector = [5, 4, 3, 2, 1]
similarity = scnet.similarityBetweenVectors(first_vector=first_vector, second_vector=second_vector)
print('Similarity between the vectors:', similarity)

```
```bash
Similarity between the vectors: 0.975
Similarity between the vectors: 1.0
Similarity between the vectors: 0.7
```
The "euclideanSimilarity" function works like the "similarityBetweenVectors" function, but the similarity calculation is done based on the Euclidean distance.
```python
from scnetwork import SCNet
scnet = SCNet()
# vectors that will be compared
first_vector = [1, 2, 3, 4, 5]
second_vector = [1, 3, 3, 4, 5]
# the "first_vector" parameter receives a vector to be compared to another vector in "second_vector"; the default value for both is an empty list
# this function only calculates the difference between elements in the same position
similarity = scnet.euclideanSimilarity(first_vector=first_vector, second_vector=second_vector) # returns a floating number between 0 and 1 with a similarity percentage
print('Similarity between the vectors:', similarity) # displays the level of similarity between the vectors

# for equal vectors the similarity returned will be 1.0 (100%)
first_vector = [1, 2, 3, 4, 5]
second_vector = [1, 2, 3, 4, 5]
similarity = scnet.euclideanSimilarity(first_vector=first_vector, second_vector=second_vector)
print('Similarity between the vectors:', similarity)

# the order of the elements is considered in the comparison
first_vector = [1, 2, 3, 4, 5]
second_vector = [5, 4, 3, 2, 1]
similarity = scnet.euclideanSimilarity(first_vector=first_vector, second_vector=second_vector)
print('Similarity between the vectors:', similarity)

```
```bash
Similarity between the vectors: 0.9333333333333333
Similarity between the vectors: 1.0
Similarity between the vectors: 0.48
```
Both functions accept vectors with different lengths in the comparison.
```python
from scnetwork import SCNet
scnet = SCNet()
# vectors that will be compared
first_vector = [1, 2, 3, 4, 5, 6, 7, 8, 9]
second_vector = [1, 3, 3, 4, 5]
similarity = scnet.similarityBetweenVectors(first_vector=first_vector, second_vector=second_vector)
print('Similarity between the vectors:', similarity) # displays the level of similarity between the vectors
similarity = scnet.euclideanSimilarity(first_vector=first_vector, second_vector=second_vector)
print('Similarity between the vectors:', similarity) # displays the level of similarity between the vectors

```
```bash
Similarity between the vectors: 0.975
Similarity between the vectors: 0.9333333333333333
```
The "overwriteWithHexadecimal" function converts any file type into an unreadable binary file with hexadecimal content, but be careful, the original file will be overwritten and will no longer exist.
```python
from scnetwork import SCNet
scnet = SCNet()
# the "overwriteWithHexadecimal" function receives the path of a local file and convert this file it to a binary file of hexadecimals
# the "file_path" parameter receives the path to a file that should be converted; the default value is an empty string
# CAUTION: THE ORIGINAL FILE WILL BE OVERWRITTEN
result = scnet.overwriteWithHexadecimal(file_path='dataset.json') # returns True if the conversion is successful, or False otherwise
if result: print('File converted SUCCESSFULLY!!') # displays a success message if the conversion is successful
else: print('ERROR converting file.') # displays a failure message if there is any error during the conversion process

```
```bash
File converted SUCCESSFULLY!!
```
The "restoreOriginalFromHexadecimal" function restores a file converted with the "overwriteWithHexadecimal" function recovering the original file.
```python
from scnetwork import SCNet
scnet = SCNet()
# the "restoreOriginalFromHexadecimal" function receives the path of a local file and restores this binary file from hexadecimals to its original contents
# the "file_path" parameter receives the path to a file that should be restored; the default value is an empty string
# CAUTION: THE ORIGINAL FILE WILL BE OVERWRITTEN
result = scnet.restoreOriginalFromHexadecimal(file_path='dataset.json') # returns True if the restoration is successful, or False otherwise
if result: print('File restored SUCCESSFULLY!!') # displays a success message if the restoration is successful
else: print('ERROR restoring file.') # displays a failure message if there is any error during the restoration process

```
```bash
File restored SUCCESSFULLY!!
```
The "outputAdaptedToInput" function adapts a language model output to a given user prompt so that the model can generalize the response to inputs other than those used in the training process.
```python
from scnetwork import SCNet
scnet = SCNet()
# the "prompt" variable will be compared to the "original_input" variable and the identified differences will be applied to the value of "original_output"
prompt = 'What is the principal language spoken in Mexico?' # input that will have its pattern transferred to the original output
original_input = 'What is the main language spoken in Spain?' # original input that is required for the output to occur
original_output = 'The main language spoken in Spain is Spanish.' # original output that would be returned if the original input was sent to the prompt
# the "outputAdaptedToInput" function will transfer the patterns from one text to another
adapted_output = scnet.outputAdaptedToInput( # returns a string with the value of the "original_output" parameter adapted to the value of the "prompt" parameter
	prompt=prompt, # receives a string with the user input; default value is an empty string
	original_input=original_input, # receives a string with the training input or the fine-tuning input; the default value is an empty string
	original_output=original_output # receives a string with the training output or the fine-tuning output; the default value is an empty string
) # the result will be a response to a given input adapted to the terms of an input other than the original
# display of the variables used in the function parameters and the final value with the patterns assimilated in the generalization
print('Prompt:', prompt) # displays the input that had the pattern assimilated
print('Original input:', original_input) # displays the input that was used to identify the differences with the prompt
print('Original output:', original_output) # displays the original output which should be adapted to the prompt standards
print('Adapted output:', adapted_output) # displays the adapted output resulting from transferring patterns from the prompt to the original output

```
```bash
Prompt: What is the principal language spoken in Mexico?
Original input: What is the main language spoken in Spain?
Original output: The main language spoken in Spain is Spanish.
Adapted output: The principal language spoken in Mexico is Spanish.
```
With the "downloadFile" function, you can locally download any file contained in a public address of WEB.
```python
from scnetwork import SCNet
scnet = SCNet()

url_path = 'https://raw.githubusercontent.com/sapiens-technology/files/refs/heads/main/dataset.txt'
file_path = './downloaded_file.txt'
# the "url_path" parameter receives a string with the web address from where the file should be downloaded; the default value is an empty string that will not download anything
# the "file_path" parameter receives a string with the local path to where the file should be downloaded; the default value is an empty string
# if "file_path" is not set, the file will be downloaded with its original name in the local directory
# if "file_path" receives an existing directory name, the file will be downloaded inside this directory with its original name
# if "file_path" receives a nonexistent directory name, the directory will be created and the file will be downloaded inside it with its original name
# if "file_path" receives an existing directory name followed by a file name, the file will be downloaded inside this directory with the defined name
# if "file_path" receives a nonexistent directory name followed by a file name, the directory will be created and the file will be downloaded inside it with the defined name
downloaded = scnet.downloadFile( # returns True if the file defined in "url_path" is downloaded successfully, or False otherwise
	url_path=url_path, # web address for the file to be downloaded
	file_path=file_path # path to the directory and/or name of the file generated with the download; if not defined the file will be downloaded with the original name in the local directory
)

if downloaded: print('File downloaded SUCCESSFULLY!!')
else: print('ERROR downloading file.')

```
```bash
File downloaded SUCCESSFULLY!!
```
If no name is set for the destination file in "file_path", the file will be saved with the original name in the current directory.
```python
from scnetwork import SCNet
scnet = SCNet()

downloaded = scnet.downloadFile('https://raw.githubusercontent.com/sapiens-technology/files/refs/heads/main/dataset.txt')
if downloaded: print('File downloaded SUCCESSFULLY!!')
else: print('ERROR downloading file.')

```
```bash
File downloaded SUCCESSFULLY!!
```
The class constructor also has parameters that can be manipulated. The "parallel_prediction" parameter receives a boolean value that when equal to True will use the cosine similarity calculation of the "vectorComparison" function with parallel computing, more recommended for hardware with GPU or in cases of use with CPU where the data set is relatively small, with powerful hardware this will make the inferences faster. However, if "parallel_prediction" is equal to False, which is its default value, the inference answers will be returned through sequential calculations with the "alternativeVectorComparison" function, which despite being slower, brings more precision answers.
```python
from scnetwork import SCNet
# if "parallel_prediction" is equal to False, the "vectorComparison" function will delegate the calculation to the "alternativeVectorComparison" function
# if "parallel_prediction" is equal to True, the "vectorComparison" function will respond with its own calculation; The default value of "parallel_prediction" is False
input_vector = [1, 2, 3, 4, 5]
vector_matrix = [
	[0, 1, 2, 3, 4],
	[5, 4, 3, 2, 1],
	[1, 2, 3, 4, 5],
	[50, 40, 30, 20, 10],
	[10, 20, 30, 40, 50]
]
# for subtle differences between data "parallel_prediction" equal to True works best
scnet = SCNet(parallel_prediction=True)
index, vector = scnet.vectorComparison(input_vector=input_vector, vector_matrix=vector_matrix)
print('Index with parallel prediction:', index)
print('Vector with parallel prediction:', vector)
print()
# since the differences between the numbers in the input vector and the answer candidates in the matrix are small, the answer will be inaccurate with "parallel_prediction" equal to False
scnet = SCNet(parallel_prediction=False)
index, vector = scnet.vectorComparison(input_vector=input_vector, vector_matrix=vector_matrix)
print('Index without parallel prediction:', index)
print('Vector without parallel prediction:', vector)

```
```bash
Index with parallel prediction: 2
Vector with parallel prediction: [1, 2, 3, 4, 5]

Index without parallel prediction: 1
Vector without parallel prediction: [5, 4, 3, 2, 1]
```
In inference predictions with language models, "parallel_prediction" works the same way. When set to True, the answers will be more accurate when the embedding numbers have little difference between them, in addition to providing faster inference on hardware that can load the entire matrix into memory. If "parallel_prediction" is set to False, which is its default value, we will have more accurate answers when the difference between the embedding numbers is greater and the calculation will be performed sequentially, which may become slower on limited hardware. You should apply tests to find out which configuration is best for your specific case.
```python
from scnetwork import SCNet
scnet = SCNet(parallel_prediction=True)
scnet.train(dataset_path='dataset.json', progress=False)
scnet.saveModel('./my_models/my_model', False) # saves the model with "parallel_prediction" equal to True

scnet = SCNet() # if "parallel_prediction" is saved with True, it will overwrite any value of "parallel_prediction" in the constructor before of loading
scnet.loadModel('./my_models/my_model', False) # loading will prioritize "parallel_prediction" equal to True from the saved model
scnet.print_predict('What is the capital of Portugal?') # incorrect answer because there is a big difference between the training embeddings

scnet = SCNet(parallel_prediction=False) # it will have no effect to set "parallel_prediction" to False when it has already been saved as True
scnet.loadModel('./my_models/my_model', False) # loading will prioritize "parallel_prediction" equal to True from the saved model
scnet.print_predict('What is the capital of Portugal?') # incorrect answer because there is a big difference between the training embeddings

scnet = SCNet(parallel_prediction=False) # "parallel_prediction" equal to False is the same as not setting "parallel_prediction"
scnet.train(dataset_path='dataset.json', progress=False)
scnet.saveModel('./my_models/my_model', False) # saves the model with "parallel_prediction" equal to False

scnet = SCNet(parallel_prediction=True) # if "parallel_prediction" was saved with False, you can change its value in the constructor before loading and this value will be the respected value
scnet.loadModel('./my_models/my_model', False) # loading will prioritize "parallel_prediction" equal to True from constructor
scnet.print_predict('What is the capital of Portugal?') # incorrect answer because there is a big difference between the training embeddings

scnet = SCNet() # the value of "parallel_prediction" will remain with the value False of the saved model
scnet.loadModel('./my_models/my_model', False) # loading will prioritize "parallel_prediction" equal to False from the saved model
scnet.print_predict('What is the capital of Portugal?') # correct answer because there is a big difference between the training embeddings

scnet = SCNet() # the effect will be the same as "SCNet(parallel_prediction=False)"
scnet.train(dataset_path='dataset.json', progress=False)
scnet.saveModel('./my_models/my_model', False) # saves the model with "parallel_prediction" equal to False

scnet = SCNet() # the value of "parallel_prediction" will remain with the value False of the saved model
scnet.loadModel('./my_models/my_model', False) # loading will prioritize "parallel_prediction" equal to False from the saved model
scnet.print_predict('What is the capital of Portugal?') # correct answer because there is a big difference between the training embeddings

```
```bash
The main language spoken in Spain is Spanish.
The main language spoken in Spain is Spanish.
The main language spoken in Spain is Spanish.
The capital of Portugal is Lisbon.
The capital of Portugal is Lisbon.
```
With the "device" parameter you can define a device to run the current inference on. GPU and TPU devices will only apply to models run with "parallel_prediction" equal to True.
```python
from scnetwork import SCNet
# with the "device" parameter equal to None, the module will automatically choose the best available device; the default value is None
scnet = SCNet(device=None) # device=None will cause inference to occur with the best device available on the current machine
scnet.loadModel('./my_models/my_model', False)
scnet.print_predict('Hello! Who are you?')
# if the set device is not available on the local machine, inference processing will be transferred to the cpu
scnet = SCNet(device='gpu', parallel_prediction=True) # device='gpu' runs the inference on gpu; requires "parallel_prediction" to be equal to True
scnet.loadModel('./my_models/my_model', False)
scnet.print_predict('Hello! Who are you?')
# "device" equals 'gpu' and "device" equals 'cuda' are interpreted as the same thing
scnet = SCNet(device='cuda', parallel_prediction=True) # device='cuda' runs the inference on gpu; requires "parallel_prediction" to be equal to True
scnet.loadModel('./my_models/my_model', False)
scnet.print_predict('Hello! Who are you?')
# both gpu processing and tpu processing require parallel prediction to be enabled
scnet = SCNet(device='tpu', parallel_prediction=True) # device='tpu' runs the inference on tpu; requires "parallel_prediction" to be equal to True
scnet.loadModel('./my_models/my_model', False)
scnet.print_predict('Hello! Who are you?')
# "device" equals 'tpu' and "device" equals 'xla' are interpreted as the same thing
scnet = SCNet(device='xla', parallel_prediction=True) # device='xla' runs the inference on tpu; requires "parallel_prediction" to be equal to True
scnet.loadModel('./my_models/my_model', False)
scnet.print_predict('Hello! Who are you?')
# use "device" equal to 'mps' to use apple's macos video memory which works like an integrated gpu
scnet = SCNet(device='mps', parallel_prediction=True) # device='mps' runs the inference with macos video memory; requires "parallel_prediction" to be equal to True
scnet.loadModel('./my_models/my_model', False)
scnet.print_predict('Hello! Who are you?')
# use "device" equal to 'cpu' to run the inference on devices without gpu, tpu or mps
# by setting cpu to device you eliminate the automatic selection of "device=None" and make execution on cpu slightly faster
scnet = SCNet(device='cpu', parallel_prediction=True) # device='cpu' runs the inference on cpu; requires "parallel_prediction" to be equal to True
scnet.loadModel('./my_models/my_model', False)
scnet.print_predict('Hello! Who are you?')

```
```bash
Hello! I am Sapiens Chat, an AI model created by Sapiens Technology.
Hello! I am Sapiens Chat, an AI model created by Sapiens Technology.
Hello! I am Sapiens Chat, an AI model created by Sapiens Technology.
Hello! I am Sapiens Chat, an AI model created by Sapiens Technology.
Hello! I am Sapiens Chat, an AI model created by Sapiens Technology.
Hello! I am Sapiens Chat, an AI model created by Sapiens Technology.
Hello! I am Sapiens Chat, an AI model created by Sapiens Technology.
```
The "user_id" parameter is used so that inference does not confuse the flow of chat models between different users.
```python
from scnetwork import SCNet
scnet = SCNet()

scnet.train(dataset_path='dataset.json')
# it is necessary to save a model and then load it before any fitting or inference
# the saving and loading do not necessarily need to be in the same code file
scnet.saveModel('./models/user-model')
scnet.loadModel('./models/user-model')

scnet.addFit(prompt='Hello!', answer='Hello everything is fine?', id=1, relationship_id=0)
scnet.addFit(prompt='Yes.', answer="Very good, I'm glad to know.", id=2, relationship_id=1)
scnet.addFit(prompt='No.', answer="What a shame, I'm so sorry for you.", id=3, relationship_id=1)
scnet.addFit(prompt='Are you a robot?', answer='Yes, I am, and you?', id=4, relationship_id=0)
scnet.addFit(prompt='No.', answer="It's good to know I'm talking to a human.", id=5, relationship_id=4)
scnet.addFit(prompt='Yes.', answer='How interesting, we will be two robots talking.', id=6, relationship_id=4)
scnet.addFit(prompt="Hey how's it going?", answer='Hi, I am fine. Can I help you with anything?', id=7, relationship_id=0)
scnet.addFit(prompt='Yes you can. I would like to know what the capital of Denmark is.', answer='The capital of Denmark is Copenhagen.', id=8, relationship_id=7)
scnet.addFit(prompt="No, you can't.", answer="Okay, if you need anything I'll be here.", id=9, relationship_id=7)

```
```bash
Reading data: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 1/1 [00:00<00:00, 12671.61it/s]
Embedding data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████| 1000000/1000000 [00:00<00:00, 18786718.56it/s]
Tokenizing data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 110.11it/s]
Training model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 127.85it/s]
Generating 91U parameters: 100%|████████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 47662.55it/s]
Saving model: 100%|██████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 5/5 [00:00<00:00, 4629.47it/s]
Loading model: 100%|████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 14585.26it/s]
```
See what happens if we try to interact with two users at the same time without setting the "user_id" parameter for each of them.
```python
from scnetwork import SCNet
# please note that when we do not set a user id for the loading, the chat cannot know what the current user's last response was
scnet = SCNet()
scnet.loadModel('./models/user-model', False)

prompt = 'Hello!'
print('User 1:', prompt)
print('Bot: ', end='', flush=True)
scnet.print_predict(prompt=prompt, stream=True, interval=0.05)
print()

scnet = SCNet()
scnet.loadModel('./models/user-model', False)
prompt = 'Hello!'
print('User 2:', prompt)
print('Bot: ', end='', flush=True)
scnet.print_predict(prompt=prompt, stream=True, interval=0.05)
print()
# as there is a relationship between the answers below and the previous dialogue and there is no separation between users, the chat will just look for the last answer given in general to answer the current prompt, regardless of who answered
scnet = SCNet()
scnet.loadModel('./models/user-model', False)
prompt = 'Yes.' # considers that the last prompt was 'Hello!'
print('User 1:', prompt)
print('Bot: ', end='', flush=True)
scnet.print_predict(prompt=prompt, stream=True, interval=0.05)
print()

scnet = SCNet()
scnet.loadModel('./models/user-model', False)
prompt = 'No.' # considers that the last prompt was 'Yes.'
print('User 2:', prompt) # understands that there is a single user sending all the prompts
print('Bot: ', end='', flush=True)
scnet.print_predict(prompt=prompt, stream=True, interval=0.05) # incorrect answer
print()

```
```bash
User 1: Hello!
Bot: Hello everything is fine?

User 2: Hello!
Bot: Hello everything is fine?

User 1: Yes.
Bot: Very good, I'm glad to know.

User 2: No.
Bot: What is the main language spoken in Spain? 
 The main language spoken in Spain is Spanish.

```
Now notice that there will be no errors if we set a different id value for each user. There is no limit to the number of users interacting simultaneously as long as each one has their own loading id in the "user_id" parameter.
```python
from scnetwork import SCNet
# please note that when we do set a user id for the loading, the chat know what the current user's last response was
scnet = SCNet(user_id=1)
scnet.loadModel('./models/user-model', False)

prompt = 'Hello!'
print('User 1:', prompt)
print('Bot: ', end='', flush=True)
scnet.print_predict(prompt=prompt, stream=True, interval=0.05)
print()

scnet = SCNet(user_id=2)
scnet.loadModel('./models/user-model', False)
prompt = 'Hello!'
print('User 2:', prompt)
print('Bot: ', end='', flush=True)
scnet.print_predict(prompt=prompt, stream=True, interval=0.05)
print()
# as there is a relationship between the answers below and the previous dialogue and there is a separation between users, the chat will just look for the last answer of current user to answer the prompt
scnet = SCNet(user_id=1)
scnet.loadModel('./models/user-model', False)
prompt = 'Yes.' # considers that the last prompt was 'Hello!'
print('User 1:', prompt)
print('Bot: ', end='', flush=True)
scnet.print_predict(prompt=prompt, stream=True, interval=0.05)
print()

scnet = SCNet(user_id=2)
scnet.loadModel('./models/user-model', False)
prompt = 'No.' # considers that the last prompt was 'Hello!'
print('User 2:', prompt) # understands that the current prompt belongs to a different user than the previous prompt
print('Bot: ', end='', flush=True)
scnet.print_predict(prompt=prompt, stream=True, interval=0.05)
print()

```
```bash
User 1: Hello!
Bot: Hello everything is fine?

User 2: Hello!
Bot: Hello everything is fine?

User 1: Yes.
Bot: Very good, I'm glad to know.

User 2: No.
Bot: What a shame, I'm so sorry for you.

```
You can also change the user id with the global variable of the class, this way it will not be necessary to rebuild the class.
```python
from scnetwork import SCNet
# use global variable named "user_id" to change user id through class object
# the numbers in "user_id" can be any integer of your choice and the order does not have to be ascending
scnet = SCNet()
scnet.loadModel('./models/user-model', False)

scnet.user_id = 1 # memorizes the current prompt in user memory 1
prompt = 'Hello!'
print('User 1:', prompt)
print('Bot: ', end='', flush=True)
scnet.print_predict(prompt=prompt, stream=True, interval=0.05)
print()

scnet.user_id = 2 # memorizes the current prompt in user memory 2
prompt = 'Hello!'
print('User 2:', prompt)
print('Bot: ', end='', flush=True)
scnet.print_predict(prompt=prompt, stream=True, interval=0.05)
print()

scnet.user_id = 1 # remembers the last interaction with the user of id number 1
prompt = 'Yes.' # considers that the last prompt was 'Hello!'
print('User 1:', prompt)
print('Bot: ', end='', flush=True)
scnet.print_predict(prompt=prompt, stream=True, interval=0.05)
print()

scnet.user_id = 2 # remembers the last interaction with the user of id number 2
prompt = 'No.' # considers that the last prompt was 'Hello!'
print('User 2:', prompt) # understands that the current prompt belongs to a different user than the previous prompt
print('Bot: ', end='', flush=True)
scnet.print_predict(prompt=prompt, stream=True, interval=0.05)
print()

```
```bash
User 1: Hello!
Bot: Hello everything is fine?

User 2: Hello!
Bot: Hello everything is fine?

User 1: Yes.
Bot: Very good, I'm glad to know.

User 2: No.
Bot: What a shame, I'm so sorry for you.

```
With the "minimum_probability_for_candidates" parameter in the constructor, you can set a minimum probability for more than one training output to be included in the inference response.
```python
from scnetwork import SCNet
# the value assigned to "minimum_probability_for_candidates" in the saved model will also apply to the same model when it is loaded
# the parameter "minimum_probability_for_candidates" takes a floating-point number between 0 and 1, representing the minimum probabilistic percentage candidate for the answer to be included in the composition of the final answer
# the default value for the "minimum_probability_for_candidates" parameter is 0.9 for 90%
scnet = SCNet(minimum_probability_for_candidates=0.5) # minimum_probability_for_candidates=0.5 will save the model so that inference accepts outputs with probability equal to or greater than 0.5
scnet.train(dataset_path='dataset.txt') # trains the model with a text file; we recommend that you use values lower than 0.8 in "minimum_probability_for_candidates" only if the dataset is not structured
scnet.saveModel('./my_models/my_model') # saves the model with the value assigned to "minimum_probability_for_candidates"

```
```bash
Reading data: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 1/1 [00:00<00:00, 12787.51it/s]
Embedding data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████| 1000000/1000000 [00:00<00:00, 17050362.81it/s]
Tokenizing data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 100.16it/s]
Training model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 116.32it/s]
Generating 500U parameters: 100%|███████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 24990.89it/s]
Saving model: 100%|██████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 5/5 [00:00<00:00, 3996.86it/s]
```
Now see the model's inference including candidate responses with probability equal to or greater than 0.5.
```python
from scnetwork import SCNet
scnet = SCNet()
# loads the model with the value of "minimum_probability_for_candidates" equal to 0.5
# "minimum_probability_for_candidates" equal to 0.5 will allow any output with probability equal to or greater than 0.5 to be part of the answer
scnet.loadModel('./my_models/my_model', False)
scnet.print_predict(prompt='Talk about AI.', stream=True, interval=0.01)

```
```bash
What is natural language processing (NLP)? 
 It is a field of AI that enables machines to understand, interpret, and generate human language, used in machine translation, chatbots, and voice assistants. 
 
 How is AI used in everyday life? 
 AI is present in virtual assistants (Siri, Alexa), recommendation systems (Netflix, Spotify), facial recognition, autonomous cars, medical diagnostics, chatbots, and much more. 
 
 What is a generative AI model? 
 It is a type of AI that can create new content, such as images, text, and music, based on patterns learned from large amounts of data. Examples include ChatGPT and DALL·E. 
 
 What is the difference between weak AI and strong AI? 
 Weak AI (or narrow AI) is designed for specific tasks, such as virtual assistants and recommendation systems. Strong AI, on the other hand, would have self-awareness and the ability to understand and reason about ai subject like a human being. 
 
 Can AI replace humans in the workforce? 
 AI can automate repetitive and analytical tasks, but it is unlikely to fully replace humans in creative, emotional, and critical thinking jobs. 
 
 What is Machine Learning and how is it related to AI? 
 Machine Learning is a subfield of AI that teaches machines to learn patterns from data without explicit programming. It allows systems to improve their performance over time with experience.
```
You can also change the value of "minimum_probability_for_candidates" via its global variable in the class constructor
```python
from scnetwork import SCNet

scnet = SCNet()
scnet.train(dataset_path='dataset.txt')
# you can use the global variable "minimum_probability_for_candidates" to change the minimum probability of candidates from any point in your code
scnet.minimum_probability_for_candidates = 0.5 # will make the saved model have 0.5 minimum probability in "minimum_probability_for_candidates"
scnet.saveModel('./my_models/my_model') # the saved model will use the last value assigned to "minimum_probability_for_candidates" before saving

```
```bash
Reading data: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 1/1 [00:00<00:00, 12633.45it/s]
Embedding data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████| 1000000/1000000 [00:00<00:00, 16774599.16it/s]
Tokenizing data: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 98.59it/s]
Training model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 114.53it/s]
Generating 500U parameters: 100%|███████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 22815.80it/s]
Saving model: 100%|██████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 5/5 [00:00<00:00, 3125.41it/s]
```
You can also use the global variable "minimum_probability_for_candidates" to override the value of this parameter in the loaded model.
```python
from scnetwork import SCNet
scnet = SCNet()
# loads the model with the value of "minimum_probability_for_candidates" equal to 0.5
scnet.loadModel('./my_models/my_model', False)
prompt = 'Talk about AI.'
print('Prediction 1:')
scnet.print_predict(prompt=prompt, stream=True, interval=0.01)
print()
scnet.minimum_probability_for_candidates = 0.9 # changes the value from 0.5 to 0.9 before prediction
print('Prediction 2:')
scnet.print_predict(prompt=prompt, stream=True, interval=0.01)

```
```bash
Prediction 1:
What is a generative AI model? 
 It is a type of AI that can create new content, such as images, text, and music, based on patterns learned from large amounts of data. Examples include ChatGPT and DALL·E. 
 
 What is the difference between weak AI and strong AI? 
 Weak AI (or narrow AI) is designed for specific tasks, such as virtual assistants and recommendation systems. Strong AI, on the other hand, would have self-awareness and the ability to understand and reason about ai subject like a human being. 
 
 How is AI used in everyday life? 
 AI is present in virtual assistants (Siri, Alexa), recommendation systems (Netflix, Spotify), facial recognition, autonomous cars, medical diagnostics, chatbots, and much more. 
 
 What is Machine Learning and how is it related to AI? 
 Machine Learning is a subfield of AI that teaches machines to learn patterns from data without explicit programming. It allows systems to improve their performance over time with experience. 
 
 What is natural language processing (NLP)? 
 It is a field of AI that enables machines to understand, interpret, and generate human language, used in machine translation, chatbots, and voice assistants. 
 
 Can AI replace humans in the workforce? 
 AI can automate repetitive and analytical tasks, but it is unlikely to fully replace humans in creative, emotional, and critical thinking jobs.

Prediction 2:
What is the difference between weak AI and strong AI? 
 Weak AI (or narrow AI) is designed for specific tasks, such as virtual assistants and recommendation systems. Strong AI, on the other hand, would have self-awareness and the ability to understand and reason about ai subject like a human being.
```
If the name parameter "show_errors" is set to True, any errors that occur in the code will be displayed, otherwise if "show_errors" is set to False, errors will be suppressed. The default value is True.
```python
from scnetwork import SCNet
# below only the first error will be displayed
# show_errors=True enables error display
scnet = SCNet(show_errors=True)
# an error will occur because there are non-numeric elements in one of the vectors
scnet.vectorComparison(input_vector=[1, 2], vector_matrix=[[1, 2], ['a', 'b']])
# show_errors=False denables error display
scnet = SCNet(show_errors=False)
# an error will occur because there are non-numeric elements in one of the vectors
scnet.vectorComparison(input_vector=[1, 2], vector_matrix=[[1, 2], ['a', 'b']])

```
```bash
ERROR in __alternativeVectorComparison: unsupported operand type(s) for -: 'str' and 'int'
```
Now check out an example with all the class constructor parameters in their respective orders and with their respective default values.
```python
from scnetwork import SCNet

scnet = SCNet(device=None, user_id=0, parallel_prediction=False, minimum_probability_for_candidates=0.9, show_errors=True)
scnet.loadModel('./models/user-model')
scnet.print_predict(prompt='Who discovered Brazil?', stream=True, interval=0.05)

```
```bash
Loading model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 4117.83it/s]
Brazil was discovered by Portuguese navigators led by Pedro Alvares Cabral in 1500.
```

## Methods
### Construtor: SCNet
Parameters
| Name                               | Description                                                                                                                                                                                       | Type  | Default Value     |
|------------------------------------|---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------|-------|-------------------|
| device                             | execution device ('cpu', 'gpu' or 'tpu')                                                                                                                                                          | str   | None              |
| user_id                            | user identification number in chat                                                                                                                                                                | int   | 0                 |
| parallel_prediction                | enables inference with parallel processing using True, or disables it using False                                                                                                                 | bool  | False             |
| minimum_probability_for_candidates | minimum probability between 0 and 1 for a training output to be included in the response                                                                                                          | float | 0.9               |
| show_errors                        | enables display of errors in runtime with True, or disables it with False                                                                                                                         | bool  | True              |

### loadJSON (function return type: dict/list): Returns the contents of a converted JSON to a Python variable.
Parameters
| Name                               | Description                                                                                                                                                                                       | Type  | Default Value     |
|------------------------------------|---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------|-------|-------------------|
| file_path                          | path to a local or web json file                                                                                                                                                                  | str   | ''                |
| string_content                     | in the absence of a local or web file, use this parameter to directly assign the json content                                                                                                     | str   | ''                |
| key_name                           | if empty it returns the full content, if defined it will return only the content of the respective key                                                                                            | str   | ''                |

### normalization (function return type: str): Returns the normalized input text in lowercase, without accents and punctuation and reduces the size of large tokens.
Parameters
| Name                               | Description                                                                                                                                                                                       | Type  | Default Value     |
|------------------------------------|---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------|-------|-------------------|
| input_text                         | content of the text to be normalized                                                                                                                                                              | str   | ''                |

### textForEmbedding (function return type: list): Returns a text transformed into a vector of embeddings.
Parameters
| Name                               | Description                                                                                                                                                                                       | Type  | Default Value     |
|------------------------------------|---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------|-------|-------------------|
| text                               | text that will be converted to a numeric vector                                                                                                                                                   | str   | ''                |
| length                             | pre-set length of the numeric vector, or None to use the actual length                                                                                                                            | int   | None              |
| encoder                            | numeric vector coding algorithm: 'gpt2'/'gpt-2', 'gpt-3'/'gpt-3.5'/'gpt-3.5-turbo', 'gpt-4', 'gpt-4-turbo', 'text-davinci-001', 'text-davinci-002' or 'text-davinci-003'                          | str   | 'gpt-4'           |

### embeddingForText (function return type: str): Returns a vector of embeddings converted to text.
Parameters
| Name                               | Description                                                                                                                                                                                       | Type  | Default Value     |
|------------------------------------|---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------|-------|-------------------|
| embedding                          | numeric list that has previously been converted from text to a vector of embeddings                                                                                                               | list  | []                |
| encoder                            | numeric vector coding algorithm: 'gpt2'/'gpt-2', 'gpt-3'/'gpt-3.5'/'gpt-3.5-turbo', 'gpt-4', 'gpt-4-turbo', 'text-davinci-001', 'text-davinci-002' or 'text-davinci-003'                          | str   | 'gpt-4'           |
| strip                              | enables the removal of lateral spaces with True, or disables it with False                                                                                                                        | bool  | True              |

### countTokens (function return type: int): Returns the total number of tokens contained in a text.
Parameters
| Name                               | Description                                                                                                                                                                                       | Type  | Default Value     |
|------------------------------------|---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------|-------|-------------------|
| text                               | input text that will have its tokens counted                                                                                                                                                      | str   | ''                |
| encoder                            | numeric vector coding algorithm: 'gpt2'/'gpt-2', 'gpt-3'/'gpt-3.5'/'gpt-3.5-turbo', 'gpt-4', 'gpt-4-turbo', 'text-davinci-001', 'text-davinci-002' or 'text-davinci-003'                          | str   | 'gpt-4'           |

### truncateTokens (function return type: str): Returns a input text with a given amount of tokens removed from its end.
Parameters
| Name                               | Description                                                                                                                                                                                       | Type  | Default Value     |
|------------------------------------|---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------|-------|-------------------|
| text                               | input text that will have its final tokens truncated                                                                                                                                              | str   | ''                |
| max_tokens                         | maximum number of tokens in the text returned after truncation                                                                                                                                    | int   | 1000000           |
| encoder                            | numeric vector coding algorithm: 'gpt2'/'gpt-2', 'gpt-3'/'gpt-3.5'/'gpt-3.5-turbo', 'gpt-4', 'gpt-4-turbo', 'text-davinci-001', 'text-davinci-002' or 'text-davinci-003'                          | str   | 'gpt-4'           |

### textSummary (function return type: str): Returns an input text with its summarized content.
Parameters
| Name                               | Description                                                                                                                                                                                       | Type  | Default Value     |
|------------------------------------|---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------|-------|-------------------|
| text                               | input text that will have its content summarized                                                                                                                                                  | str   | ''                |
| max_tokens                         | maximum number of tokens in the text returned after summarized                                                                                                                                    | int   | 1000000           |
| encoder                            | numeric vector coding algorithm: 'gpt2'/'gpt-2', 'gpt-3'/'gpt-3.5'/'gpt-3.5-turbo', 'gpt-4', 'gpt-4-turbo', 'text-davinci-001', 'text-davinci-002' or 'text-davinci-003'                          | str   | 'gpt-4'           |

### getTrainingText (function return type: str): Returns the text used in training and fine-tuning the model.
Parameters
| Name                               | Description                                                                                                                                                                                       | Type  | Default Value     |
|------------------------------------|---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------|-------|-------------------|
| max_tokens                         | maximum number of tokens in returned text                                                                                                                                                         | int   | None              |

### alternativeVectorComparison (function return type: tuple): Returns the index and vector of an array that are closest to an input vector using the percentage difference between embeddings.
Parameters
| Name                               | Description                                                                                                                                                                                       | Type  | Default Value     |
|------------------------------------|---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------|-------|-------------------|
| input_vector                       | vector that will be compared to the rows of the matrix                                                                                                                                            | list  | []                |
| vector_matrix                      | matrix that will have its rows compared to the vector                                                                                                                                             | list  | []                |

### vectorComparison (function return type: tuple): Returns the index and vector of an array that are closest to an input vector using cosines similarity.
Parameters
| Name                               | Description                                                                                                                                                                                       | Type  | Default Value     |
|------------------------------------|---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------|-------|-------------------|
| input_vector                       | vector that will be compared to the rows of the matrix                                                                                                                                            | list  | []                |
| vector_matrix                      | matrix that will have its rows compared to the vector                                                                                                                                             | list  | []                |
| stream                             | enables reading the matrix using streaming with True, or disables it with False                                                                                                                   | bool  | True              |

### similarityBetweenVectors (function return type: float): Returns a probability between 0 and 1 with the level of similarity between two embeddings vectors considering the difference in value and position between them.
Parameters
| Name                               | Description                                                                                                                                                                                       | Type  | Default Value     |
|------------------------------------|---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------|-------|-------------------|
| first_vector                       | first numeric vector of comparison                                                                                                                                                                | list  | []                |
| second_vector                      | second numeric vector of comparison                                                                                                                                                               | list  | []                |

### euclideanSimilarity (function return type: float): Returns a probability between 0 and 1 with the level of similarity between two embeddings vectors considering only the difference in value between elements that occupy the same position.
Parameters
| Name                               | Description                                                                                                                                                                                       | Type  | Default Value     |
|------------------------------------|---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------|-------|-------------------|
| first_vector                       | first numeric vector of comparison                                                                                                                                                                | list  | []                |
| second_vector                      | second numeric vector of comparison                                                                                                                                                               | list  | []                |

### overwriteWithHexadecimal (function return type: bool): Returns True if the file is converted into a binary file of hexadecimals, or False otherwise.
Parameters
| Name                               | Description                                                                                                                                                                                       | Type  | Default Value     |
|------------------------------------|---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------|-------|-------------------|
| file_path                          | path to the file to be converted                                                                                                                                                                  | str   | ''                |

### restoreOriginalFromHexadecimal (function return type: bool): Returns True if a file previously converted to binary with hexadecimals is restored, or False otherwise.
Parameters
| Name                               | Description                                                                                                                                                                                       | Type  | Default Value     |
|------------------------------------|---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------|-------|-------------------|
| file_path                          | path to the file to be converted                                                                                                                                                                  | str   | ''                |

### outputAdaptedToInput (function return type: str): Returns a string with an answer adapted to a question.
Parameters
| Name                               | Description                                                                                                                                                                                       | Type  | Default Value     |
|------------------------------------|---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------|-------|-------------------|
| prompt                             | user prompt                                                                                                                                                                                       | str   | ''                |
| original_input                     | prompt with the user input                                                                                                                                                                        | str   | ''                |
| original_output                    | response that will should be adapted to the user's prompt                                                                                                                                         | str   | ''                |

### downloadFile (function return type: bool): Returns True if a download is successful, or False otherwise.
Parameters
| Name                               | Description                                                                                                                                                                                       | Type  | Default Value     |
|------------------------------------|---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------|-------|-------------------|
| url_path                           | online web address for the public file to be downloaded                                                                                                                                           | str   | ''                |
| file_path                          | destination path of downloaded file                                                                                                                                                               | str   | ''                |

### addHiddenLayer (function return type: bool): Returns True if a hidden layer is successfully added to the network architecture, or False otherwise.
Parameters
| Name                               | Description                                                                                                                                                                                       | Type  | Default Value     |
|------------------------------------|---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------|-------|-------------------|
| num_neurons                        | number of neurons that the hidden layer that is being built will have                                                                                                                             | int   | 0                 |

### train (function return type: bool): Returns True if the neural network training is successful, or False otherwise.
Parameters
| Name                               | Description                                                                                                                                                                                       | Type  | Default Value     |
|------------------------------------|---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------|-------|-------------------|
| dataset_path                       | local path or a web path to a file in text or json format that will be used as the data source for training the model                                                                             | str   | ''                |
| string                             | text with all training data or with data to complement the content of the txt/json file                                                                                                           | str   | ''                |
| precision                          | percentage between 0 and 1 of elements of the embeddings vector that will remain                                                                                                                  | float | 1.0               |
| tokenizer                          | numeric vector coding algorithm (same as encoder): 'gpt2'/'gpt-2', 'gpt-3'/'gpt-3.5'/'gpt-3.5-turbo', 'gpt-4', 'gpt-4-turbo', 'text-davinci-001', 'text-davinci-002' or 'text-davinci-003'        | str   | 'gpt-4'           |
| context_window                     | size of the context window in the user input prompt (use float('inf') to maintain an infinite context window)                                                                                     | int   | float('inf')      |
| end_tag                            | subtext used as separator of input and output pairs in the training dataset                                                                                                                       | str   | '\n\n'            |
| validate                           | percentage between 0 and 1 of the training data that will be separated for validation                                                                                                             | float | 0.0               |
| progress                           | if True enables the progress bars that displays the current stage of training, if False will hide the progress bars                                                                               | bool  | True              |

### saveModel (function return type: bool): Returns True if the model is saved successfully, or False otherwise.
Parameters
| Name                               | Description                                                                                                                                                                                       | Type  | Default Value     |
|------------------------------------|---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------|-------|-------------------|
| model_path                         | local path to the directory where the model should be generated, followed by the name used for the model files                                                                                    | str   | ''                |
| progress                           | if True enables the progress bars that displays the current stage of saving, if False will hide the progress bars                                                                                 | bool  | True              |

### loadModel (function return type: bool): Returns True if the model is loaded successfully, or False otherwise.
Parameters
| Name                               | Description                                                                                                                                                                                       | Type  | Default Value     |
|------------------------------------|---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------|-------|-------------------|
| model_path                         | path to the directory where the model files are located, seguido do nome de um arquivo ".scnet" ou ".sccon"                                                                                       | str   | ''                |
| progress                           | if True enables the progress bar that displays the current stage of loading, if False will hide the progress bar                                                                                  | bool  | True              |

### addFit (function return type: bool): Returns True if an of "input/prompt/question" and "output/answer" pair is successfully added to the current model, or False otherwise.
Parameters
| Name                               | Description                                                                                                                                                                                       | Type  | Default Value     |
|------------------------------------|---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------|-------|-------------------|
| prompt                             | input text for a given response to be obtained                                                                                                                                                    | str   | ''                |
| answer                             | response to be returned when the input is sent for model prediction                                                                                                                               | str   | ''                |
| id                                 | unique identifier for the current pair                                                                                                                                                            | int   | 0                 |
| relationship_id                    | identifier of some other pair that will be related to the current pair                                                                                                                            | int   | 0                 |

### train (function return type: bool): Returns True if training with the HurNet network is successful, or False otherwise.
Parameters
| Name                               | Description                                                                                                                                                                                       | Type  | Default Value     |
|------------------------------------|---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------|-------|-------------------|
| interaction                        | if True, enable interactive computation for subtle pattern abstraction, if False, disable                                                                                                         | bool  | True              |
| activation_function                | string with the name of the activation function to be used for non-linear data abstraction                                                                                                        | str   | 'linear'          |
| bias                               | positive or negative floating number used to add bias to the network                                                                                                                              | float | 0                 |
| learning_rate                      | number multiplied by the network weights to control the learning speed                                                                                                                            | float | 1                 |
| stochastic_factor                  | if True, adds randomness to the weights, if False, keep the original weights                                                                                                                      | bool  | False             |
| progress                           | if True enables the progress bars that displays the current stage of adjustment, if False will hide the progress bars                                                                             | bool  | True              |

### predict (function return type: str/generator object): Returns a string with the response to the user prompt.
Parameters
| Name                               | Description                                                                                                                                                                                       | Type  | Default Value     |
|------------------------------------|---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------|-------|-------------------|
| prompt                             | prompt with the input for which a response is desired                                                                                                                                             | str   | ''                |
| max_tokens                         | maximum number of tokens in the returned response                                                                                                                                                 | int   | None              |
| min_fit_probability                | minimum probability between 0 and 1 for a fine-tuning response to be returned                                                                                                                     | float | 0.7               |
| min_probability                    | minimum probability between 0 and 1 for a training response to be returned, if this probability is not reached, the return will be an empty string                                                | float | 0.01              |
| generalization                     | if True, enables generalization of the response based on the user prompt, if False, does not generalize                                                                                           | bool  | True              |
| stream                             | if True, enables the return of a token generator that will return one token at a time, if False, returns all tokens in a single response                                                          | bool  | False             |
| interval                           | if greater than zero, will wait an interval in seconds between one token and another when "stream" is activated                                                                                   | float | 0.0               |

### print_predict (method without return): Displays a string with the model prediction/inference result.
Parameters
| Name                               | Description                                                                                                                                                                                       | Type  | Default Value     |
|------------------------------------|---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------|-------|-------------------|
| prompt                             | prompt with the input for which a response is desired                                                                                                                                             | str   | ''                |
| max_tokens                         | maximum number of tokens in the displayed response                                                                                                                                                | int   | None              |
| min_fit_probability                | minimum probability between 0 and 1 for a fine-tuning response to be displayed                                                                                                                    | float | 0.7               |
| min_probability                    | minimum probability between 0 and 1 for a training response to be displayed, if this probability is not reached, the display will be an empty string                                              | float | 0.01              |
| generalization                     | if True, enables generalization of the response based on the user prompt, if False, does not generalize                                                                                           | bool  | True              |
| stream                             | if True, enables the display of a token generator that will display one token at a time, if False, displays all tokens in a single response                                                       | bool  | False             |
| interval                           | if greater than zero, will wait an interval in seconds between one token and another when "stream" is activated                                                                                   | float | 0.0               |

Check out now a comparison between a conventional Transformer architecture model and our SCNet architecture with the HurNet network.
```python
with open('./dataset.txt', 'r', encoding='utf-8') as file: content = file.read() # reading the contents of the text file
from scnetwork import SCNet # import the scnet class
count_tokens = SCNet().countTokens(text=content) # counts the number of tokens in the training dataset
print(f'The current dataset has {count_tokens} tokens.') # displays the total number of tokens in the current dataset

```
```bash
The current dataset has 451 tokens.
```
```bash
pip install torch tiktoken
```
```python
# this is a code of a transformer algorithm for gpt models; it belongs to sapiens technology® and its unauthorized use by third parties is strictly prohibited
# !pip install torch tiktoken or !pip install torch==2.4.1 and then !pip install tiktoken==0.4.0
class GPTModel: # main class for a standard pre-trained generative transformer model
    def __init__(self): # gpt model architecture builder
        from torch.utils.data import Dataset, DataLoader
        from torch import nn, triu, ones
        from torch.nn import Module, functional as F, utils
        from torch import no_grad, tensor, int64, multinomial, cat, save, load
        from json import load as json_load
        from tiktoken import get_encoding
        from torch import optim
        from tqdm import tqdm
        from os import path as os_path, makedirs as os_makedirs
        from torch import cuda, device, backends        
        if cuda.is_available(): local_device = device('cuda')
        elif backends.mps.is_available(): local_device = device('mps')
        else: local_device = device('cpu')
        self.__Dataset = Dataset
        self.__Module = Module
        self.__nn = nn
        self.__tensor = tensor
        self.__triu = triu
        self.__ones = ones
        self.__no_grad = no_grad
        self.__device = local_device
        self.__F = F
        self.__int64 = int64
        self.__multinomial = multinomial
        self.__cat = cat
        self.__json_load = json_load
        self.__get_encoding = get_encoding
        self.__DataLoader = DataLoader
        self.__optim = optim
        self.__utils = utils
        self.__tqdm = tqdm
        self.__os_path = os_path
        self.__os_makedirs = os_makedirs
        self.__save = save
        self.__load = load
        self.__model = None
        self.__encode = None
        self.__block_size = 500
        self.__decode = None
        self.__string = ''
        self.__vocab_size = 0
        self.__char_to_idx = {}
        self.__idx_to_char = {}
        self.__tokenizer = 'sapi'
        self.__batch_size = 32
        self.__embedding_dim = 384
        self.__number_heads = 6
        self.__number_layers = 6
        self.__dropout = 0.1
        self.__optimizer = None
        self.__learning_rate = 3e-4
        self.__eval_interval = 500
        self.__train = False
        self.parameters_number = 0
        class TextDataset(self.__Dataset): # class for processing training data
            def __init__(self, data={}, block_size=0): self.data, self.block_size = data, block_size
            def __len__(self): return len(self.data) - self.block_size
            def __getitem__(self, index=0):
                input_sequence = self.data[index:index + self.block_size]
                target_sequence = self.data[index + 1:index + self.block_size + 1]
                return input_sequence, target_sequence
        self.__TextDataset = TextDataset
        class Transformer(self.__Module): # building transformer architecture
            def __init__(self, outer=None, vocab_size=0, embedding_dim=0, number_heads=0, number_layers=0, dropout=None, block_size=0):
                super().__init__()
                self.outer = outer
                self.embedding = outer._GPTModel__nn.Embedding(vocab_size, embedding_dim)
                self.pos_encoder = outer._GPTModel__nn.Parameter(outer._GPTModel__tensor([]).new_zeros(1, block_size, embedding_dim))
                self.transformer = outer._GPTModel__nn.TransformerDecoder(outer._GPTModel__nn.TransformerDecoderLayer(d_model=embedding_dim, nhead=number_heads, dropout=dropout), num_layers=number_layers)
                self.fc_out = outer._GPTModel__nn.Linear(embedding_dim, vocab_size)
                self.dropout = outer._GPTModel__nn.Dropout(dropout)
                self.block_size = block_size
            def forward(self, input_tensor=[]):
                outer = self.outer
                batch_size, seq_len = input_tensor.size()
                positions = self.pos_encoder[:, :seq_len, :].to(input_tensor.device)
                embedded = self.dropout(self.embedding(input_tensor) + positions)
                transposed = embedded.transpose(0, 1)
                mask = outer._GPTModel__triu(outer._GPTModel__ones(seq_len, seq_len, device=input_tensor.device) * float('-inf'), diagonal=1)
                output = self.transformer(transposed, transposed, tgt_mask=mask)
                output = output.transpose(0, 1)
                return self.fc_out(output)
        self.__Transformer = Transformer
    def __compute_loss(self, loader=[]): # function for computing network loss rate
        self.__model.eval()
        total_loss = 0
        with self.__no_grad():
            for input_batch, target_batch in loader:
                input_batch, target_batch = input_batch.to(self.__device), target_batch.to(self.__device)
                logits = self.__model(input_batch)
                loss = self.__F.cross_entropy(logits.view(-1, logits.size(-1)), target_batch.view(-1))
                total_loss += loss.item()
        return total_loss / len(loader)
    def __format_params(self, number_params=0): # function for formatting the number of network parameters
        if number_params < 1_000: return f'{number_params}U'
        elif number_params < 1_000_000: return f'{number_params // 1_000}K'
        elif number_params < 1_000_000_000: return f'{number_params // 1_000_000}M'
        elif number_params < 1_000_000_000_000: return f'{number_params // 1_000_000_000}B'
        else: return f'{number_params // 1_000_000_000_000}T'
    def __generate_tokens(self, prompt='', max_tokens=500, temperature=1.0): # function to generate the predicted tokens in the inference
        self.__model.eval()
        encoded_prompt = self.__encode(prompt)
        input_tensor = self.__tensor(encoded_prompt, dtype=self.__int64).unsqueeze(0).to(self.__device)
        with self.__no_grad():
            tokens_generated = 0
            while True:
                conditioned_input = input_tensor[:, -self.__block_size:] if input_tensor.size(1) > self.__block_size else input_tensor
                logits = self.__model(conditioned_input)
                logits = logits[:, -1, :] / temperature
                probs = self.__F.softmax(logits, dim=-1)
                next_token = self.__multinomial(probs, num_samples=1)
                input_tensor = self.__cat((input_tensor, next_token), dim=1)
                token = next_token.item()
                decoded_token = self.__decode([token])
                if tokens_generated == 0 and '\n' in decoded_token: continue
                tokens_generated += 1
                yield decoded_token
                if (tokens_generated >= max_tokens and decoded_token[-1] in {'.', '\n', '!', '?', ';'}) or (tokens_generated >= (max_tokens*2)): break
    def train(self, dataset_path='', precision=0.5, tokenizer='sapi', context_window=500, progress=True): # function for training a conventional transformer model
        try:
            """
                Arguments:
                    dataset_path: receives a string with the address of a txt or json file for model training
                    precision: receives a float with a target value for the precision of weights adjustment in backpropagation; backpropagation will only stop if this target is reached
                    tokenizer: receives a string with the value 'sapi' to use sapiens' tokenizer, or 'gpt' to use the generative pre-trained transformer tokenizer
                    context_window: receives an integer with the limit for the context window to be created for the model
                    progress: receives a boolean indicating whether the progress bar will be displayed or hidden
            """
            dataset_path = str(dataset_path).strip()
            precision = float(precision) if type(precision) in (bool, int, float) else 0.5
            tokenizer = str(tokenizer).lower().strip()
            context_window = max((1, int(context_window))) if type(context_window) in (bool, int, float) else 500
            progress = bool(progress) if type(progress) in (bool, int, float) else True
            self.__block_size = context_window
            loss_limit = max(0, min(1, 1 - precision))
            is_txt, is_json, formatted_sequences = dataset_path.endswith('.txt'), dataset_path.endswith('.json'), []
            if not is_txt and not is_json: raise ValueError('Unsupported file format. Use .txt or .json')
            if is_txt:
                with open(dataset_path, 'r', encoding='utf-8') as file: text_data = file.read()
            elif is_json:
                with open(dataset_path, 'r', encoding='utf-8') as file: json_data = self.__json_load(file)
                if type(json_data) == dict:
                    data_key = list(json_data.keys())[0]
                    pairs = json_data[data_key]
                else: pairs = json_data
                formatted_sequences = [str(pair[list(pair.keys())[0]]+'\n'+pair[list(pair.keys())[1]]).strip() for pair in pairs]
                text_data = '\n\n'.join(formatted_sequences)
            if len(self.__string) > 0: text_data += '\n\n'+self.__string
            text_data = text_data.strip()
            if tokenizer == 'sapi':
                chars = sorted(list(set(text_data)))
                self.__vocab_size = len(chars)
                self.__char_to_idx = {char: index for index, char in enumerate(chars)}
                self.__idx_to_char = {index: char for index, char in enumerate(chars)}
                self.__encode = lambda string: [self.__char_to_idx[char] for char in string]
                self.__decode = lambda indices: ''.join([self.__idx_to_char[index] for index in indices])
            else:
                encode = self.__get_encoding('gpt2')
                self.__vocab_size = encode.n_vocab
                self.__encode = encode.encode
                self.__decode = encode.decode
            data = self.__tensor(self.__encode(text_data), dtype=self.__int64)
            split_point = int(0.8 * len(data))
            train_data = data[:split_point]
            val_data = data[split_point:]
            if len(train_data) < self.__block_size:
                if len(train_data) > 1: self.__block_size = len(train_data) - 1
                else: raise ValueError('Dataset too small for training. Add more data.')
            self.__tokenizer = tokenizer
            train_dataset = self.__TextDataset(train_data, self.__block_size)
            val_dataset = self.__TextDataset(val_data, self.__block_size)
            train_loader = self.__DataLoader(train_dataset, batch_size=self.__batch_size, shuffle=True)
            val_loader = self.__DataLoader(val_dataset, batch_size=self.__batch_size, shuffle=False)
            self.__model = self.__Transformer(self, self.__vocab_size, self.__embedding_dim, self.__number_heads, self.__number_layers, self.__dropout, self.__block_size).to(self.__device)
            self.__optimizer = self.__optim.AdamW(self.__model.parameters(), lr=self.__learning_rate)
            scheduler = self.__optim.lr_scheduler.ReduceLROnPlateau(self.__optimizer, mode='min', factor=0.5, patience=3)
            epoch, step, best_val_loss = 0, 0, float('inf')            
            while True:
                self.__model.train()
                total_train_loss = 0
                str_epoch = str(epoch+1).rjust(10, '0')
                for input_batch, target_batch in train_loader:
                    input_batch, target_batch = input_batch.to(self.__device), target_batch.to(self.__device)
                    logits = self.__model(input_batch)
                    loss = self.__F.cross_entropy(logits.view(-1, logits.size(-1)), target_batch.view(-1))
                    self.__optimizer.zero_grad()
                    loss.backward()
                    self.__utils.clip_grad_norm_(self.__model.parameters(), 1.0)
                    self.__optimizer.step()
                    total_train_loss += loss.item()
                    if step > 0 and step % self.__eval_interval == 0:
                        val_loss = self.__compute_loss(val_loader)
                        scheduler.step(val_loss)
                        if val_loss < best_val_loss: best_val_loss = val_loss
                    step += 1
                avg_train_loss = total_train_loss / len(train_loader)
                if avg_train_loss <= loss_limit:
                    if progress: print()
                    break
                elif progress:
                    format_str = '{desc}: {percentage:3.0f}%|{bar:10}| {n_fmt}/{total_fmt} [{elapsed}<{remaining}, {rate_fmt}]'
                    current_precision = max(0, min(1, 1 - avg_train_loss))
                    str_current_precision = f'{current_precision:.4f}'.ljust(5, '0')
                    str_precision = f'{precision:.4f}'.ljust(5, '0')
                    train_loader = self.__tqdm(train_loader, desc=f'Epoch {str_epoch} - current precision is {str_current_precision}; aiming for precision >= {str_precision} in training', bar_format=format_str)
                epoch += 1
            return True
        except Exception as error:
            print('ERROR in train: ' + str(error))
            return False
    def saveModel(self, model_path='', progress=True): # function to save a pre-trained model
        try:
            """
                Arguments:
                    model_path: receives a string with the address and name of the model file to be generated
                    progress: receives a boolean indicating whether the progress bar will be displayed or hidden
            """
            model_path = str(model_path).strip()
            progress = bool(progress) if type(progress) in (bool, int, float) else True
            if self.__model is None: raise ValueError('Model is not initialized. Call train or loadModel first.')
            number_params = sum(p.numel() for p in self.__model.parameters())
            formatted_params = self.__format_params(number_params)
            if isinstance(model_path, str):
                directory, file_name = self.__os_path.split(model_path)
                if not file_name: file_name = 'model.gpt'
                elif not file_name.endswith('.gpt'): file_name += '.gpt'
            else: directory, file_name = str(model_path), 'model.gpt'
            if directory and not self.__os_path.exists(directory): self.__os_makedirs(directory)
            save_path = self.__os_path.join(directory, file_name)
            save_dict = {'model_state_dict': self.__model.state_dict(), 'tokenizer': self.__tokenizer, 'vocab_size': self.__vocab_size, 'block_size': self.__block_size}
            if self.__tokenizer == 'sapi': save_dict['char_to_idx'], save_dict['idx_to_char'] = self.__char_to_idx, self.__idx_to_char
            if progress:
                for _ in self.__tqdm(range(10), desc=f'Saving model with {formatted_params} parameters', leave=False): self.__save(save_dict, save_path)
            else: self.__save(save_dict, save_path)
            self.__train = True
            return True
        except Exception as error:
            print('ERROR in saveModel: ' + str(error))
            return False
    def loadModel(self, model_path='', progress=True): # function to load a previously saved pre-trained model
        try:
            """
                Arguments:
                    model_path: receives a string with the address and name of the model file to be loaded
                    progress: receives a boolean indicating whether the progress bar will be displayed or hidden
            """
            model_path = str(model_path).strip()
            progress = bool(progress) if type(progress) in (bool, int, float) else True
            if len(model_path) > 0:
                directory, file_name = self.__os_path.split(model_path)
                if not file_name: file_name = 'model.gpt'
                elif not file_name.endswith('.gpt'): file_name += '.gpt'
            else: directory, file_name = str(model_path), 'model.gpt'
            model_file = self.__os_path.join(directory, file_name)
            if progress:
                for _ in self.__tqdm(range(10), desc='Loading model', leave=False): checkpoint = self.__load(model_file, map_location=self.__device)
            else: checkpoint = self.__load(model_file, map_location=self.__device)
            self.__tokenizer = checkpoint['tokenizer']
            self.__vocab_size = checkpoint['vocab_size']
            self.__block_size = checkpoint['block_size']
            if self.__tokenizer == 'sapi':
                self.__char_to_idx = checkpoint['char_to_idx']
                self.__idx_to_char = checkpoint['idx_to_char']
                self.__encode = lambda string: [self.__char_to_idx[char] for char in string]
                self.__decode = lambda indices: ''.join([self.__idx_to_char[index] for index in indices])
            else:
                encode = self.__get_encoding('gpt2')
                self.__encode = encode.encode
                self.__decode = encode.decode
            self.__model = self.__Transformer(self, self.__vocab_size, self.__embedding_dim, self.__number_heads, self.__number_layers, self.__dropout, self.__block_size).to(self.__device)
            self.__model.load_state_dict(checkpoint['model_state_dict'])
            number_params = sum(p.numel() for p in self.__model.parameters())
            self.parameters_number, self.__optimizer, self.__train = number_params, None, True
            return True
        except Exception as error:
            print('ERROR in loadModel: ' + str(error))
            return False
    def addFit(self, prompt='', answer=''): # function to add fine-tuning to a dataset before training, or to a previously loaded model
        try:
            """
                Arguments:
                    prompt: receives a string with the input sample to be added to the current model
                    answer: receives a string with the output sample to be added to the current model
            """
            prompt = str(prompt).strip()
            answer = str(answer).strip()
            if not self.__train: self.__string += prompt+'\n'+answer+'\n\n'
            else:
                if self.__model is None: raise ValueError('Model is not initialized. Call train or loadModel first.')
                if self.__optimizer is None: self.__optimizer = self.__optim.AdamW(self.__model.parameters(), lr=self.__learning_rate)
                formatted = prompt+'\n'+answer+'\n\n'
                encoded = self.__encode(formatted)
                if len(encoded) > self.__block_size: encoded = encoded[:self.__block_size]
                input_tensor = self.__tensor(encoded[:-1], dtype=self.__int64).unsqueeze(0).to(self.__device)
                target_tensor = self.__tensor(encoded[1:], dtype=self.__int64).unsqueeze(0).to(self.__device)
                self.__model.train()
                logits = self.__model(input_tensor)
                loss = self.__F.cross_entropy(logits.view(-1, logits.size(-1)), target_tensor.view(-1))
                self.__optimizer.zero_grad()
                loss.backward()
                self.__utils.clip_grad_norm_(self.__model.parameters(), 1.0)
                self.__optimizer.step()
            return True
        except Exception as error:
            print('ERROR in addFit: ' + str(error))
            return False
    def predict(self, prompt='', max_tokens=500, stream=False): # function to return the inference result
        try:
            """
                Arguments:
                    prompt: receives a string with the input for which an output is desired
                    max_tokens: receives an integer with an approximate number for the maximum tokens to be generated in the response
                    stream: receives a boolean indicating whether the response will be returned token by token or all at once
            """
            prompt = str(prompt).strip()
            max_tokens = max((1, int(max_tokens))) if type(max_tokens) in (bool, int, float) else 500
            stream = bool(stream) if type(stream) in (bool, int, float) else False
            if self.__model is None: raise ValueError('Model is not initialized. Call train or loadModel first.')
            if stream: return self.__generate_tokens(prompt, max_tokens)
            tokens = list(self.__generate_tokens(prompt, max_tokens))
            return ''.join(tokens)
        except Exception as error:
            print('ERROR in predict: ' + str(error))
            return ''
    def print_predict(self, prompt='', max_tokens=500, stream=False): # method to display the inference result
        try:
            """
                Arguments:
                    prompt: receives a string with the input for which an output is desired
                    max_tokens: receives an integer with an approximate number for the maximum tokens to be generated in the response
                    stream: receives a boolean indicating whether the response will be displayed token by token or all at once
            """
            prompt = str(prompt).strip()
            max_tokens = max((1, int(max_tokens))) if type(max_tokens) in (bool, int, float) else 500
            stream = bool(stream) if type(stream) in (bool, int, float) else False
            if self.__model is None: raise ValueError('Model is not initialized. Call train or loadModel first.')
            if stream:
            	[print(token, end='', flush=True) for token in self.__generate_tokens(prompt, max_tokens)]
            	print()
            else: print(self.predict(prompt, stream=False))
        except Exception as error:
            print('ERROR in print_predict: ' + str(error))
# this is a code of a transformer algorithm for gpt models; it belongs to sapiens technology® and its unauthorized use by third parties is strictly prohibited

```
```python
# ---> insert here the code of the GPTModel class defined above...
from time import time # import time module
start = time() # marks the initial time

transformer = GPTModel() # instantiation of the transformer class for the gpt model
# training the model with a text-based dataset using achieving 97% accuracy and the gpt tokenizer and with a context window of 200 tokens
transformer.train(dataset_path='./dataset.txt', precision=0.97, tokenizer='gpt', context_window=200, progress=True)
transformer.saveModel(model_path='./transformer_model.gpt', progress=True) # saving the pre-trained generative model

end = time() # marks the end time
time_spent_transformer = abs(start-end) # calculates the time difference between the start and end of execution
print(f'\nRuntime: {time_spent_transformer} seconds.') # displays the total time spent in seconds

```
Runned on a Macbook M3 Max with 48GB of VRAM.
```bash
Epoch 0000000001 - current precision is 0.0000; aiming for precision >= 0.9700 in training: 100%|██████████| 6/6 [00:02<00:00,  2.82it/s]
Epoch 0000000002 - current precision is 0.0000; aiming for precision >= 0.9700 in training: 100%|██████████| 6/6 [00:01<00:00,  3.08it/s]
Epoch 0000000003 - current precision is 0.0000; aiming for precision >= 0.9700 in training: 100%|██████████| 6/6 [00:01<00:00,  3.05it/s]
Epoch 0000000004 - current precision is 0.0000; aiming for precision >= 0.9700 in training: 100%|██████████| 6/6 [00:01<00:00,  3.11it/s]
Epoch 0000000005 - current precision is 0.0000; aiming for precision >= 0.9700 in training: 100%|██████████| 6/6 [00:01<00:00,  3.11it/s]
Epoch 0000000006 - current precision is 0.0000; aiming for precision >= 0.9700 in training: 100%|██████████| 6/6 [00:01<00:00,  3.11it/s]
Epoch 0000000007 - current precision is 0.0000; aiming for precision >= 0.9700 in training: 100%|██████████| 6/6 [00:01<00:00,  3.11it/s]
Epoch 0000000008 - current precision is 0.2517; aiming for precision >= 0.9700 in training: 100%|██████████| 6/6 [00:01<00:00,  3.11it/s]
Epoch 0000000009 - current precision is 0.4527; aiming for precision >= 0.9700 in training: 100%|██████████| 6/6 [00:01<00:00,  3.11it/s]
Epoch 0000000010 - current precision is 0.5855; aiming for precision >= 0.9700 in training: 100%|██████████| 6/6 [00:01<00:00,  3.11it/s]
Epoch 0000000011 - current precision is 0.6756; aiming for precision >= 0.9700 in training: 100%|██████████| 6/6 [00:01<00:00,  3.12it/s]
Epoch 0000000012 - current precision is 0.7414; aiming for precision >= 0.9700 in training: 100%|██████████| 6/6 [00:01<00:00,  3.10it/s]
Epoch 0000000013 - current precision is 0.7851; aiming for precision >= 0.9700 in training: 100%|██████████| 6/6 [00:01<00:00,  3.11it/s]
Epoch 0000000014 - current precision is 0.8246; aiming for precision >= 0.9700 in training: 100%|██████████| 6/6 [00:01<00:00,  3.11it/s]
Epoch 0000000015 - current precision is 0.8532; aiming for precision >= 0.9700 in training: 100%|██████████| 6/6 [00:01<00:00,  3.02it/s]
Epoch 0000000016 - current precision is 0.8700; aiming for precision >= 0.9700 in training: 100%|██████████| 6/6 [00:01<00:00,  3.10it/s]
Epoch 0000000017 - current precision is 0.8881; aiming for precision >= 0.9700 in training: 100%|██████████| 6/6 [00:01<00:00,  3.11it/s]
Epoch 0000000018 - current precision is 0.9040; aiming for precision >= 0.9700 in training: 100%|██████████| 6/6 [00:01<00:00,  3.06it/s]
Epoch 0000000019 - current precision is 0.9201; aiming for precision >= 0.9700 in training: 100%|██████████| 6/6 [00:01<00:00,  3.10it/s]
Epoch 0000000020 - current precision is 0.9299; aiming for precision >= 0.9700 in training: 100%|██████████| 6/6 [00:01<00:00,  3.06it/s]
Epoch 0000000021 - current precision is 0.9327; aiming for precision >= 0.9700 in training: 100%|██████████| 6/6 [00:01<00:00,  3.10it/s]
Epoch 0000000022 - current precision is 0.9291; aiming for precision >= 0.9700 in training: 100%|██████████| 6/6 [00:01<00:00,  3.05it/s]
Epoch 0000000023 - current precision is 0.9442; aiming for precision >= 0.9700 in training: 100%|██████████| 6/6 [00:01<00:00,  3.10it/s]
Epoch 0000000024 - current precision is 0.9536; aiming for precision >= 0.9700 in training: 100%|██████████| 6/6 [00:01<00:00,  3.10it/s]
Epoch 0000000025 - current precision is 0.9588; aiming for precision >= 0.9700 in training: 100%|██████████| 6/6 [00:01<00:00,  3.10it/s]
Epoch 0000000026 - current precision is 0.9639; aiming for precision >= 0.9700 in training: 100%|██████████| 6/6 [00:01<00:00,  3.09it/s]
Epoch 0000000027 - current precision is 0.9681; aiming for precision >= 0.9700 in training: 100%|██████████| 6/6 [00:01<00:00,  3.08it/s]
Epoch 0000000028 - current precision is 0.9691; aiming for precision >= 0.9700 in training: 100%|██████████| 6/6 [00:01<00:00,  3.08it/s]

                                                                                                                                                                     
Runtime: 60.74804615974426 seconds.
```
```python
from time import time # import time module
start = time() # marks the initial time

from scnetwork import SCNet # import of the module semantic class
scnet = SCNet() # instantiation of the semantic class for the scnet model
# training the model with a text-based dataset using the gpt tokenizer
scnet.train(dataset_path='./dataset.txt', progress=True)
scnet.saveModel(model_path='./scnet_model/scnet_file', progress=True) # saving the pre-trained generative model

end = time() # marks the end time
time_spent_scnet = abs(start-end) # calculates the time difference between the start and end of execution
print(f'\nRuntime: {time_spent_scnet} seconds.') # displays the total time spent in seconds

```
Runned on a Macbook M3 Max with 48GB of VRAM.
The **SCNet** network was **109 times** faster in training than a conventional Transformer network using the same dataset.
```bash
Reading data: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 1/1 [00:00<00:00, 12985.46it/s]
Embedding data: 100%|████████████████████████████████████████████████████████████████████████████████████████████████| 1000000/1000000 [00:00<00:00, 15804303.10it/s]
Tokenizing data: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 92.96it/s]
Training model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 108.02it/s]
Generating 500U parameters: 100%|████████████████████████████████████████████████████████████████████████████████████████████████████| 6/6 [00:00<00:00, 5136.93it/s]
Saving model: 100%|██████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 5/5 [00:00<00:00, 2117.69it/s]

Runtime: 0.5549192428588867 seconds.
```
Check below the difference between the times spent on fine-tuning training with the two architectures.
```python
# ---> insert here the code of the GPTModel class defined above...
from time import time # import time module
start = time() # marks the initial time

transformer = GPTModel() # instantiation of the transformer class for the gpt model
transformer.loadModel(model_path='./transformer_model.gpt', progress=True) # loading the pre-trained generative model
# sample data for fine-tuning
transformer.addFit(prompt='What does the acronym GPT mean in language models?', answer='GPT in language models stands for Generative Pre-trained Transformer.')
transformer.addFit(prompt='Tell me what is the meaning of GPT in language models.', answer='GPT in language models stands for Generative Pre-trained Transformer.')
transformer.addFit(prompt='What is the capital of Spain?', answer='The capital of Spain is Madrid.')
transformer.addFit(prompt='Tell me the name of the capital of Spain.', answer='The capital of Spain is Madrid.')

transformer.saveModel(model_path='./adjusted_transformer.gpt', progress=True) # saving the pre-trained generative model

end = time() # marks the end time
time_spent_transformer = abs(start-end) # calculates the time difference between the start and end of execution
print(f'\nRuntime: {time_spent_transformer} seconds.') # displays the total time spent in seconds

```
```bash
                                                                                                                                                                     
Runtime: 5.164402961730957 seconds.
```
```python
from time import time # import time module
start = time() # marks the initial time

from scnetwork import SCNet # import of the module semantic class
scnet = SCNet() # instantiation of the semantic class for the scnet model
scnet.loadModel(model_path='./scnet_model/scnet_file', progress=True) # loading the pre-trained generative model
# sample data for fine-tuning
scnet.addFit(prompt='What does the acronym GPT mean in language models?', answer='GPT in language models stands for Generative Pre-trained Transformer.')
scnet.addFit(prompt='Tell me what is the meaning of GPT in language models.', answer='GPT in language models stands for Generative Pre-trained Transformer.')
scnet.addFit(prompt='What is the capital of Spain?', answer='The capital of Spain is Madrid.')
scnet.addFit(prompt='Tell me the name of the capital of Spain.', answer='The capital of Spain is Madrid.')
# does not require saving

end = time() # marks the end time
time_spent_scnet = abs(start-end) # calculates the time difference between the start and end of execution
print(f'\nRuntime: {time_spent_scnet} seconds.') # displays the total time spent in seconds

```
Runned on a Macbook M3 Max with 48GB of VRAM.
The **SCNet** network was **10 times** faster in training with fine-tuning than a conventional Transformer network using the same dataset.
```bash
Loading model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 3727.32it/s]

Runtime: 0.5296740531921387 seconds.
```
Now see an inference test with the generated models.
```python
# ---> insert here the code of the GPTModel class defined above...
from time import time # import time module
start = time() # marks the initial time

transformer = GPTModel() # instantiation of the transformer class for the gpt model
transformer.loadModel(model_path='./transformer_model.gpt', progress=True) # loading the pre-trained generative model

prompt = 'What are the main types of machine learning?' # prompt for inference test
transformer.print_predict(prompt=prompt, max_tokens=50, stream=True) # infers token by token up to approximately 50 tokens

end = time() # marks the end time
time_spent_transformer = abs(start-end) # calculates the time difference between the start and end of execution
print(f'\nRuntime: {time_spent_transformer} seconds.') # displays the total time spent in seconds

```
Runned on a Macbook M3 Max with 48GB of VRAM.
```bash
The three main types are:                                                                                                                                            
- **Supervised Learning**: The model learns from labeled data.
- **Unsupervised Learning**: The model identifies patterns without labels.
- **Reinforcement Learning**: The model learns through trial and error, receiving rewards or penalties.

Runtime: 4.530112981796265 seconds.
```
```python
from time import time # import time module
start = time() # marks the initial time

from scnetwork import SCNet # import of the module semantic class
scnet = SCNet() # instantiation of the semantic class for the scnet model
scnet.loadModel(model_path='./scnet_model/scnet_file', progress=True) # loading the pre-trained generative model

prompt = 'What are the main types of machine learning?' # prompt for inference test
scnet.print_predict(prompt=prompt, stream=True) # infers all required tokens one by one

end = time() # marks the end time
time_spent_scnet = abs(start-end) # calculates the time difference between the start and end of execution
print(f'\nRuntime: {time_spent_scnet} seconds.') # displays the total time spent in seconds

```
Runned on a Macbook M3 Max with 48GB of VRAM.
The **SCNet** network was **9 times** faster in inferring than a conventional Transformer network using the same dataset.
```bash
Loading model: 100%|█████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 7/7 [00:00<00:00, 3519.56it/s]
The three main types are:
- **Supervised Learning**: The model learns from labeled data.
- **Unsupervised Learning**: The model identifies patterns without labels.
- **Reinforcement Learning**: The model learns through trial and error, receiving rewards or penalties.

Runtime: 0.5193250179290771 seconds.
```

## Contributing

We do not accept contributions that may result in changing the original code.

Make sure you are using the appropriate version.

## License

This is proprietary software and its alteration and/or distribution without the developer's authorization is not permitted.
